

#if !defined(__THQ_defs_h__)
#define __THQ_defs_h__


	template <typename T, typename TI>
	inline T atomicloadadd(T& var, TI val)
	{
		return __atomic_fetch_add(&var, val, __ATOMIC_SEQ_CST);
	}

	template <typename T, typename TI>
	inline T atomicaddload(T& var, TI val)
	{
		return __atomic_fetch_add(&var, val, __ATOMIC_SEQ_CST);
	}

	template <typename T>
	inline T atomicload(T& var)
	{
		return __atomic_load_n(&var, __ATOMIC_SEQ_CST);
	}

	template <typename T, typename TI>
	inline void atomicstore(T& var, TI val)
	{
		__atomic_store_n(&var, val, __ATOMIC_SEQ_CST);
		return;
	}

	template <typename T, typename TI>
	inline T atomicxchg(T& var, TI val)
	{
		return __atomic_exchange_n(&var, val, __ATOMIC_SEQ_CST);
	}

namespace THQsupport
{
#if !defined(THQCPPVER)
#define LSLEEP
#if defined(LSLEEP)
	static void THQFATTR2
	NSsleepf(double nanosecsd)
	{
		double secsd = nanosecsd * 1.0e-9;
		int secsi64;
		asm ("cvttsd2si %1,%0" : "=r"(secsi64) : "xm"(secsd));
		int64_t secsdt = secsi64;
		double secslessthan1 = secsd - secsdt;
		double nsecslessthan1 = secslessthan1 * 1.0e9;
		int64_t nsecs = nsecslessthan1;

		struct timespec ts = {(time_t) secsdt, (long) nsecs};
		struct timespec tsrem = ts;
		struct timespec tsnext = ts;
		
		::fprintf(stderr,"NSsleepf arguments: ts.tv_sec: %ld, ts.tv_nsec: %ld\n",
			(long int) ts.tv_sec, (long int) ts.tv_nsec);

		while( nanosleep(&tsrem, &tsnext) != 0)
		{
			int eval = errno;
			if (eval == EINTR)
			{
				tsrem = tsnext;
			}
			else
			{
				perror("NSsleepf -- nanosleep");
				break;
			}
		}
	}

	static void THQFATTR2
	NSsleep(int64_t nanosecs)
	{
		int64_t secs = nanosecs / 1000000000 ;
		int64_t nsecs = nanosecs - (secs * 1000000000);

		struct timespec ts = {(time_t) secs, (long) nsecs};
		struct timespec tsrem = ts;
		struct timespec tsnext = ts;

		while( nanosleep(&tsrem, &tsnext) != 0)
		{
			int eval = errno;
			if (eval == EINTR)
			{
				tsrem = tsnext;
			}
			else
			{
				perror("NSsleep -- nanosleep");
				break;
			}
		}
	}

	static inline void
	USsleep(int32_t usecs)
	{
		int64_t nsecs = usecs * 1000;
		NSsleep(nsecs);
	}

	static inline void
	MSsleep(int32_t mssarg)
	{
		int64_t nsecs = mssarg * 1000000;
		NSsleep(nsecs);
	}

#endif
#if defined(LSLEEP)
#if defined(USETIMEOFDAY)
	static inline bool
	gettime(timespec* ts)
	{
		struct timeval tv;
		if (gettimeofday(&tv, nullptr))
			return false;

		ts->tv_sec = tv.tv_sec;
		ts->tv_nsec = tv.tv_usec * 1000;
		return true;
	}

#else
	static inline bool
	gettime(timespec* ts)
	{
		if( clock_gettime(CLOCK_REALTIME, ts))
			return false;
		return true;
	}
#endif
#endif

#if !defined(ONWINDOWS)
	static timespec THQFATTR2
	Ltimeradd_ms(const timespec* ts, int offsetmsec)
	{
		timespec localts = *ts;

		localts.tv_nsec += offsetmsec * 1000000;
		if (localts.tv_nsec >= 1000000000)
		{
			int64_t secadd = localts.tv_nsec / 1000000000;
			localts.tv_nsec %= 1000000000;
			localts.tv_sec += secadd;
		}

		return localts;
	}

	static timespec THQFATTR2
	Ltimeradd_ns(const timespec* ts, int offsetnsec)
	{
		timespec localts = *ts;

		localts.tv_nsec += offsetnsec;
		if (localts.tv_nsec >= 1000000000)
		{
			int64_t secadd = localts.tv_nsec / 1000000000;
			localts.tv_nsec %= 1000000000;
			localts.tv_sec += secadd;
		}

		return localts;
	}

	static timespec THQFATTR2
	Ltimeradd(const timespec* ts, const timespec* tsoffset)
	{
		timespec localts = *ts;

		localts.tv_nsec += tsoffset->tv_nsec;
		localts.tv_sec += tsoffset->tv_sec;
		if (localts.tv_nsec >= 1000000000)
		{
			int64_t secadd = localts.tv_nsec / 1000000000;
			localts.tv_nsec %= 1000000000;
			localts.tv_sec += secadd;
		}

		return localts;
	}

	static timespec THQFATTR2
	gettime_with_offset_ns(int offsetnsec)
	{
		timespec tstmp;

		if (gettime(&tstmp))
		{
			return (timespec) {0,0};
		}

		return Ltimeradd_ns(&tstmp, offsetnsec);
	}

	static timespec THQFATTR2
	gettime_with_offset_ms(int offsetmsec)
	{
		timespec tstmp;

		if (gettime(&tstmp))
		{
			return (timespec) {0,0};
		}

		return Ltimeradd_ms(&tstmp, offsetmsec);
	}

#endif
#endif
};

#endif
