//
//
// Copyright(C) 2017,2018,219 John S Dyson
// All rights reserved.
//
// Refer to audioproc.cpp for license
// 

#include <sys/types.h>

#if !defined(__AGCFILTERS__H__)
#define __AGCFILTERS__H__ 1
//#include <math.h>

#define FIRARRAYTYPE static const

#include "itqA.h"
#include "varray.h"
#include "vecmath.h"
#include "accuminfo.h"
#include "itq-qdefs.h"

#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Wconversion"

#pragma clang diagnostic pop

#define AGCFNOINLINE __attribute__((always_inline)) //__attribute__((noinline))
#define AGCFNOINLINE0 __attribute__((noinline))

template <typename VDEF>
inline varrayT<VDEF>
sratetmod(varrayT<VDEF> tval)
{
	return Vselect((tval == varrayT<VDEF>(1.0)), varrayT<VDEF>(1.0), tval);
}

template <typename T>
inline T
sratetmod(T tval)
{
	return (tval == T(1.0)) ? T(1.0) : (tval * T(aspec.SRATETMULT));
}

template <typename T, typename CT = double>
class hpfT
{
	T coeff;
	T zval;

public:
	hpfT(void) : coeff(), zval() {}

	hpfT(const hpfT<T, CT>& init) : coeff(init.coeff), zval(init.zval) {}

	hpfT(T c, T z = T{}): coeff(c), zval(z) {}
	
	hpfT(CT c, CT z = CT(0)) : hpfT(T(c), T(z)) {}

	hpfT<T, CT>&
	operator=(hpfT<T, CT> src)
	{
		coeff = src.coeff;
		zval = src.zval;
		return *this;
	}

	T
	next(T in)
	{
		T out = in - zval;

		zval = zval + out * coeff;

		return out;
	}

	T
	operator()(T in)
	{
		return next(in);
	}

	void
	setcoeff(T newcoeff)
	{
		coeff = newcoeff;
	}

	void
	setcoeff(CT newcoeff)
	{
		coeff = T(newcoeff);
	}
};

class hpfAs
{
	float coeff;
	Asample zval;

public:
	hpfAs(float c = 0.0f) :
		coeff(c), zval(0.0f, 0.0f)
	{}

	Asample
	next(const Asample& in)
	{
		Asample out = in - zval;

		zval = zval + out * coeff;

		return out;
	}

	Asample
	operator()(const Asample& in)
	{
		return next(in);
	}
};

class hpf4As
{
	float coeff[4];
	Asample zval[4];

public:

	hpf4As(float c0, float c1, float c2, float c3) :
		coeff(), zval()
	{
		for (int i = 0; i < 4; i++)
		{
			zval[i] = Asample{};
		}
		coeff[0] = c0;
		coeff[1] = c1;
		coeff[2] = c2;
		coeff[3] = c3;
	}

	hpf4As(float c0 = 0.0f) :
		hpf4As(c0, c0, c0, c0) {}

	hpf4As(varrayT<TYPEvd4x1> cvals) :
		hpf4As(float(cvals[0]), float(cvals[1]), float(cvals[2]), float(cvals[3])) {}

	hpf4As(varrayT<TYPEvf4x1> cvals) :
		hpf4As(cvals[0], cvals[1], cvals[2], cvals[3]) {}

	void
	setcoeffs(float c0, float c1, float c2, float c3)
	{
		coeff[0] = c0;
		coeff[1] = c1;
		coeff[2] = c2;
		coeff[3] = c3;
	}
		
	void
	next(const Asample* in, Asample* out)
	{
		for (int i = 0; i < 4; i++)
		{
			out[i] = in[i] - zval[i];
			zval[i] = zval[i] + out[i] * coeff[i];
		}
		return;
	}

	void
	operator()(const Asample* in, Asample* out)
	{
		next(in, out);
		return;
	}
};

template <int NFLT>
class XsfdnfAs GCQ
{
	static constexpr int REALNFLT = ((NFLT + 1) / 2) * 2;
	//static constexpr int REALNFLT = NFLT;
	int inptr;
	float rawbuf0[REALNFLT*2];
	float rawbuf1[REALNFLT*2];
	const float * filter;

	void
	newdata(const Asample& in)
	{
		float tmp = in.item0();
		rawbuf0[inptr] = tmp;
		rawbuf0[inptr+NFLT] = tmp;

		tmp = in.item1();
		rawbuf1[inptr] = tmp;
		rawbuf1[inptr+NFLT] = tmp;
		inptr += 1;
		if (inptr >= NFLT)
			inptr = 0;
	}

	Asample 
	basicfilt(void) const
	{
		
		float out0 = 0.0f, out1 = 0.0f;
		const float * rbptr0 = &rawbuf0[inptr];
		const float * rbptr1 = &rawbuf1[inptr];
		for (int i = 0; i < REALNFLT; i++)
		{
			float f = filter[i];
			out0 += rbptr0[i] * f;
			out1 += rbptr1[i] * f;
		}
		return Asample{out0, out1};
	}


public:
	XsfdnfAs(const float *filtera = nullptr) AGCFNOINLINE : inptr(0), rawbuf0(), rawbuf1(), filter(filtera) 
	{
		for (int i = 0; i < REALNFLT*2; i++)
		{
			rawbuf0[i] = 0.0f;
			rawbuf1[i] = 0.0f;
		}
	}

	void setfilt(const float *filtera)
	{
		filter = filtera;
	}

	Asample AGCFNOINLINE
	next(const Asample& input)
	{
		newdata(input);
		return basicfilt();
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}


};

template <int NFLT>
class sfdnfAs GCQ
{
	static constexpr int REALNFLT = ((NFLT + 1) / 2) * 2;
	float __attribute__((aligned(32))) rawbuf[REALNFLT*4];
	const float * __attribute__((aligned(32))) filter;

	Asample prev;
	int inptr;

	void
	newdata(const Asample& in)
	{
		float tmp0 = in.item0();
		float tmp1 = in.item1();
		rawbuf[inptr] = tmp0;
		rawbuf[inptr+1] = tmp1;
		rawbuf[inptr+NFLT*2] = tmp0;
		rawbuf[inptr+1+NFLT*2] = tmp1;

		inptr += 2;
		if (inptr >= NFLT*2)
			inptr = 0;
	}

	Asample 
	basicfilt(void)
	{
		
		float out0 = 0.0f, out1 = 0.0f;
		const float * rbptr = &rawbuf[inptr];
		for (int i = 0; i < REALNFLT; i += 1 )
		{
			float f = filter[i];
			int dindex = i * 2;
			out0 += rbptr[dindex] * f;
			out1 += rbptr[dindex+1] * f;
		}
#if 0
		if (NFLT & 1)
		{
			Asample rtval = prev;
			prev = Asample(out0, out1);
			return rtval;
		}
#endif
		return Asample{out0, out1};
	}


public:
	sfdnfAs(const float * __attribute__((aligned(32))) filtera = nullptr) AGCFNOINLINE : rawbuf(), filter(filtera), prev(), inptr(0)
	{
		for (int i = 0; i < REALNFLT*4; i++)
		{
			rawbuf[i] = 0.0f;
		}
	}

	void setfilt(const float *filtera)
	{
		filter = filtera;
	}

	Asample AGCFNOINLINE
	next(const Asample& input)
	{
		newdata(input);
		return basicfilt();
	}

	template <typename PV>
	PV
	operator()(const PV& input)
	{
		return next(input);
	}


};

template <int NFLT, typename VDEF=TYPEvd4x1>
class sfdnfVa
{
	using VT=varrayT<VDEF>;
	using FT=typename VDEF::FT;
	VT prev;
	VT rawbuf[NFLT*2];
	const float * filter;
	int inptr;

	void
	newdata(VT in)
	{
		int nextinptr = inptr + 1;
		if (nextinptr >= NFLT)
			nextinptr = 0;

		rawbuf[inptr] = in;
		rawbuf[inptr+NFLT] = in;

		inptr = nextinptr;
	}

	VT 
	basicfilt(void) 
	{
		int i;
		VT fout = VT(FT(0.0));
		int rbbase = inptr;
		const VT * rbptr = (const VT *) &rawbuf[rbbase];
		for (i = 0; i < NFLT; i += 4)
		{
			fout += rbptr[i] * filter[i];
			fout += rbptr[i+1] * filter[i+1];
			fout += rbptr[i+2] * filter[i+2];
			fout += rbptr[i+3] * filter[i+3];
		}
		for (int j = i; j < NFLT; j++)
		{
			fout += rbptr[j] * filter[j];
		}
		VT rtval;
#if 0
		if (NFLT & 1)
		{
			rtval = prev;
			prev = fout;
		}
		else
#endif
		{
			rtval = fout;
		}
			
		return rtval;
	}


public:
	sfdnfVa(const float * const filtera = nullptr) AGCFNOINLINE : prev(FT(0)), inptr(0), rawbuf(), filter(filtera)
	{
		for (int i = 0; i < NFLT*2; i++)
			rawbuf[i] = VT(FT(0.0));
	}

	VT AGCFNOINLINE
	next(const VT& input)
	{
		newdata(input);
		return basicfilt();
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	void
	setfilt(const float* const newfilt)
	{
		filter = newfilt;
	}
};

template <int NFLT, typename IOTYPE = TYPEvd4x1>
class sfdnfVd4x2X1
{
	using ft = sfdnfVa<NFLT,TYPEvf8x1>;

	using fltv = varrayT<TYPEvf8x1>;

	using iotypev = varrayT<IOTYPE>;
	using iotypeftype = typename IOTYPE::FT;
	

	ft filt;

public:
	sfdnfVd4x2X1(const float * const filtera = nullptr) : filt(filtera)
	{}


	void AGCFNOINLINE
	next(iotypev v0in, iotypev v1in, iotypev& v0out, iotypev& v1out)
	{
		fltv ftmpin {
			float(v0in[0]), float(v0in[1]), float(v0in[2]), float(v0in[3]),
			float(v1in[0]), float(v1in[1]), float(v1in[2]), float(v1in[3])
		};

		fltv ftmpout = filt(ftmpin);

		iotypev dtmpout0 {
			iotypeftype(ftmpout[0]), iotypeftype(ftmpout[1]), iotypeftype(ftmpout[2]), iotypeftype(ftmpout[3])
		};

		iotypev dtmpout1 {
			iotypeftype(ftmpout[4]), iotypeftype(ftmpout[5]), iotypeftype(ftmpout[6]), iotypeftype(ftmpout[7])
		};

		v0out = dtmpout0;
		v1out = dtmpout1;
	}

	
	template <typename PV>
	void
	operator()(const PV& input0, const PV& input1, PV& output0, PV& output1)
	{
		next(input0, input1, output0, output1);
	}


	void
	setfilt(const float* const newfilt)
	{
		filt.setfilt(newfilt);
	}
};
	

template <int MAXDLY, typename DT = float>
class sVdly GCQ
{
	int inptr;
	DT rawbuf[2*MAXDLY];

	void
	newdata(DT in)
	{
		rawbuf[inptr] = in;
		rawbuf[inptr+MAXDLY] = in;
		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdly() : inptr(0), rawbuf()
	{
		for (int i = 0; i < 2*MAXDLY; i++)
		{
			rawbuf[i] = DT{0};
		}
	}

	const DT&
	next(const DT& in)
	{
		newdata(in);
		return rawbuf[inptr];
	}

	DT
	dly(int vdly) const
	{
		return rawbuf[inptr + (MAXDLY-1) - vdly];
	}

	DT
	next(DT in, int vdly)
	{
		newdata(in);
		return dly(vdly);
	}

	DT operator()(const DT& input)
	{
		return next(input);
	}

	DT operator()(const DT& input, int delay)
	{
		return next(input, delay);
	}

	const DT operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

	int maxdelay(void) const {
		return MAXDLY;
	}

};

template <int NDLY>
class sdlyAs GCQ
{
	int inptr;
	Asample rawbuf[(NDLY < 1)?1:NDLY];

	void
	newdata(const Asample& in)
	{
		rawbuf[inptr] = in;
		inptr += 1;
		if (inptr >= NDLY)
			inptr = 0;
	}

public:
	sdlyAs() : inptr(0), rawbuf()
	{
		for (int i = 0; i < NDLY; i++)
		{
			rawbuf[i] = Asample(0.0f,0.0f);
		}
	}

	Asample 
	next(const Asample& in)
	{
		if (NDLY < 1)
		{
			return in;
		}
		else
		{
			newdata(in);
			return rawbuf[inptr];
		}
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	int maxdelay(void) const {
		return NDLY;
	}

};

template <int MAXDLY>
class sVdlyAs GCQ
{
	int inptr;
	Asample rawbuf[2*MAXDLY];

	void
	newdata(const Asample& in)
	{
	//	::memcpy(&rawbuf[inptr], &in, sizeof(Asample));
	//	::memcpy(&rawbuf[inptr+MAXDLY], &in, sizeof(Asample));
		rawbuf[inptr] = in;
		rawbuf[inptr+MAXDLY] = in;
		inptr += 1;
		if (inptr >= MAXDLY)
			inptr = 0;
	}

public:
	sVdlyAs() : inptr(0), rawbuf()
	{
#pragma clang loop unroll(disable)
		for (int i = 0; i < 2*MAXDLY; i++)
		{
			rawbuf[i] = Asample(0.0f,0.0f);
		}
	}

	const Asample&
	next(const Asample& in)
	{
		newdata(in);
		return rawbuf[inptr];
	}

	const Asample& 
	next(const Asample& in, int vdly)
	{
		newdata(in);
		return rawbuf[inptr + (MAXDLY-1) - vdly];
	}

	const Asample&
	dly(int dlyamt) const
	{
		return rawbuf[inptr + (MAXDLY-1) - dlyamt];
	}

	const Asample&
	minusmax(int dlyamt) const
	{
		return dly(MAXDLY - dlyamt);
	}

	const Asample& operator()(const Asample& input)
	{
		return next(input);
	}

	const Asample& operator()(const Asample& input, int delay)
	{
		return next(input, delay);
	}

	const Asample& operator()(int dlyamt) const
	{
		return dly(dlyamt);
	}

	int maxdelay(void) const {
		return MAXDLY;
	}

};

template <typename ST>
void 
calcbilinear2(ST* x, ST *y, ST w0, ST t, ST b0, ST b1, ST b2, ST a0, ST a1, ST a2)
{
	ST k = w0 / tan(w0 * t * ST(0.5));
	//ST k = 2.0 / t;
	//ST k = ST(w0);
	ST k2 = k * k;
	ST denom = a0 * k2 + a1 * k + a2;
	ST invdenom = ST(1.0) / denom;

	x[2] = (b0 * k2 + b1 * k + b2) * invdenom;

	x[1] = (ST(2.0) * b2 - ST(2.0) * b0 * k2) * invdenom;

	x[0] = (b0 * k2 - b1 * k + b2) * invdenom;

	y[1] = ST(-1.0) * (ST(2.0) * a2 - ST(2.0) * a0 * k2) * invdenom;

	y[0] = ST(-1.0) * (a0 * k2 - a1 * k + a2) * invdenom;

#if 0
	if (0 && ac.info)
	{
		fprintf(stderr,"w0: %15.10lg, k: %15.10lg, denom: %15.10lg\n", double(w0), double(k), double(denom));
		fprintf(stderr,"x[0]: %15.10lg, x[1]: %15.10lg, x[2]: %15.10lg\n",
			double(x[0]), double(x[1]), double(x[2]));
		fprintf(stderr,"y[0]: %15.10lg, y[1]: %15.10lg\n",
			double(y[0]), double(y[1]));
	}
#endif
}

class dfIIRHPF
{
	static constexpr int ORDER = 2;
	int finptr;
	int iinptr;

	double Fbuf[(ORDER+1)*2];
	double Ibuf[(ORDER+1)*2];

	double Ffilter[ORDER+1];
	double Ifilter[ORDER];

	double srate;
	double stime;
	double lsterms[7];

	double
	basicfilt(double in) 
	{

		Fbuf[iinptr] = in;
		Fbuf[iinptr+(ORDER + 1)] = in;

		double out = 0.0;

		const double * Fbufptr = &Fbuf[iinptr + (ORDER + 1)];
		for (int i = ORDER; i >= 0; --i)
		{
			out += *(Fbufptr - i) * Ffilter[(ORDER - i)];
		}

		const double * Ibufptr = &Ibuf[iinptr + (ORDER + 1)];
		for (int i = ORDER; i > 0; --i)
		{
			out += *(Ibufptr - i) * Ifilter[(ORDER  - i)];
		}

		Ibuf[iinptr] = out;
		Ibuf[iinptr + (ORDER + 1)] = out;

		iinptr += 1;
		if (iinptr >= (ORDER + 1))
			iinptr = 0;
			
		//return out * a;
		return out;
	}

public:

	dfIIRHPF(const double *sterms) : finptr(0), iinptr(0), Fbuf(), Ibuf(), Ffilter(), Ifilter()
	{
		srate= aspec.inSRATE;
		stime = 1.0 / srate;

		calcbilinear2<double>(Ffilter, Ifilter, sterms[0], stime, sterms[1], sterms[2], sterms[3], sterms[4], sterms[5], sterms[6]);
		for (int i = 0; i < 7; i++)
		{
			lsterms[i] = sterms[i];
		}

		for (int i = 0; i < (ORDER + 1)*2; i++)
		{
			Fbuf[i] = 0.0;
			Ibuf[i] = 0.0;
		}
	}

	void NOINLINE
	initHPF(float initval, int maxiterations = 10000)
	{
		for (int i = 0; i < maxiterations; i++)
		{
			next(initval);
		}
	}

	void NOINLINE
	initHPF(double initval, int maxiterations = 10000)
	{
		for (int i = 0; i < maxiterations; i++)
		{
			next(initval);
		}
	}

	double AGCFNOINLINE
	next(double input)
	{
		return basicfilt(input);
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

};


namespace cfilt
{
	static constexpr double PISQ = M_PI * M_PI;
	static constexpr double PI = M_PI;
	static constexpr double TWOPI = double(2.0) * M_PI;
	static constexpr double TWOPISQ = (TWOPI * TWOPI);

	static constexpr double freqlf = 160.0;
	static constexpr double freqmf = 2600.0;
	static constexpr double freqhf = 8000.0;

//	static constexpr double qlf = 0.840;
//	static constexpr double qmf = 0.707;
//	static constexpr double qhf = 0.707;

	static constexpr double qlf = 0.840;
	static constexpr double qmf = 0.420;
	static constexpr double qhf = 0.707;

	static constexpr double hp0[7] =
	{
		freqlf * TWOPI,
		1.0, 0.0, 0.0,			/* b0, b1, b2 */
		1.0, freqlf * TWOPI / qlf, (freqlf * freqlf * TWOPISQ)  /* a0, a1, a2 */
	};

	static constexpr double hp1[7] =
	{
		freqmf * TWOPI,
		1.0, 0.0, 0.0,					/* b0, b1, b2 */
		1.0, freqmf * TWOPI / qmf, (freqmf * freqmf * TWOPISQ)	/* a0, a1, a2 */
	};

	static constexpr double hp2[7] =
	{
		freqhf * TWOPI,
		1.0, 0.0, 0.0,					/* b0, b1, b2 */
		1.0, freqhf * TWOPI / qhf, (freqhf * freqhf * TWOPISQ) /* a0, a1, a2 */
	};


};


class IBPF
{
	using DFTYPEq=dfIIRHPF;
	DFTYPEq fhp0L, fhp0R;
	DFTYPEq fhp1L, fhp1R;
	DFTYPEq fhp2L, fhp2R;

public:
	IBPF(void) :
		fhp0L(cfilt::hp0),
		fhp0R(cfilt::hp0),

		fhp1L(cfilt::hp1),
		fhp1R(cfilt::hp1),

		fhp2L(cfilt::hp2),
		fhp2R(cfilt::hp2)

	{

	}

	void AGCFNOINLINE
	next( const Asample& inA, Asample* out)
	{

		double rL = double(inA.item0());
		double rR = double(inA.item1());
		double inL = fhp0L(rL);
		double inR = fhp0R(rR);

		double band0L = rL - inL;
		double band0R = rR - inR;

//		double hp1L = fhp1L(inL);
//		double hp1R = fhp1R(inR);

		double hp1L = fhp1L(rL);
		double hp1R = fhp1R(rR);

		double hp2L = fhp2L(rL);
		double hp2R = fhp2R(rR);

//		double band1L = hp1L - hp2L;
//		double band1R = hp1R - hp2R;

//		double band2L = hp2L;
//		double band2R = hp2R;

//		double band1L = hp1L - band0L;
//		double band1R = hp1R - band0R;

#if 1
		double band1L = inL - hp1L;
		double band1R = inR - hp1R;

		double band2L = hp1L;
		double band2R = hp1R;
#else
#if 1
static constexpr double mix21 = 3.0;
static constexpr double mix21div = 1.0 / (mix21 + 1);
		double band1L = (((inL - hp1L) * mix21) + hp1L) * mix21div;
		double band1R = (((inR - hp1R) * mix21) + hp1R) * mix21div;

		double band2L = (hp1L + (band1L * mix21)) * mix21div;
		double band2R = (hp1R + (band1R * mix21)) * mix21div;
#endif
#endif


		out[0] = Asample{float(band0L), float(band0R)} * 1.00f;
		out[1] = Asample{float(band1L), float(band1R)} * 1.00f;
		out[2] = Asample{float(band2L), float(band2R)} * 1.00f;

		return;
	}

	int
	maxdelay(void) const
	{
		return 2;
	}

	void
	operator()(const Asample& in, Asample* out)
	{
		return next(in, out);
	}
};

#endif
//
// vim: set ts=4:
//


