//
//
// Copyright(C) 2017,2018,2019 John S Dyson
// All rights reserved.
//
// Refer to the file audioproc.cpp for license.
// 

#define _GNU_SOURCE  1           /* See feature_test_macros(7) */
#include <sys/types.h>

#include <sys/types.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
// #include <math.h>
#include <ctype.h>
#include <pthread.h>
#include <sched.h>

#if defined(GDEBUG)
#define GC_DEBUG
#endif
#include <malloc.h>
#include <signal.h>

#include <xmmintrin.h>
#include <pmmintrin.h>
#include <immintrin.h>

#include "itqA.h"
#include "itq-typedefs.h"
#include "itq-qdefs.h"
#include "vtype.h"
#include "agcfilters.h"

cq0 demptyqueue;
cqtype0 dinput;
cqtype0 dinputpost;
cqtype0 dfilter[MAXNFPASSES];
cqtype0 dendproc;
cqtype0 doutput;

constexpr int MAXTXTBSIZE=256;

constexpr int NUMTHREADS=4;

int numthreads = 0;

//
// Argument processing and storage (thread common) RO on all threads except main and UI
//

#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Weffc++"
static struct arg_context acINIT;
struct arg_context ac;
#pragma GCC diagnostic pop

extern void* THreadfile(const void *threadinfo);
extern void* THinput(const void *threadinfo);
extern void* THinputpost(const void *threadinfo);
extern void* THfilter(const void *threadinfo);
extern void* THendproc(const void *threadinfo);
extern void* THwritefile(const void *threadinfo);

struct thread_init_list tinit[]=
{
	{THreadfile, 0, "readfile", PTHREAD_MUTEX_INITIALIZER, PTHREAD_COND_INITIALIZER, false, 0},
	{THinput, 0, "inputproc", PTHREAD_MUTEX_INITIALIZER, PTHREAD_COND_INITIALIZER, false, 0},
	{THendproc, 0, "endproc", PTHREAD_MUTEX_INITIALIZER, PTHREAD_COND_INITIALIZER, false, 0},
	{THwritefile, 0, "writefile", PTHREAD_MUTEX_INITIALIZER, PTHREAD_COND_INITIALIZER, false, 0},
};

static pthread_t	allthreads[NUMTHREADS];


audiochunk THQALIGNED preallocaudiochunk[NAUDIOCHUNKSXTRA];

extern void initcqs(void);

#if !defined(WINDOWS64)
namespace affinity
{
	static constexpr int MAXCPUS = 32;
	int ncpus;
	int nnoht;
	bool affinityset;
	cpu_set_t* origaffinity;
	cpu_set_t* noht;
	cpu_set_t* yesht;
};
#endif

//
// setup the affinity masks
//
void
initaffinity(void)
{
#if !defined(WINDOWS64)
	affinity::origaffinity = CPU_ALLOC(affinity::MAXCPUS);
	CPU_ZERO_S(affinity::MAXCPUS, affinity::origaffinity);
	affinity::noht = CPU_ALLOC(affinity::MAXCPUS);
	CPU_ZERO_S(affinity::MAXCPUS, affinity::noht);
	affinity::yesht = CPU_ALLOC(affinity::MAXCPUS);
	CPU_ZERO_S(affinity::MAXCPUS, affinity::yesht);
	affinity::affinityset = false;

	if (ac.affinity == 0)
		return;


	if (sched_getaffinity((pid_t) 0, CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::origaffinity))
	{
		perror("getaffinity");
		return;
	}

	affinity::ncpus = 0;
	for (int i = 0; i < affinity::MAXCPUS; i++)
	{
		if (CPU_ISSET_S(i, CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::origaffinity))
			affinity::ncpus++;
	}

	if ((ac.affinity == -1) && (affinity::ncpus <= 4))
		return;

	CPU_ZERO_S(CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::noht);

	for (int i = 0; i < affinity::ncpus; i++)
	{
		CPU_SET_S(i, CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::yesht);
	}

	affinity::nnoht = ac.affinity;

	if (affinity::nnoht > affinity::MAXCPUS)
		affinity::nnoht = affinity::MAXCPUS;
	else if (affinity::nnoht < 0)
		affinity::nnoht = affinity::ncpus / 2;

	for (int i = 0; i < affinity::nnoht; i++)
	{
		CPU_SET_S(i, CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::noht);
	}

	unsigned long int cpuhtmask = 0;
	unsigned long int cpunohtmask = 0;
	for (int i = 0; i < affinity::ncpus; i++)
		if ( CPU_ISSET_S(i, CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::yesht))
			cpuhtmask |= (1UL << i);

	for (int i = 0; i < affinity::ncpus; i++)
		if ( CPU_ISSET_S(i, CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::noht))
			cpunohtmask |= (1UL << i);

	fprintf(stderr,"HT cpu mask: %lx, NOHT cpu mask: %lx\n", cpuhtmask, cpunohtmask);
#endif
}
	
	
	

//
// Set current thread not to share with other threads
//
void
setnoht(void)
{
#if !defined(WINDOWS64)
	if ((affinity::nnoht == 0) || (ac.affinity == 0))
		return;
	if (sched_setaffinity((pid_t) 0, CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::noht))
		perror("setaffinity -- setnoht");
#endif
}

//
// Set current thread to share with other threads
//
void
setht(void)
{
#if !defined(WINDOWS64)
	if ((affinity::nnoht == 0) || (ac.affinity == 0))
		return;
	if (sched_setaffinity((pid_t) 0, CPU_ALLOC_SIZE(affinity::MAXCPUS), affinity::yesht))
		perror("setaffinity -- setht");
#endif
}

void
doqueueinits(void)
{
	initcqs();
}

void NOINLINE
dumpqueuestatus(void)
{
	demptyqueue.dumpqueuestatus();
	dinput.dumpqueuestatus();
	for (int i = 0; i < MAXNFPASSES; i++)
		dfilter[i].dumpqueuestatus();
	dendproc.dumpqueuestatus();
	doutput.dumpqueuestatus();
	
}

void NOINLINE
initcqs(void)
{
		for (int i = 0; i < NAUDIOCHUNKSXTRA; i++)
		{
			paudiochunk t = &preallocaudiochunk[i];
			t->setindex(i);
			demptyqueue.THQwqspaceandinsert(t);
		}
}


static void NOINLINE
thread_errormessage(int UNUSED index, int UNUSED errorcode)
{
	
	char emsgbuf[emsgbufsize];
	int eno = errno;
	if (errorcode)
		eno = errorcode;
#if !defined(NOSTRERROR)
	char *estr = ::strerror_r(eno, emsgbuf, emsgbufsize - 1);
	::fprintf(stderr,"Thread#%d(\"%s\"), %s\n", index, tinit[index].threadname, estr);
#else
	::fprintf(stderr,"Thread#%d(\"%s\"), %s\n", index, tinit[index].threadname, strerror(eno));
#endif
}

static void NOINLINE
closeallthreads(void)
{
	for (int index = 0; index < NUMTHREADS; index++)
	{
		if (allthreads[index] == pthread_t(0))
			continue;
		thread_init_list *curinit = &tinit[index];
		::pthread_mutex_destroy(&curinit->thread_local_mutex);
		::pthread_cond_destroy(&curinit->thread_local_cond);
		::pthread_kill(allthreads[index], SIGTERM);
	}
}

static void NOINLINE UNUSED
rundownallthreads(void)
{
	for (int index = 0; index < NUMTHREADS; index++)
	{
		if (allthreads[index] == pthread_t(0))
			continue;
		if (!::pthread_kill(allthreads[index], 0))
		{
			int rtsts = 0;
			::pthread_join(allthreads[index], (void **) &rtsts);
			allthreads[index] = pthread_t(0);
		}
	}
}

static void NOINLINE
wait_all_threads(void)
{
	int rtsts = 0;
	int *rtstsp = &rtsts;
	char threadexited[NUMTHREADS] = {0};
	while (1)
	{
		int threadsdone = 0;
		for (int index = 0; index < NUMTHREADS; index++)
		{
			if ((allthreads[index] == pthread_t(0)) || threadexited[index])
			{
				threadsdone++;
				continue;
			}

			if (threadexited[index] || ::pthread_kill(allthreads[index], 0))
			{
				threadsdone++;
				continue;
			}

			if (::pthread_join(allthreads[index], (void **) &rtstsp) == 0)
			{
				threadsdone++;
				threadexited[index] = 1;
				allthreads[index] = pthread_t(0);
			}
		}

		if (threadsdone >= NUMTHREADS)
			return;
	}
}

static void* NOINLINE
thread_start(void *threadarg)
{
	thread_init_list *actualthreadargs = (thread_init_list *) threadarg;
#if 0
	fprintf(stderr,"Trying to start thread: %s\n", actualthreadargs->threadname);
#endif

	//
	// We don't let the thread actually run until 'let_threads_run'
	// is called.  This allows for time consuming setup & preparation.
	//
	::pthread_mutex_lock(&actualthreadargs->thread_local_mutex);
	while (!actualthreadargs->startok)
	{
		::pthread_cond_wait(&actualthreadargs->thread_local_cond, &actualthreadargs->thread_local_mutex);
	}

	::pthread_mutex_unlock(&actualthreadargs->thread_local_mutex);

#if 0
	fprintf(stderr,"Thread released: %s\n", actualthreadargs->threadname);
#endif
	setfpmode();

	return (*actualthreadargs->subr)(actualthreadargs);
}

static void NOINLINE
thread_release(int index)
{
	thread_init_list *curinit = &tinit[index];
	::pthread_mutex_lock(&curinit->thread_local_mutex);
	curinit->startok = true;
	::pthread_cond_broadcast(&curinit->thread_local_cond);
	::pthread_mutex_unlock(&curinit->thread_local_mutex);
}

//
// Shorter version of pthread_create
//
int NOINLINE
create_thread(pthread_t *newthreadid, void *(*thfunc)(void *), void *threadarg)
{
	int rtval;
#if 0
	pthread_attr_t attr;
	if (!pthread_attr_init(&attr))
	{
		size_t oldstacksize = 0;
		size_t oldguardsize = 0;
		
		pthread_attr_getstacksize(&attr, &oldstacksize);
		pthread_attr_getguardsize(&attr, &oldguardsize);
//		::fprintf(stderr,"oldstacksize: %luM, oldguardsize: %luk\n", oldstacksize/(1024*1024), oldguardsize/1024);
		pthread_attr_setstacksize(&attr, 4096 * 65536);
		pthread_attr_setguardsize(&attr, 65536);

		pthread_attr_getstacksize(&attr, &oldstacksize);
		pthread_attr_getguardsize(&attr, &oldguardsize);
//		::fprintf(stderr,"newstacksize: %luM, newguardsize: %luk\n", oldstacksize/(1024*1024), oldguardsize/1024);

//		pthread_attr_setstacksize(&attr, 64 * 65536);
//		pthread_attr_setguardsize(&attr, 65536);

		//pthread_attr_setstacksize(&attr, threadstacksize);
		//pthread_attr_setguardsize(&attr, threadstackguardsize);
		rtval = ::pthread_create(newthreadid, &attr, thfunc, threadarg);
		pthread_attr_destroy(&attr);
		return rtval;
	}
#endif

	rtval = ::pthread_create(newthreadid, nullptr, thfunc, threadarg);
	return rtval;
}

static bool NOINLINE
init_threads(void)
{
	numthreads = NUMTHREADS;
	//
	// Init thread data structures
	//
	for (int index = 0; index < NUMTHREADS; index++)
	{
		thread_init_list *curinit = &tinit[index];
		allthreads[index] = pthread_t(0);
		::pthread_mutex_init(&curinit->thread_local_mutex, nullptr);
		::pthread_cond_init(&curinit->thread_local_cond, nullptr);
		curinit->startok = false;
	}

	//
	// Create and start threads, but hold them up until 'thread_release'
	// lets the threads actually run.
	//
	for (int index = 0; index < NUMTHREADS; index++)
	{
		thread_init_list *curinit = &tinit[index];
		int thrtval = create_thread(&allthreads[index], thread_start, (void *) curinit);
		if (thrtval != 0)
		{
			thread_errormessage(index, thrtval);
			closeallthreads();
			return false;
		}
#if 0
		fprintf(stderr,"Created thread: \"%s\" at %lu, 0x%lx\n", curinit->threadname,  (unsigned long) allthreads[index], allthreads[index]);
#endif
	}

	return true;
}

static void NOINLINE
let_threads_run(void)
{
	for (int index = 0; index < NUMTHREADS; index++)
	{
		thread_release(index);
	}
}

static void NOINLINE
threads_start(void)
{
	init_threads();
	let_threads_run();
}


//
// Command line arguements allow for cloud IP to be a DNS name.   However, due to internals of the
// Socketserver interface, the SS IP must be an IP addr.
//
// Other code examples didn't seem to use getopts -- getopts does too many good/nice things to ignore.
//
enum optcode {
	INPUTCODE, OUTPUTCODE, INFOCODE, CMDCODE, FLOATOUTCODE, RAWFILTCODE,
	HELPCODE, IDCODE, CRATIOCODE,
	INGAINCODE, OUTGAINCODE, ONECODE, GCLIPCODE, LLOGCODE, FULLOUTCODE,
	INGAINLCODE, OUTGAINLCODE,
	INGAINRCODE, OUTGAINRCODE, FILTERTESTCODE, HPFCODE, TONECODE, MAXGAINCODE,
	WIDEFILTCODE, ALLOWWIDEBANDCODE, MAXLEVELCODE,
	FLATTHRESHCODE, FLEVELCODE, AFFINITYCODE, XNRCODE,
	BOGUSCODE=999999
};

struct optcodenamemap
{
	enum optcode opc;
	const char *name;
};

static optcodenamemap UNUSED optcodenamemap[] =
	{
		{ INPUTCODE, "INPUTCODE" }, { OUTPUTCODE, "OUTPUTCODE" }, { WIDEFILTCODE, "WIDEFILTCODE" },
		{ INFOCODE, "INFOCODE" }, { CMDCODE, "CMDCODE" }, { GCLIPCODE, "GCLIPCODE" }, { RAWFILTCODE, "RAWFILTCODE" },
		{ INGAINCODE, "INGAINCODE" }, { FLEVELCODE, "FLEVELCODE" }, { LLOGCODE, "LLOGCODE" },
		{ OUTGAINCODE, "OUTGAINCODE" }, { ONECODE, "ONECODE" }, { HPFCODE, "HPFCODE" }, 
		{ HELPCODE, "HELPCODE" }, { FULLOUTCODE, "FULLOUTCODE" },
		{ IDCODE, "IDCODE" }, { XNRCODE, "XNRCODE" },
		{ INGAINLCODE, "INGAINLCODE" }, { OUTGAINLCODE, "OUTGAINLCODE" }, { MAXGAINCODE, "MAXGAINCODE" },
		{ INGAINRCODE, "INGAINRCODE" }, { OUTGAINRCODE, "OUTGAINRCODE" },
		{ FILTERTESTCODE, "FILTERTESTCODE" }, { FLOATOUTCODE, "FLOATOUTCODE"}, { AFFINITYCODE, "AFFINITYCODE" },
		{ ALLOWWIDEBANDCODE, "ALLOWWIDEBANDCODE" }, { TONECODE, "TONECODE" },
		{ CRATIOCODE, "CRATIOCODE" }, { MAXLEVELCODE, "MAXLEVELCODE" },
		{ BOGUSCODE, nullptr}
	};

static const char dfltchar='.';

static int pacode;
void digest_args(void);

		const char*
		printaimode(void)
		{
			return "NULL";
		}
			


#if 0
static bool gainset = false;
static float gainval = 0.0f;
static bool thresholdset = false;
static float thresholdval = 0.0f;
#endif
static char *previouscmd = nullptr;

namespace da
{
	extern float TONEOFFSET;
};

static int NOINLINE
process_args(int argc, char *const argv[])
{
	int c, opt_index, __attribute__((unused)) optlen;
	int error = 0;
	static struct option long_options[] = {
		{"output",	required_argument, &pacode, (int) OUTPUTCODE},
		{"input",	required_argument, &pacode, (int) INPUTCODE},
		{"help", optional_argument, &pacode, (int) HELPCODE},
		{"info",	optional_argument, &pacode, (int) INFOCODE},
		{"ing", required_argument, &pacode, (int) INGAINCODE},
		{"ingain", required_argument, &pacode, (int) INGAINCODE},
		{"thresh", required_argument, &pacode, (int) TONECODE},
		{"tone", required_argument, &pacode, (int) TONECODE},
		{"cmd",	required_argument, &pacode, (int) CMDCODE},
		{"outg",	required_argument, &pacode, (int) OUTGAINCODE},
		{"outgain",	required_argument, &pacode, (int) OUTGAINCODE},
		{"floatout",	no_argument, &pacode, (int) FLOATOUTCODE},
		{"mode",	required_argument, &pacode, (int) ONECODE},
		{"rawfilt",	optional_argument, &pacode, (int) RAWFILTCODE},
		{"affinity", optional_argument, &pacode, (int) AFFINITYCODE },
		{"xnr", no_argument, &pacode, (int) XNRCODE},
		{"cratio", required_argument, &pacode, (int) CRATIOCODE},
		{"allowwideband", optional_argument, &pacode, (int) ALLOWWIDEBANDCODE},
		{"maxgain", required_argument, &pacode, (int) MAXGAINCODE},
		{"maxlevel", required_argument, &pacode, (int) MAXLEVELCODE},
		{nullptr,       no_argument      , &pacode, (int) BOGUSCODE}
	};

	//
	// reset scan
	//
//	optind = 1;
	while (1)
	{
		opt_index = -1;
		c = ::getopt_long_only(argc, argv, "", long_options, &opt_index);
#if defined(CMDDEBUG)
		::fprintf(stderr,"getopt returns %d\n", c);
#endif
		if (c == -1)
		{
			break;
		}
		c = pacode;
		const static int optargtempsize=4096;
		enum optcode rtcode = (enum optcode) pacode;
		char optargtemp[optargtempsize];
		memset(optargtemp, 0,sizeof optargtemp);
#if defined(CMDDEBUG)
		::fprintf(stderr,"argument: optarg: %s, opt_index: %d, code: %d, name: \"%s\"\n", optarg, opt_index, rtcode, optlookup(rtcode));
#endif

		int optarglen = 0;
		if (long_options[opt_index].has_arg)
		{
#if defined(CMDDEBUG)
		::fprintf(stderr,"has arg\n");
#endif
			if (optarg)
				optarglen = ::strlen(optarg);
			if (optarglen)
			{
				if (optarglen > (optargtempsize - 1))
				{
					::fprintf(stderr,"argument \"%s\" is too big -- aborting\n", optarg);
					::exit(1);
				}

				// no worry about null terminate - -already handled with prezero above.
				if (optarg[0] == '\"' || optarg[0] == '\'')
				{
					::memcpy(optargtemp, optarg+1, (size_t) optarglen - 2);
				}
				else
				{
					::memcpy(optargtemp, optarg, (size_t) optarglen);
				}
#if defined(CMDDEBUG)
				::fprintf(stderr,"Optarg == \"%s\"\n", optargtemp);
#endif
			}
		}

#if defined(CMDDEBUG)
		::fprintf(stderr,"rtcode: %d\n", rtcode);
#endif

		switch (rtcode)
		{

		case MAXLEVELCODE:
			{
				if (optargtemp[0] == '\0')
				{
					ac.maxlevel = -1.0;
					break;
				}
				ac.maxlevel = (atof(optargtemp));
				break;
			}
		case MAXGAINCODE:
			{
				fprintf(stderr, "maxgain: %s\n", optargtemp);
				if (optargtemp[0] == '\0')
				{
					ac.maxgain = 9.0f;
					break;
				}
				ac.maxgain = atof(optargtemp);
				break;
			}

		case XNRCODE:
			{
				if (optargtemp[0] == '\0')
				{
					ac.xnr = 1;
					break;
				}
				ac.xnr = atoi(optargtemp);
				break;
			}
		case IDCODE:
			{
				if (ac.id != NULL)
					free(ac.id);
				ac.id = strndup(optargtemp, 8);
			}
			break;
		case CRATIOCODE:
			{
				ac.cratio = atof(optargtemp);
			}
			break;

		case AFFINITYCODE:
			{
				if (optargtemp[0] == '\0')
				{
					ac.affinity = -1;
				}
				else
				{
					ac.affinity = atoi(optargtemp);
				}
			}
			break;
				
		case FULLOUTCODE:
			{
				ac.fullout = !ac.fullout;
			}
			break;

		case HELPCODE:
			{
				fprintf(stderr,"This program is currently command line oriented, and normally uses pipe & stdin/stdio mechanisms\n");
				fprintf(stderr,"Example usage:\n");
				fprintf(stderr,"da-avx --info --thresh=-15.50 <infile.wav >outfile.wav\n");
				fprintf(stderr,"  the input/output wav files must be between 44.1k and 124k -- 192k is pushing it\n");
				fprintf(stderr,"  the file format can be 16bit signed, 24bit signed and 32bit floating point\n");
				fprintf(stderr,"   16bit signed is the CD bit layout, 24bit signed is the flac layout, and\n");
				fprintf(stderr,"   32bit floating point is the ideal format for minimal quality loss and zero headroom issues\n");
				fprintf(stderr,"---\n");
				fprintf(stderr,"hints for people not comfortable with command line:\n");
				fprintf(stderr," note the use of TWO dashes in front of switch names\n");
				fprintf(stderr," the threshold number is usually between -16.0 and -15.0 dB, have to detect by ear\n");
				fprintf(stderr,"  (if you have the DolbyA tone, it is usually approx 0.70 to 1.7dB greater than the\n");
				fprintf(stderr,"   input RMS level as detected by the program -- I still haven't fully determined the\n");
				fprintf(stderr,"   value, but that will give you a start for now.  I actually think that it is the 0.7dB\n");
				fprintf(stderr," there are also --input=filename.wav and --output=filename.wav switches for those\n");
				fprintf(stderr," not comfortable with the traditional input/output file redirection syntax.\n");
				fprintf(stderr,"You can also pipe the input/output to/from conversion programs or playout programs.\n");
				fprintf(stderr,"\n\n");
			}
			break;

		case CMDCODE:
			{
				fprintf(stderr,"--cmd is removed -- shutting down program to avoid errors!!!!\n");
				exit(1);
			}
			break;

		case HPFCODE:
			{
				if (optargtemp[0] == '\0')
				{
					ac.hpf = 1;
					break;
				}

				ac.hpf = atoi(optargtemp);
				break;
			}

		case LLOGCODE:
			ac.llog = atoi(optargtemp);
			break;

		case RAWFILTCODE:
				if (optargtemp[0] == '\0')
				{
					ac.rawfilt = true;
				}
				else
				{
					ac.rawfilt = strtol(optargtemp, NULL, 16);
				}
			break;

		case ALLOWWIDEBANDCODE:
			if (optargtemp[0] == '\0')
			{
				ac.allowwideband = true;
			}
			else
			{
				if (!strncmp(optargtemp, "no", 2))
					ac.allowwideband = false;
				else if (!strncmp(optargtemp, "off", 2))
					ac.allowwideband = false;
				else if (!strncmp(optargtemp, "0", 1))
					ac.allowwideband = false;
				else if (!strncmp(optargtemp, "yes", 2))
					ac.allowwideband = true;
				else if (!strncmp(optargtemp, "on", 2))
					ac.allowwideband = true;
				else if (!strncmp(optargtemp, "1", 1))
					ac.allowwideband= true;
				else {
					fprintf(stderr, "--allowwideband=<no,off,0,yes,on,1> bw > 21.5kHz if sample rate > 88k\n");
				}
			}
				
			break;

		case WIDEFILTCODE:
			if (optargtemp[0] == '\0')
			{
				ac.skip2nd = 1;
				ac.widefilt = 0.46f;
				break;
			}
			ac.widefilt = atof(optargtemp);
			if (ac.widefilt <= 0.0)
			{
				ac.skip2nd = 1;
				if (ac.widefilt < -1.5)
					ac.skip2nd =2;
				ac.widefilt = 0.0;
			}
			else
			{
				if (ac.widefilt > 0.46f)
				{
					ac.skip2nd = 1;
					ac.widefilt = 0.46f;
				}
				if (ac.widefilt < 0.10f)
				{
					ac.skip2nd = 1;
					ac.widefilt = 0.10f;
				}
			}
			break;

		case FLATTHRESHCODE:
			ac.flatthreshold = atof(optargtemp);
			break;

		case FILTERTESTCODE:
			{
				fprintf(stderr,"FILTER TEST MODE ENABLED --info displays filter levels now\n");
			}
			ac.filtertest = true;
			break;

		case FLEVELCODE:
			{
				ac.flevel = true;
			}
			break;

		case GCLIPCODE:
			{
				if (optargtemp[0] == '\0')
				{
					ac.gclip = 0.01;
				}
				else
				{
					ac.gclip = atof(optargtemp);
				}
			}
			break;

		case INGAINCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --ingain\n");
					break;
				}
				float gaintmp = atof(optargtemp);
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--ingain=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
					
				if (ac.ingainset)
				{
					fprintf(stderr,"--ingain switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.ingainval, gaintmp);
				}
				ac.ingainset = true;
				ac.ingainval = gaintmp;
			}
			break;

		case INGAINLCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --ingainl\n");
					break;
				}
				float gaintmp = atof(optargtemp);
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--ingainl=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
					
				if (ac.ingainlset)
				{
					fprintf(stderr,"--ingainl switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.ingainlval, gaintmp);
				}
				ac.ingainlset = true;
				ac.ingainlval = gaintmp;
			}
			break;

		case INGAINRCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --ingainr\n");
					break;
				}
				float gaintmp = atof(optargtemp);
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--ingainr=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
					
				if (ac.ingainrset)
				{
					fprintf(stderr,"--ingainr switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.ingainrval, gaintmp);
				}
				ac.ingainrset = true;
				ac.ingainrval = gaintmp;
			}
			break;


		case OUTGAINCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --outgain\n");
					break;
				}
				float gaintmp = atof(optargtemp);
#if 0
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--outgain=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
#endif
					
				if (ac.outgainset)
				{
					fprintf(stderr,"--outgain switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.outgainval, gaintmp);
				}
				ac.outgainset = true;
				ac.outgainval = gaintmp;
			}
			break;

		case OUTGAINLCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --outgainl\n");
					break;
				}

				float gaintmp = atof(optargtemp);
#if 0
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--outgainl=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
#endif
					
				if (ac.outgainlset)
				{
					fprintf(stderr,"--outgainl switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.outgainlval, gaintmp);
				}
				ac.outgainlset = true;
				ac.outgainlval = gaintmp;
			}
			break;

		case OUTGAINRCODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the --outgainr\n");
					break;
				}
				float gaintmp = atof(optargtemp);
#if 0
				if (previouscmd != nullptr)
				{
					fprintf(stderr,"--outgainr=%s switch used & ignored, because --cmd=\"%s\" was already used\n",
						optargtemp, previouscmd);
					break;
				}
#endif
					
				if (ac.outgainrset)
				{
					fprintf(stderr,"--ingainr switch value previously set to \"%fdB\", changing to \"%fdB\"\n",
						ac.outgainrval, gaintmp);
				}
				ac.outgainrset = true;
				ac.outgainrval = gaintmp;
			}
			break;

		case TONECODE:
			{
				if (optargtemp[0] == '\0')
				{
					fprintf(stderr,"please specify a value for the threshold\n");
					break;
				}
				float tonetmp = atof(optargtemp);
				float threshtmp = tonetmp + da::TONEOFFSET;
				ac.thresholdset = true;
				ac.thresholdval = threshtmp;
				if (ac.info > 100)
				{
					fprintf(stderr,"Tone setting: %.2f, threshold equiv: %.2f\n", tonetmp, threshtmp);
				}
			}
			break;
	
		//
		// input file
		//
		case INPUTCODE:
			
			if (optargtemp[0] == '\0')
				break;
#if defined(CMDDEBUG)
		::fprintf(stderr,"optargtemp: %s\n", optargtemp);
#endif
			if (ac.infile != nullptr)
				::free(ac.infile);
			ac.infile = ::strdup(optargtemp);
#if defined(CMDDEBUG)
		::fprintf(stderr,"new infile: %s\n", ac.infile);
#endif
			break;

		case ONECODE:
			ac.one = atoi(optargtemp);
			break;

		//
		// cl_port
		//
		case OUTPUTCODE:
			if (optargtemp[0] == '\0')
				break;
#if defined(CMDDEBUG)
		::fprintf(stderr,"optargtemp: %s\n", optargtemp);
#endif
			if (ac.outfile != nullptr)
				::free(ac.outfile);
			ac.outfile = ::strdup(optargtemp);
#if defined(CMDDEBUG)
		::fprintf(stderr,"new outfile: %s\n", ac.outfile);
#endif
			break;

		//
		// loud debug
		//
		case INFOCODE:
			if (optargtemp[0] == '\0')
			{
				ac.info = 1;
				break;
			}
			ac.info = atoi(optargtemp);
			//fprintf(stderr,"setting info to be: %d\n", ac.info);
			break;

		case BOGUSCODE:
			break;

		case FLOATOUTCODE:
			ac.rareqfloatout = 1;
			if (ac.info > 0)
				::fprintf(stderr,"32 bit floating point output: enabled\n");
			break;
			
		}
	}

	digest_args();

	if (true || (ac.info > 4))
	{
		fprintf(stderr,"Audio \"Simple compressor\" %s -- Author John S. Dyson -- Copyright 2017,2018,2019\nLicense terms are in audioproc.cpp & accompanying documentation\n", programVersion);
		if (ac.info) fprintf(stderr,"Required CPU type: %s\n", VNAME);
	}

	if (previouscmd == nullptr)
	{
#if 0
		if (!ac.gainset)
		{
			ac.gainval = 0.0f;
		}
#endif
		char previouscmd[2048];
		int nc = snprintf(previouscmd, sizeof previouscmd, "(l)");
		if (nc != -1)
		{
#if 0
			if (ac.info)
			{
				fprintf(stderr,"using threshold = %f, gain = %f\n", ac.thresholdval, ac.gainval);
			}
#endif
		}
	}

	return error;
}

extern int setupinfile(void);

static char inbuf[16384];
static char outbuf[16384];

static void displayconfig(FILE *);

static void
wakesig(int __attribute__((unused)) sig)
{
	return;
}

extern void runexp(void);

pid_t mainpid = 0;
static void acinit();


void
digest_args(void)
{
	if (!ac.thresholdlset && ac.thresholdset)
	{
		ac.thresholdlval = ac.thresholdval;
		ac.thresholdlset = true;
	}
	if (!ac.thresholdrset && ac.thresholdset)
	{
		ac.thresholdrval = ac.thresholdval;
		ac.thresholdrset = true;
	}

	if (!ac.ingainlset && ac.ingainset)
	{
		ac.ingainlval = ac.ingainval;
		ac.ingainlset = true;
	}
	if (!ac.ingainrset && ac.ingainset)
	{
		ac.ingainrval = ac.ingainval;
		ac.ingainrset = true;
	}

	if (!ac.outgainlset && ac.outgainset)
	{
		ac.outgainlval = ac.outgainval;
		ac.outgainlset = true;
	}
	if (!ac.outgainrset && ac.outgainset)
	{
		ac.outgainrval = ac.outgainval;
		ac.outgainrset = true;
	}

	ac.LINthresholdlval = dB2lin(ac.thresholdlval);
	ac.LINthresholdrval = dB2lin(ac.thresholdrval);
	ac.LINingainlval = dB2lin(ac.ingainlval);
	ac.LINingainrval = dB2lin(ac.ingainrval);
	ac.LINoutgainlval = dB2lin(ac.outgainlval);
	ac.LINoutgainrval = dB2lin(ac.outgainrval);
	

}

int
main(int argc, char * const argv[])
{

#if !defined(WINDOWS64)
	fcntl(0, F_SETPIPE_SZ, 524288 * 2);

	fcntl(1, F_SETPIPE_SZ, 524288 * 2);
#endif

	mainpid = getpid();
	acinit();

	process_args(argc, argv);

//	if (!ac.thresholdset && !(ac.thresholdlset && ac.thresholdrset))
//	{
//		::fprintf(stderr,"***FATAL*** Setting the threshold is not optional\n");
//		::exit(1);
//	}

	if (ac.info == 0)
	{
//		::fprintf(stderr,"Some help:  NEEDS TO BE COMPLETED AGAIN\n");
	}

	if ((ac.infile == nullptr) || !::strncmp(ac.infile, "-", 2))
	{
		ac.infp = fdopen(dup(0), "rb");
		if (ac.infile == nullptr)
			ac.infile = strdup("Using stdin");
	}
	else
	{
		ac.infp = fopen(ac.infile, "rb");
	}

	if (ac.infp == nullptr)
	{
#if !defined(NOSTRERROR)
		char sbuf[MAXTMPSTR + 1];
		const char *rtstr = lstrerror_r(errno, sbuf, sizeof sbuf);
		::fprintf(stderr,"open error \"%s\", code: %d, for input file: \"%s\"\n", rtstr, errno, ac.infile);
#else
		const char *rtstr = strerror(errno);
		::fprintf(stderr,"open error \"%s\", code: %d, for input file: \"%s\"\n", rtstr, errno, ac.infile);
#endif
		::exit(1);
	}

#if defined(CMDDEBUG)
	::fprintf(stderr, "trying to open output -- \"%s\"\n", ac.outfile);
#endif

	if (ac.outfile == nullptr)
	{
		ac.outfp = fdopen(dup(1), "wb");
#if defined(CMDDEBUG)
	::fprintf(stderr, "duped output fd: 1\n");
#endif
	}
	else
	{
		ac.outfp = fopen(ac.outfile, "w+b");
	}

#if defined(CMDDEBUG)
	::fprintf(stderr, "opened output -- \"%s\", fd: %d\n", ac.outfile, fileno(ac.outfp));
#endif
	if (ac.outfp == nullptr)
	{
#if !defined(NOSTRERROR)
		char sbuf[MAXTMPSTR + 1];
		const char *rtstr = lstrerror_r(errno, sbuf, sizeof sbuf);
		::fprintf(stderr,"open error \"%s\", code: %d, for create/output file: \"%s\"\n", rtstr, errno, ac.outfile);
#else
		const char *rtstr = strerror(errno);
		::fprintf(stderr,"open error \"%s\", code: %d, for create/output file: \"%s\"\n", rtstr, errno, ac.outfile);
#endif
		::exit(1);
	}

	setvbuf(ac.infp, inbuf, _IOFBF, sizeof inbuf);
	setvbuf(ac.outfp, outbuf, _IOFBF, sizeof outbuf);

#if defined(CMDDEBUG)
	::fprintf(stderr, "file %s open with fd %d, file %s open with fd %d\n",
		ac.infile, fileno(ac.infp), ac.outfile, fileno(ac.outfp));
#endif

	if (setupinfile())
	{
		fprintf(stderr,"Failure to start -- input file is incompatible\n");
		exit(1);
	}

	const char *aitype = printaimode();
	if (ac.info >= 10)
		fprintf(stderr,"Antialiasing mode: %s\n", aitype);
	if (ac.widefilt != 0.0f)
		fprintf(stderr,"Bandwidth: %lf\n", double(aspec.inSRATE * ac.widefilt));

	initaffinity();
	initcqs();
	setsignal();
	forcettyupdate();
	accuminfo::initaccumulatelevel();
//	runexp();
	threads_start();

	if (ac.info > 0x100)
		fprintf(stderr,"waiting for thread rundown\n");
	wait_all_threads();
	::exit(0);
}

static void NOINLINE
acinit(void)
{
	setfpmode();
	acINIT.infile = nullptr;
	acINIT.outfile = nullptr;
	acINIT.infp = nullptr;
	acINIT.outfp = nullptr;
	acINIT.rafloatout = -1;
	acINIT.rareqfloatout = 0;
	acINIT.info = 0;
	acINIT.outgainset = false;
	acINIT.outgainval = 0.0f;
	acINIT.outgainlset = false;
	acINIT.outgainlval = 0.0f;
	acINIT.outgainrset = false;
	acINIT.outgainrval = 0.0f;
	acINIT.thresholdset = false;
	acINIT.thresholdval = -0.80f;
	acINIT.thresholdlset = false;
	acINIT.thresholdlval = -0.80f;
	acINIT.thresholdrset = false;
	acINIT.thresholdrval = -0.80f;
	acINIT.ingainset = false;
	acINIT.ingainval = 0.0f;
	acINIT.ingainlset = false;
	acINIT.ingainlval = 0.0f;
	acINIT.ingainrset = false;
	acINIT.ingainrval = 0.0f;

	acINIT.LINthresholdlval = 0.0f;
	acINIT.LINthresholdrval = 0.0f;
	acINIT.LINingainlval = 0.0f;
	acINIT.LINingainrval = 0.0f;
	acINIT.LINoutgainlval = 0.0f;
	acINIT.LINoutgainrval = 0.0f;
	acINIT.one = 0;
	acINIT.bandenable = 0xff;
	acINIT.filtertest = false;
	acINIT.flevel = false;
	acINIT.flatthreshold = 7.5f;
	acINIT.gclip = -5.0;
	acINIT.rawfilt = 0x00;
	acINIT.widefilt = 0.0f;
	acINIT.skip2nd = 1;
	acINIT.llog = 0;
	acINIT.hpf = 0;
	acINIT.decaymode = 0;
	acINIT.id = nullptr;
	acINIT.nchannels = 2;
	acINIT.bext = nullptr;
	acINIT.bextdesc = nullptr;
	acINIT.bextorig = nullptr;
	acINIT.bextorigref = nullptr;
	acINIT.fullout = true;
	acINIT.affinity = 0;
	acINIT.xnr = 0;
	acINIT.cratio = 0.50f;
	acINIT.maxgain = 9.0f;
	acINIT.maxlevel = -1.5f;
	acINIT.allowwideband = true;

	ac = acINIT;
}


//
// vi: set ts=4:
//
