
//
//
// Copyright(C) 2017,2018,2019 John S Dyson
// All rights reserved.
//
// Refer to audioproc.cpp for license.
// 

#include <sys/types.h>
#include "vecmath.h"
#include "vtype.h"



#if !defined(__ITQA_QDEFS__)
#define __ITQA_QDEFS__ 1

#include "THQCPPbasena.h"

static constexpr int NBANDS = 4;
static constexpr int MAXNFPASSES = 8;

namespace da
{
	struct fpdata
	{
		int np, curnfpasses;
		bool wideband;
		bool lastpass;
		bool xfilt, split1;
		bool supermode, xsupermode;
		bool hfilt, hspeed;
		bool disableall;
		int dbamode;

		float flipfactorstart;
		float flipfactor;

		Asample sigin;
		vtype8 pgain;
		vtype8 hpgain;
		vtype8 dgain;
		vtype8 pgc;
		float pgainl1slow;
		float pgainr1slow;
		float hpgainl1slow;
		float hpgainr1slow;
		Asample sigoutbuf[NBANDS];
		Asample mflow, mfhigh;
		Asample hf0low, hf0high;
		Asample hf1low, hf1high;
		Asample hf1AxN45, hf1aboveN45;

		float totaldelay;
		Asample sigout;
	};

	using fpdatap = fpdata*;
};



static const int NAUDIOCHUNKSXTRA=32;
//static const int NOBUFPAYLOAD=24;


#if 1
struct THQALIGNED audiopayload
{
	int samplecount;
	da::fpdatap samples[AUDIOCHUNKSIZE];
	audiopayload(void) : samplecount(), samples()
	{
		for (int i = 0; i < AUDIOCHUNKSIZE; i++)
		{
			samples[i] = new da::fpdata;
		}
		
	}

	~audiopayload()
	{
		for (int i = 0; i < AUDIOCHUNKSIZE; i++)
		{
			delete samples[i];
			samples[i] = nullptr;
		}
	}
};
#else
struct THQALIGNED audiopayload
{
	int samplecount;
	da::fpdata *samples;
	audiopayload(void) : samplecount(), samples()
	{
		samples = new da::fpdata[AUDIOCHUNKSIZE];
/*
		for (int i = 0; i < AUDIOCHUNKSIZE; i++)
		{
			samples[i] = new da::fpdata;
		}
*/
		
	}

	~audiopayload()
	{
/*
		for (int i = 0; i < AUDIOCHUNKSIZE; i++)
		{
			delete samples[i];
			samples[i] = nullptr;
		}
*/
		if (samples != nullptr)
			delete [] samples;
	}
};


#endif

//
// 'audiochunk' is the low-brow name of the data chunks sent between each thread.
//

typedef THQpayload <audiopayload> audiochunk;
typedef audiochunk *paudiochunk;

//
// We use the 'queue' for most message passing, but in cases where locality of reference
// is needed and order of access is not needed, then the stack might be better performing.
//
// I could probably organize the stack and queue code to have more in common, but isn't critical
// at this point.  If I did organize for more code in common, it might be slightly easier to
// maintain and might have slightly better locality of reference for code cache.
//
typedef THQstack<paudiochunk> achunkstack;
typedef THQqueue<paudiochunk> achunkqueue;

//
// The audio chunks are preallocated -- dynamic allocation seems to be unneeded because of
// large system memory sizes nowadays and cpu time being a little more important.
//
extern audiochunk preallocaudiochunk[NAUDIOCHUNKSXTRA];

//extern achunkqueue freeaudiochunks;

extern void initfreeaudiochunks(void);
extern paudiochunk findfreeaudiochunk(void);


typedef THQ<achunkstack, paudiochunk, NAUDIOCHUNKSXTRA, 1, true> cq0;

typedef THQ<achunkqueue, paudiochunk, 3, 1> cqtype0;

extern cq0 demptyqueue;	// anywhere between 0 and NAUDIOCHUNK entries inclusive
extern cqtype0 dinput;
extern cqtype0 dinputpost;
extern cqtype0 dfilter[MAXNFPASSES];
extern cqtype0 dendproc;
extern cqtype0 doutput;

extern int64_t cq0indexlog[NAUDIOCHUNKSXTRA];

void initcqs(void);
void logcq0(int index);
void dumplogcq0(void);
void doqueueinits(void);

extern void dumprtqueuestats();

#endif

//
// vim: set ts=4:
//
