
//
//
// Copyright(C) 2017,2018,2019 John S Dyson
// All rights reserved.
//
// Refer to audioproc.cpp for license
// 

#include <sys/types.h>

#if !defined(__ITQA_DEFS__)
#define __ITQA_DEFS__ 1

#include "itqA.h"

class physaudiosample;
struct physaudiosamplefloat;

#define TDINLINE __attribute__((always_inline))

static short UNUSED
inltrunc(float fpin)
{
	return short(rint(fpin));
}

extern int limitcount;
extern int hlimitcount;

static inline void addlimitcount(int incr)
{
	__atomic_add_fetch(&limitcount, incr, __ATOMIC_SEQ_CST);
}

static inline int getlimitcount(void)
{
	return __atomic_load_n(&limitcount, __ATOMIC_SEQ_CST);
}

static inline int getresetlimitcount(void)
{
	return __atomic_exchange_n(&limitcount, 0, __ATOMIC_SEQ_CST);
}

static inline void addhlimitcount(int incr)
{
	__atomic_add_fetch(&hlimitcount, incr, __ATOMIC_SEQ_CST);
}

static inline int gethlimitcount(void)
{
	return __atomic_load_n(&hlimitcount, __ATOMIC_SEQ_CST);
}

static inline int getresethlimitcount(void)
{
	return __atomic_exchange_n(&hlimitcount, 0, __ATOMIC_SEQ_CST);
}


class Asample
{

public:
	sv_phys_t sv;
	static int itemperline;

#define AsMultFlt(srcrt, val) srcrt.__multflt(val)
#define AsAddAs(srcrt, val) srcrt.__AddAs(val)

	Asample(void) : sv()
	{
		sv.sva[0] = 0;
		sv.sva[1] = 0;
	}

#if 0
	Asample(double v0, double v1) : sv()
	{
		sv.sva[0] = float(v0);
		sv.sva[1] = float(v1);
	}
#endif

	Asample(float v0, float v1) : sv()
	{
		sv.sva[0] = v0;
		sv.sva[1] = v1;
	}

#if 0
	Asample( int v0, int v1) : sv()
	{
		sv.sva[0] = v0;
		sv.sva[1] = v1;
	}
#endif

	Asample(const physaudiosample& src);
	Asample(const physaudiosamplefloat& src);
	Asample& operator=(const physaudiosample& src);
	Asample& operator=(const physaudiosamplefloat& src);

	Asample& ALWAYS_INLINE operator*=(const float g)
	{
		sv.sva[0] *= g;
		sv.sva[1] *= g;
		return *this;
	}

	Asample& ALWAYS_INLINE multpair(float v0, float v1)
	{
		sv.sva[0] *= v0;
		sv.sva[1] *= v1;
		return *this;
	}

	Asample& ALWAYS_INLINE __multflt(const float g)
	{
		sv.sva[0] *= g;
		sv.sva[1] *= g;
		return *this;
	}

	Asample& ALWAYS_INLINE operator*=(const Asample& src)
	{
		sv.sva[0] *= src.sv.sva[0];
		sv.sva[1] *= src.sv.sva[1];
		return *this;
	}

	Asample& ALWAYS_INLINE operator/=(const float g)
	{
		float divisor = 1.0f / g;
		sv.sva[0] *= divisor;
		sv.sva[1] *= divisor;
		return *this;
	}

	Asample& ALWAYS_INLINE operator=(const Asample& src) 
	{
		if (&src == this)
			return *this;

	    sv.sva[0] = src.sv.sva[0];
	    sv.sva[1] = src.sv.sva[1];
		return *this;
	}

	Asample& ALWAYS_INLINE operator+=(const Asample& src)
	{
		sv.sva[0] += src.sv.sva[0];
		sv.sva[1] += src.sv.sva[1];
		return *this;
	}


	Asample& ALWAYS_INLINE __AddAs(Asample src)
	{
		sv.sva[0] += src.sv.sva[0];
		sv.sva[1] += src.sv.sva[1];
		return *this;
	}

	Asample& ALWAYS_INLINE operator-=(const Asample& src)
	{
		sv.sva[0] -= src.sv.sva[0];
		sv.sva[1] -= src.sv.sva[1];
		return *this;
	}

	void asprintf(FILE* outfile = stderr);
	char *asprint(void) const;

	float sumsq(void) const TDINLINE
	{
		float total0 = sv.sva[0] * sv.sva[0];
		float total1 = sv.sva[1] * sv.sva[1];
		float total = (total0 + total1);
		return (total <= GBLminfloat) ? GBLminfloat : total;
	}

	float Xmsumsq(void) const // NOINLINE
	{
		float total0 = sv.sva[0] * sv.sva[0];
		float total1 = sv.sva[1] * sv.sva[1];
		float total = (total0 > total1) ? total0 : total1;
			
		return (total <= GBLminfloat) ? GBLminfloat : total;
	}

	float msumsq(void) const // TDINLINE
	{
		return Xmsumsq();
	}

	float sumabs(void) const TDINLINE
	{
		
		float total0 = sv.sva[0];
		float total1 = sv.sva[1];
		total0 = (total0 < 0.0f) ? (-1.0f * total0) : total0;
		total1 = (total1 < 0.0f) ? (-1.0f * total1) : total1;
		float total = 1.4f * ((total0 > total1) ? total0 : total0);
		return (total <= GBLminfloat ? GBLminfloat : total);
	}
	
	float item(int index) const TDINLINE
	{
		return sv.sva[index];
	}

	float item0(void) const TDINLINE
	{
		return sv.sva[0];
	}

	float item1(void) const TDINLINE
	{
		return sv.sva[1];
	}

	float& item0(void) TDINLINE
	{
		return sv.sva[0];
	}

	float& item1(void) TDINLINE
	{
		return sv.sva[1];
	}

	void setitem(float v, int index) TDINLINE
	{
		sv.sva[index] = v;
	}

	void setitem0(float v) TDINLINE
	{
		sv.sva[0] = v;
	}

	void setitem1(float v) TDINLINE
	{
		sv.sva[1] = v;
	}

	Asample MminAs(float v) const TDINLINE
	{
		return Asample(minf(sv.sva[0], v), minf(sv.sva[1], v));
	}

	Asample MmaxAs(float v) const TDINLINE
	{
		return Asample(maxf(sv.sva[0], v), maxf(sv.sva[1], v));
	}

	const v2sf&
	vec() const
	{
		return sv.svv;
	}

	v2sf&
	vec()
	{
		return sv.svv;
	}

};

static UNUSED float Tmax(float a, float b)
{
	return (a > b) ? a : b;
}

static UNUSED Asample Asmax(const Asample& src0, const Asample& src1)
{
	return Asample( Tmax(src0.item0(), src1.item0()), Tmax(src0.item1(), src1.item1()));
}

static UNUSED Asample operator*(const Asample& src0, const Asample src1)
{
	return Asample(src0.item0() * src1.item0(), src0.item1() * src1.item1());
}

static UNUSED Asample operator*(const Asample& src0, float src1)
{
	return Asample(src0.item0() * src1, src0.item1() * src1);
}

static UNUSED Asample operator*(float& src0, const Asample& src1)
{
	return Asample(src0 * src1.item0(), src0 * src1.item1());
}

static UNUSED Asample operator+(const Asample& src0, const Asample& src1)
{
	return Asample(src0.item0() + src1.item0(), src0.item1() + src1.item1());
}

static UNUSED Asample operator-(const Asample& src0, const Asample& src1)
{
	return Asample(src0.item0() - src1.item0(), src0.item1() - src1.item1());
}

static UNUSED Asample operator/(const Asample& src0, const Asample& src1)
{
	return Asample(src0.item0() / src1.item0(), src0.item1() / src1.item1());
}

static UNUSED Asample operator/(const Asample& src0, float src1)
{
	return Asample(src0.item0() / src1, src0.item1() / src1);
}

static UNUSED Asample operator/(float src0, const Asample& src1)
{
	return Asample(src0 / src1.item0(), src0 / src1.item1());
}

// class Asample;

typedef Asample *pAsample;


void postfilt(Asample* asr);


namespace lm
{
        static inline Asample Mmin(Asample v0, float v1)
        {
                return v0.MminAs(v1);
        }

        static inline Asample Mmax(Asample v0, float v1)
        {
                return v0.MmaxAs(v1);
        }
};

class physaudiosample
{
public:
	int16_t s0, s1;
	void zero(void);
		
	physaudiosample(void) : s0(int16_t(0)), s1(int16_t(0)) {}
	physaudiosample(const physaudiosample& src) : s0(src.s0), s1(src.s1) {}
	physaudiosample(const Asample& src) : s0(cvtMAXto16(src.item0())), s1(cvtMAXto16(src.item1())) {}

	void dump(void);

	physaudiosample&
	operator=(const physaudiosample& src)
	{
		s0 = src.s0;
		s1 = src.s1;
		return *this;
	}

};

typedef struct threebytes
{
	unsigned char b[3];
public:
	void zero(void);
	threebytes(void);
	threebytes(const threebytes& init);
	threebytes& operator=(const threebytes& src);
	threebytes(float av);
	float tofloat(void) const;
} threebytes_t;

class physaudiosample24
{
public:
	threebytes_t s0, s1;
	void zero(void);
	physaudiosample24(void);
	physaudiosample24(const Asample& src);
	physaudiosample24& operator=(const physaudiosample24& src);
};

struct physaudiosamplefloat
{
	float stereo[2];

	void zero(void)
	{
		stereo[0] = 0.0f;
		stereo[1] = 0.0f;
	}

public:
	physaudiosamplefloat(void) : stereo()
	{
		zero();
	}

	physaudiosamplefloat(const Asample& src) : stereo()
	{
		stereo[0] = src.item0();
		stereo[1] = src.item1();
	}

	physaudiosamplefloat& operator=(const physaudiosamplefloat& src)
	{
		if (&src == this)
			return *this;

		stereo[0] = src.stereo[0];
		stereo[1] = src.stereo[1];
		return *this;
	}
};


#include "itq-cmpdefs.h"

//typedef engine_hardlimit<float> hardlimit_float;
//typedef engine_hardlimit<Asample, float> hardlimit_Asample;
//typedef engine_slimitT<float> engine_slimitf;


#endif
//
// vim: set ts=4:
//

