
//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 



#if !defined(__THQbase_H__)
#define __THQbase_H__ 1

#include <sys/types.h>

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <errno.h>
//#include <math.h>
#include <string.h>
#include <signal.h>

#include <vector>
#include <list>
#include <map>
#include <deque>
#include <queue>
#include <stack>

using std::queue;
using std::vector;
using std::list;
using std::map;
using std::stack;
using std::deque;

//#include "autop.h"

#define THQINLINE __attribute__((always_inline))

#define THQNOINLINE __attribute__((noinline))

#define THQFATTR0 
#define THQFATTR1 
#define THQFATTR2

#include "THQdefs.h"

//
// Inter thread buffering, uses pthread mutex/cond variable based synchronization.
//

enum class THQpayloadstate {
	STNOINFO = 0,
	STEOF = 1
};

//
// Data sent between threads "class THQpayload"
// Data communications mgmt is handled by "class Tachunkqueue" and
//   "class THQstack".  Tachunkclass provides more immediate
//   reuse of data chunks -- POSSIBLY providing better locality of
//   reference.  I found that if the system is very busy with the
//   program, then the reuse because less likely.  It seems to be
//   a good idea to try to implement the more recent reuse if possible,
//   it just might help and doesn't add anything to the complexity.
//

//
// This is the raw data that is sent between threads
//
#if 0
template <typename ASTYPE>
class THQpayload
{
	int index;
	THQpayloadstate state;
	ASTYPE* payloadbuf;

public:
	THQpayload() THQFATTR0 :
		index(0), state(THQpayloadstate::STNOINFO), payloadbuf(new ASTYPE{}) {}

	THQpayload(const THQpayload<ASTYPE>& init) THQFATTR0 :
		index(init.index), state(THQpayloadstate::STNOINFO),
		payloadbuf(new ASTYPE{})
	{
		*payloadbuf = *init.payloadbuf;
	}

	~THQpayload() THQFATTR0
	{
		delete [] payloadbuf;
		payloadbuf = nullptr;
	}

	THQpayload& THQFATTR0
	operator=(const THQpayload<ASTYPE> & src)
	{
		if (&src == this)
			return *this;

		index = src.index;
		state = src.state;
		if (payloadbuf != nullptr)
			delete payloadbuf;
		payloadbuf = new ASTYPE{};
		*payloadbuf = *src.payloadbuf;
		return *this;
	}

	//
	// allocate/unallocate is not fully implemented, so don't worry if you see some stuff
	// that doesn't seem to be correct.
	//
	void
	unallocate(void)
	{
	}

	void
	allocate(void)
	{
	}

	bool
	isallocated(void)
	{
	}

	//
	// Each raw data buffer has an 'index' or 'id' for logging/statistics purposes.
	//
	int 
	getindex(void) const
	{
		return index;
	}

	void 
	setindex(int newindex)
	{
		index = newindex;
	}

	//
	// This supports access to the raw data in the buffer.
	//
	const ASTYPE&
	payload(void) const
	{
		return *payloadbuf;
	}

	ASTYPE&
	payload(void)
	{
		return *payloadbuf;
	}

	bool
	iseof(void) const
	{
		return state == THQpayloadstate::STEOF;
	}

	void
	seteof(void)
	{
		state = THQpayloadstate::STEOF;
	}

};
#else

template <typename ASTYPE>
class THQpayload
{
	int index;
	THQpayloadstate state;
	THQALIGNED ASTYPE payloadbuf;

public:
	THQpayload() THQFATTR0 :
		index(0), state(THQpayloadstate::STNOINFO), payloadbuf() {}

	THQpayload(const THQpayload<ASTYPE>& init) THQFATTR0 :
		index(init.index), state(THQpayloadstate::STNOINFO),
		payloadbuf()
	{
		payloadbuf = init.payloadbuf;
	}

	~THQpayload() THQFATTR0
	{
		payloadbuf = nullptr;
	}

	THQpayload& THQFATTR0
	operator=(const THQpayload<ASTYPE> & src)
	{
		if (&src == this)
			return *this;

		index = src.index;
		state = src.state;
		payloadbuf = src.payloadbuf;
		return *this;
	}

	//
	// allocate/unallocate is not fully implemented, so don't worry if you see some stuff
	// that doesn't seem to be correct.
	//
	void
	unallocate(void)
	{
	}

	void
	allocate(void)
	{
	}

	bool
	isallocated(void)
	{
		return false;
	}

	//
	// Each raw data buffer has an 'index' or 'id' for logging/statistics purposes.
	//
	int 
	getindex(void) const
	{
		return index;
	}

	void 
	setindex(int newindex)
	{
		index = newindex;
	}

	//
	// This supports access to the raw data in the buffer.
	//
	const ASTYPE&
	payload(void) const
	{
		return payloadbuf;
	}

	ASTYPE&
	payload(void)
	{
		return payloadbuf;
	}

	bool
	iseof(void) const
	{
		return state == THQpayloadstate::STEOF;
	}

	void
	seteof(void)
	{
		state = THQpayloadstate::STEOF;
	}

};

#endif

template <class ELEMENT>
	class THQstack {

	protected:
		std::deque<ELEMENT> c;
	public:

		THQstack(void) THQFATTR0 : c() {}

		THQstack(const THQstack& init) THQFATTR0 : c(init.c) {}


		typename std::deque<ELEMENT>::size_type THQFATTR0 size() const
		{
			return c.size();
		}

		bool THQFATTR0 empty(void) const
		{
			return c.empty();
		}

		void THQFATTR0 additem(const ELEMENT& e)
		{
			c.push_back(e);
		}

		ELEMENT THQFATTR0 getnextitem(void)
		{
			//
			// Somehow check for error -- lets avoid exceptions
			//
			ELEMENT e(c.back());
			c.pop_back();
			return e;
		}

		ELEMENT& THQFATTR0 checknextitem(void)
		{
			return c.back();
		}
	};

template <class ELEMENT>
	class THQqueue {
	
	protected:
		std::deque<ELEMENT> c;

	public:
		THQqueue(void) THQFATTR0 : c() {}

		THQqueue(const THQqueue& init) THQFATTR0 : c(init.c) {}

		typename std::deque<ELEMENT>::size_type THQFATTR0 size() const
		{
			return c.size();
		}

		bool THQFATTR0 empty() const
		{
			return c.empty();
		}

		void THQFATTR0 additem(const ELEMENT& e)
		{
			c.push_back(e);
		}

		ELEMENT THQFATTR0 getnextitem(void)
		{
			ELEMENT e(c.front());
			c.pop_front();
			return e;
		}

		ELEMENT& THQFATTR0 checknextitem(void)
		{
			return c.front();
		}
	};


template <typename QTYPE, typename DTYPE, int MAXQSIZE, int QWAKEUPSIZE, bool NOBLOCKSTORE=false>
THQALIGNED class THQ
{
	int curqcount;
	THQALIGNED QTYPE aq;
	THQALIGNED pthread_mutex_t qlock;
	THQALIGNED pthread_cond_t waitdataavail;
	THQALIGNED pthread_cond_t waitqueueavail;
	char *qname;
	int64_t waitdataavailcount;
	int64_t signaldataavailcount;
	int64_t waitqueueavailcount;
	int64_t signalqueueavailcount;
	int64_t signalqueueavailoptimize;
	bool waitingondataavail;
	bool waitingonqueuespaceavail;

#if defined(DEFAULTWAKEUP)
	void THQFATTR2
	local_pthread_cond_timewait50msec(pthread_cond_t* cond, pthread_mutex_t* mtx)
	{
		timespec waittimeout = THQsupport::gettime_with_offset_ms(50);
		pthread_cond_timedwait(cond, mtx, &waittimeout);
	}
#endif

#if defined(PRINTFTRACE)
	void THQFATTR2
	waitqmsgs(int msgid)
	{
		switch (msgid)
		{
		case 0:
			::fprintf(stderr,"THQwqdataread: queue \"%s\" is empty, waiting on data available\n", qname);
			break;
		case 1:
			::fprintf(stderr,"THQwqdataread: queue \"%s\" error -- empty queue with curqcount == 0\n", qname);
			break;
		case 2:
			::fprintf(stderr,"THQwqdataread: queue \"%s\" broadcast to cond variable waitqavail\n", qname);
			break;
		case 3:
			::fprintf(stderr,"THQwqdataread: queue \"%s\" broadcast to cond variable waitqavail\n", qname);
			break;
		default:
			break;
		}
	}

	void THQFATTR2
	waitqrundownmsg(int msgid, int msgarg = 0)
	{
		switch (msgid)
		{
		case 0:
			::fprintf(stderr,"THQwqrundown: queue %s rundown attempt %d\n", qname, msgarg);
			break;
		case 1:
			::fprintf(stderr,"THQwqrundown: queue %s rundown complete\n", qname);
			break;
		default:
			break;
		}
	}

	void THQFATTR2
	waitqspacemsg(int msgid, bool foreof = false)
	{
		switch(msgid)
		{
		case 0:
			::fprintf(stderr,"THQwqspace: queue %s waiting for some space%s -- curqcount == %d\n",
				qname, foreof ? " for EOF processing" : "", curqcount);
			break;
		case 1:
			::fprintf(stderr,"THQwqspace: queue %s space found%s -- curqcount == %d\n",
				qname, foreof ? " for EOF processing" : "", curqcount);
			break;
		default:
			break;
		}
	}

	void THQFATTR2
	settingnewqname(int msgid, const char *qnargorig, const char *qnargnew)
	{
		switch(msgid)
		{
		case 0:
			::fprintf(stderr,"setnewqname: original name: %s, newname: %s\n",
				(qnargorig != nullptr) ? qnargorig : "<unknown or undef>",
				(qnargnew != nullptr) ? qnargnew : "<set to null name>");
			break;
		default:
			break;
		}
	}

#endif

	THQ(const THQ& init);
	THQ& operator=(const THQ& src);
			
public:
#if defined(__clang__)
	THQ(const char *qn = nullptr) : curqcount(0), aq(),
		qlock(), waitdataavail(),
		waitqueueavail(), qname((qn == nullptr) ? nullptr : (::strdup(qn))),
		waitdataavailcount(0), signaldataavailcount(0), waitqueueavailcount(0), signalqueueavailcount(0),
		signalqueueavailoptimize(0),
		waitingondataavail(false), waitingonqueuespaceavail(false)
	{
		pthread_mutex_init(&qlock, nullptr);
		pthread_cond_init(&waitdataavail, nullptr);
		pthread_cond_init(&waitqueueavail, nullptr);
	}
#else
	THQ(const char *qn = nullptr) : curqcount(0), aq(),
		qlock(PTHREAD_MUTEX_INITIALIZER), waitdataavail(PTHREAD_COND_INITIALIZER),
		waitqueueavail(PTHREAD_COND_INITIALIZER), qname((qn == nullptr) ? nullptr : (::strdup(qn))),
		waitdataavailcount(0), signaldataavailcount(0), waitqueueavailcount(0), signalqueueavailcount(0),
		signalqueueavailoptimize(0),
		waitingondataavail(false), waitingonqueuespaceavail(false) {}
#endif

	~THQ() THQFATTR0
	{
		if (qname != nullptr)
		{
			::free(qname);
			qname = nullptr;
		}
	}

	void THQFATTR1
	setqname(const char* newqname)
	{
#if defined(PRINTFTRACE) && (PRINTFTRACE > 4)
		settingnewqname(0, qname, newqname);
#endif
		if (qname != nullptr)
		{
			::free(qname);
			qname = nullptr;
		}

		if (newqname != nullptr)
		{
			qname = ::strdup(newqname);
		}
	}


	DTYPE THQFATTR1
	THQwqdataread(void)
	{
		pthread_mutex_lock(&qlock);
		while (aq.empty())
		{
			waitingondataavail = true;
			++waitdataavailcount;
#if defined(PRINTFTRACE) && (PRINTFTRACE > 8)
			waitqmsgs(0);
#endif

#if defined(DEFAULTWAKEUP)
			local_pthread_cond_timewait50msec(&waitdataavail, &qlock);
#else
			pthread_cond_wait(&waitdataavail, &qlock);
#endif
		}

		DTYPE dqentry = aq.getnextitem();

		waitingondataavail = false;

		//
		// Count the freed up queue space
		// and wakeup anyone who is waiting.
		//
		curqcount -= 1;

		//
		// We send signal if we need more.
		//
		if (curqcount <= QWAKEUPSIZE)
		{
			if (waitingonqueuespaceavail)
			{
					pthread_cond_signal(&waitqueueavail);
					++signalqueueavailcount;
			}
			else
			{
				++signalqueueavailoptimize;
			}
		}

		pthread_mutex_unlock(&qlock);
		return dqentry;
	}

	void THQFATTR1
	THQwqrundown(void)
	{
		//
		// since the purposeof this routine is to rundown the queue -- basically
		// when the system might not be totally sane, we try to rundown the queue
		// for a limited number of times/amount of time then we give-up.
		//
		int rundowncount = 10;
		int timeoutmult = 1;

#pragma clang loop unroll(disable)
		for (int ntries = 0; ntries < rundowncount; ntries++)
		{
			pthread_mutex_lock(&qlock);
			{
#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
				waitqrundownmsg(0, ntries);
#endif
				timespec tstime = THQsupport::gettime_with_offset_ms(100 * rundowncount);
				pthread_cond_timedwait(&waitdataavail, &qlock, &tstime);
				rundowncount++;
				timeoutmult *= 2;
				if (timeoutmult > 100)
					timeoutmult = 100;
			}
			if (aq.empty())
			{
				pthread_mutex_unlock(&qlock);
#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
				waitqrundownmsg(1);
#endif
				break;
			}
			pthread_mutex_unlock(&qlock);
		}

	}

	void THQFATTR1
	THQwqspaceandinsert(DTYPE dqentry)
	{
		pthread_mutex_lock(&qlock);

		if (!NOBLOCKSTORE)
		{
			while (curqcount >= MAXQSIZE)
			{
				++waitqueueavailcount;
				waitingonqueuespaceavail = true;
#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
				waitqspacemsg(0, dqentry->iseof());
#endif
#if defined(DEFAULTWAKEUP)
				local_pthread_cond_timewait50msec(&waitqueueavail, &qlock);
#else
				pthread_cond_wait(&waitqueueavail, &qlock);
#endif
			}
		}

		//
		// Have to save iseof here, because we don't theoretically own dqentry once
		// we have queued it...  Yeah, I know that we haven't released the mutex yet,
		// but trying to be as clean as possible.
		//
		bool iseof = dqentry->iseof();
		waitingonqueuespaceavail = false;
		curqcount += 1;
		aq.additem(dqentry);

		if (waitingondataavail)
		{
			pthread_cond_signal(&waitdataavail);
			++signaldataavailcount;
		}
#if defined(PRINTFTRACE) && (PRINTFTRACE > 9)
		waitqspacemsg(1, iseof);
#endif
		pthread_mutex_unlock(&qlock);
		return;
	}

	void THQFATTR1
	THQwqspaceandeof(DTYPE dqentry)
	{
		dqentry->seteof();
		waitqspaceandinsert(dqentry);
		return;
	}

	void
	dumpqueuestatus(void)
	{
#if 1
		::fprintf(stderr,"queue %s: waitd: %ld, signald: %ld, waitq: %ld, signalq: %ld, signalqopt: %ld\n",
			qname, (int64_t) waitdataavailcount, (int64_t) signaldataavailcount, (int64_t) waitqueueavailcount,
			(int64_t) signalqueueavailcount, (int64_t) signalqueueavailoptimize);
#else
		::fprintf(stderr,"queue %s: waitdata: %d, signaldata: %d, waitqueue: %d, signalqueue: %d, signalqueueopt: %d\n",
			qname, (int) waitdataavailcount, (int) signaldataavailcount, (int) waitqueueavailcount,
			(int) signalqueueavailcount, (int) signalqueueavailoptimize);
#endif
	}

	int
	THQqcount(void) const
	{
		return curqcount;
	}

};

#endif

//
// vim: set ts=4:
//
