
struct int24
{
	char val[3];

	static void
	cvtfpto24(float in, char *out)
	{
		int tmpout;
		double tmpin = double(in);
		tmpin *= 8388608.0;

		tmpout = int(tmpin);

		out[0] = tmpout & 0xff;
		tmpout >>= 8;
		out[1] = tmpout & 0xff;
		tmpout >>= 8;
		out[2] = tmpout & 0xff;
	}

	static float
	cvt24tofp(char *in)
	{
		int32_t tmpin;
		int32_t tmpv;

		//
		// do it this way to extend the sign
		//
		tmpv = in[0];
		tmpv &= 0xff;
		tmpv <<= 8;
		tmpin = tmpv;

		tmpv = in[1];
		tmpv &= 0xff;
		tmpv <<= 16;
		tmpin |= tmpv;

		tmpv = in[2];
		tmpv &= 0xff;
		tmpv <<= 24;
		tmpin |= tmpv;

		tmpin >>= 8;

		float tmpout = tmpin;
		tmpout /= (8388608.0);

		return tmpout;
	}
		
		

public:
	int24(void) : val()
	{
		val[0] = 0;
		val[1] = 0;
		val[2] = 0;
	}

	int24(float inval) : val()
	{
		cvtfpto24(inval, val);
	}

	void
	load(float inval)
	{
		cvtfpto24(inval, val);
	}

	float
	store(void)
	{
		return cvt24tofp(val);
	}

	char*
	addr(void)
	{
		return &val[0];
	}
};

//
// Audio file read/write routines
//
class readaudio
{
	FILE *fp;
	physaudiosample inbuf[RAINBUFSIZE];

	readaudio(const readaudio& init);
	readaudio& operator=(const readaudio& src);

public:
	readaudio(FILE *fparg = nullptr):
		fp(fparg), inbuf()
	{
		if (fparg == nullptr)
		{
			::fprintf(stderr,"Opening data source with invalid FD\n");
			::exit(1);
		}

		physaudiosample patmp;
		for (int i = 0; i < RAINBUFSIZE; i++)
			inbuf[i] = patmp;
	}

	~readaudio(void)
	{
//		fclose(fp);
		fflush(fp);
		fp = nullptr;
	}

	void
	newinfile(const char *fn)
	{
		char tmpfn[1024];
		::snprintf(tmpfn, sizeof tmpfn, "/music/%s", fn);

		for (int i = 0; tmpfn[i]; i++)
		{
			if (tmpfn[i] < ' ')
				tmpfn[i] = '\0';
		}

		FILE *newinfp = fopen(tmpfn, "r");
		if (newinfp)
		{
			::fprintf(stderr,"Starting %s\n", fn);
			fclose(fp);
			fp = newinfp;
		}
		else
		{
			::perror("open error");
			::fprintf(stderr,"Cannot find file %s\n", tmpfn);
		}
	}

	int NOINLINE
	readASbuf(Asample *ibuf, int maxread)
	{
		int i;
		int nshorts = 2 * maxread;
		short tmpbuf[nshorts];
		size_t nr = 0;

		if (ac.nchannels == 2)
		{
retry:
			nr = fread(tmpbuf, 2*sizeof(short), maxread, fp);
			if (nr == 0)
			{
				if (errno == EINTR)
				{
					usleep(1000);
					goto retry;
				}
			}


			for (i = 0; i < nr; i += 1)
			{
				int ri = 2 * i;
				ibuf[i] = Asample(cvt16toMAX(float(tmpbuf[ri])), cvt16toMAX(float(tmpbuf[ri+1])));
			}
		}
		else if (ac.nchannels == 1)
		{
retry1:
			nr = fread(tmpbuf, sizeof(short), maxread, fp);
			if (nr == 0)
			{
				if (errno == EINTR)
				{
					usleep(1000);
					goto retry1;
				}
			}


			for (i = 0; i < nr; i += 1)
			{
				int ri = i;
				ibuf[i] = Asample(cvt16toMAX(float(tmpbuf[ri])), cvt16toMAX(float(tmpbuf[ri])));
			}
		}
			

		return nr;
	}

	
};

class readaudiofloat
{
	FILE *fp;
	physaudiosamplefloat inbuf[RAINBUFSIZE];
	readaudiofloat(const readaudiofloat& init);
	readaudiofloat& operator=(const readaudiofloat& src);

public:
	readaudiofloat(FILE *fparg = nullptr):
		fp(fparg), inbuf()
	{
		if (fparg == nullptr)
		{
			::fprintf(stderr,"Opening data source with invalid FD\n");
			::exit(1);
		}

		physaudiosamplefloat patmp;
		for (int i = 0; i < RAINBUFSIZE; i++)
			inbuf[i] = patmp;
	}

	~readaudiofloat()
	{
	//	fclose(fp);
		fflush(fp);
		fp = nullptr;
	}

	int NOINLINE
	readASbuf(Asample *ibuf, int maxread)
	{
		int i;
		int nfloats = 2*maxread;
		float tmpbuf[nfloats];
		int nr;

		if (fp == NULL)
			return 0;

		if (ac.nchannels == 2)
		{
retry:
			nr = fread(tmpbuf, 2*sizeof(float), maxread, fp);
			if (nr == 0)
			{
				if (errno == EINTR)
				{
					usleep(1000);
					goto retry;
				}
				return 0;
			}

			for (i = 0; i < nr; i += 1)
			{
				int ri = 2 * i;
				ibuf[i] = Asample(tmpbuf[ri], tmpbuf[ri+1]);
			}
		}
		else if (ac.nchannels == 1)
		{
retry1:
			nr = fread(tmpbuf, sizeof(float), maxread, fp);
			if (nr == 0)
			{
				if (errno == EINTR)
				{
					usleep(1000);
					goto retry1;
				}
				return 0;
			}

			for (i = 0; i < nr; i += 1)
			{
				int ri = i;
				ibuf[i] = Asample(tmpbuf[ri], tmpbuf[ri]);
			}
		}
			
		return nr;
	}

};

class writeaudio
{
	int cursample;
	FILE *fp;
	int writeiter;
	physaudiosample outbuf[RAOUTBUFSIZE];

	writeaudio(const writeaudio& init);
	writeaudio& operator=(const writeaudio& src);

	int NOINLINE
	realwrite(const void * const ob, const int nbytestowrite)
	{
		int nbtowr = nbytestowrite;
		int rtval = 0;
		const char *obptr = (const char *) ob;
		while (nbtowr > 0)
		{
			int wrval;
			//
			// If the output file descriptor has been closed, we just flush
			//
			if (ac.outfp)
				wrval = fwrite(obptr, 1, nbtowr, fp);
			else
				wrval = nbtowr;
			if (wrval == -1)
			{
				return -1;
			}

			nbtowr -= wrval;
			rtval += wrval;
			obptr += wrval;
		}
		writeiter += 1;
		return rtval;
	}
			

	void NOINLINE
	dowr(void)
	{
		int __attribute__((unused)) nw;
		if (cursample == 0)
			return;
		int nbytes = cursample * sizeof(physaudiosample);
		nw = realwrite((void *) &outbuf[0], nbytes);
		cursample = 0;
		return;
	}

public:
	writeaudio(FILE *fparg) NOINLINE:
		cursample(0), fp(fparg), writeiter(0), outbuf()
	{
		for (int i = 0; i < RAOUTBUFSIZE; i++)
			outbuf[i] = physaudiosample();
	}

	~writeaudio() NOINLINE
	{
		if (fp == nullptr)
			return;

		if (ac.outfp)
		{
			if (cursample > 0)
				dowr();
	//		fclose(fp);
			fflush(fp);
	
		}
		cursample = 0;
		fp = nullptr;

	}
		

#if 0
	void 
	outsample_pa(const physaudiosample& val)
	{
		outbuf[cursample] = val;
		++cursample;
		if (cursample >= RAOUTBUFSIZE)
			dowr();
		return;
	}

	void 
	outsample(const Asample& asval)
	{
		outsample_pa(asval);
		return;
	}

	bool
	flush(void)
	{
		if (cursample != 0)
			dowr();
		return true;
	}
#endif
		
};

class writeaudiofloat
{
	FILE *fp;
	int writeiter;
	float outbuf[2 * RAOUTBUFSIZE];

	writeaudiofloat(const writeaudiofloat& init);
	writeaudiofloat& operator=(const writeaudiofloat& src);

public:
	writeaudiofloat(FILE *fparg)  NOINLINE :
		fp(fparg), writeiter(0), outbuf()
	{
		::memset(outbuf, 0, sizeof outbuf);
	}

	~writeaudiofloat() NOINLINE
	{
	}


	NOINLINE int
	writebuf(const void * const ob, const int nbytestowrite)
	{
		size_t nbtowr = nbytestowrite;
		size_t rtval = 0;
		const char *obptr = (const char *) ob;
		while (nbtowr > 0)
		{
			size_t wrval;

			//
			// If the output file descriptor has been closed, we just flush
			// JSDNEW
			if (fp)
				wrval = fwrite( obptr, 1, nbtowr, fp);
			else
				wrval = nbtowr;

			if (wrval == (size_t) -1)
			{
				char emsgbuf[emsgbufsize] = {0};
				//
				// Warning -- GNU version
				//
#if !defined(NOSTRERROR)
				char *estr = ::strerror_r(errno, emsgbuf, emsgbufsize - 1);
				::fprintf(stderr,
					"output file (float) write error -- (%d) \"%s\"\n",
					errno, estr);
#else
				::fprintf(stderr,
					"output file (float) write error -- (%d) \"%s\"\n",
					errno, strerror(errno));
#endif
				
				return -1;
			}

			nbtowr -= wrval;
			rtval += wrval;
			obptr += wrval;
		}
		writeiter += 1;
		return int(rtval);
	}

	void
	closefile(void)
	{
//		fclose(fp);
		fflush(fp);
		fp = nullptr;
	}
};

class readaudioint24
{
	FILE *fp;
	int24 inbuf[RAINBUFSIZE];
	readaudioint24(const readaudioint24& init);
	readaudioint24& operator=(const readaudioint24& src);

public:
	readaudioint24(FILE *fparg = nullptr):
		fp(fparg), inbuf()
	{
		if (fparg == nullptr)
		{
			::fprintf(stderr,"Opening data source with invalid FD\n");
			::exit(1);
		}

		for (int i = 0; i < RAINBUFSIZE; i++)
			inbuf[i] = int24{};
	}

	~readaudioint24()
	{
	//	fclose(fp);
		fflush(fp);
		fp = nullptr;
	}

	int NOINLINE
	readASbuf(Asample *ibuf, int maxread)
	{
		int i;
		int nvals = 2*maxread;
		int24 tmpbuf[nvals];
		if (fp == NULL)
			return 0;
		int nr;
		
		if (ac.nchannels == 2)
		{
retry:
			nr = fread((void *) tmpbuf, 2 * sizeof(int24), maxread, fp);
			if (nr == 0)
			{
				if (errno == EINTR)
				{
					usleep(1000);
					goto retry;
				}
				return 0;
			}

			for (i = 0; i < nr; i += 1)
			{
				int ri = 2 * i;

				ibuf[i] = Asample(tmpbuf[ri].store(), tmpbuf[ri+1].store());
		//	fprintf(stderr,"(%f,%f)", ibuf[i].item0(), ibuf[i].item1());
			}
		}
		else if (ac.nchannels == 1)
		{
retry1:
			nr = fread((void *) tmpbuf, sizeof(int24), maxread, fp);
			if (nr == 0)
			{
				if (errno == EINTR)
				{
					usleep(1000);
					goto retry1;
				}
				return 0;
			}

			for (i = 0; i < nr; i += 1)
			{
				int ri = i;
				ibuf[i] = Asample(tmpbuf[ri].store(), tmpbuf[ri].store());
			}
		}
			
			
			
		return nr;
	}

};

#if 0
class writeaudiofloat
{
	FILE *fp;
	int writeiter;
	float outbuf[2 * RAOUTBUFSIZE];

	writeaudiofloat(const writeaudiofloat& init);
	writeaudiofloat& operator=(const writeaudiofloat& src);

public:
	writeaudiofloat(FILE *fparg)  NOINLINE :
		fp(fparg), writeiter(0), outbuf()
	{
		::memset(outbuf, 0, sizeof outbuf);
	}

	~writeaudiofloat() NOINLINE
	{
	}


	NOINLINE int
	writebuf(const void * const ob, const int nbytestowrite)
	{
		size_t nbtowr = nbytestowrite;
		size_t rtval = 0;
		const char *obptr = (const char *) ob;
		while (nbtowr > 0)
		{
			size_t wrval;

			//
			// If the output file descriptor has been closed, we just flush
			// JSDNEW
			if (fp)
				wrval = fwrite( obptr, 1, nbtowr, fp);
			else
				wrval = nbtowr;

			if (wrval == (size_t) -1)
			{
				char emsgbuf[emsgbufsize] = {0};
				//
				// Warning -- GNU version
				//
#if !defined(NOSTRERROR)
				char *estr = ::strerror_r(errno, emsgbuf, emsgbufsize - 1);
				::fprintf(stderr,
					"output file (float) write error -- (%d) \"%s\"\n",
					errno, estr);
#else
				::fprintf(stderr,
					"output file (float) write error -- (%d) \"%s\"\n",
					errno, strerror(errno));
#endif
				
				return -1;
			}

			nbtowr -= wrval;
			rtval += wrval;
			obptr += wrval;
		}
		writeiter += 1;
		return int(rtval);
	}

	void
	closefile(void)
	{
//		fclose(fp);
		fflush(fp);
		fp = nullptr;
	}
};
#endif


