//
// Copyright(C) 2017,2018 John S Dyson
// All rights reserved.

//
//
// This code couldn't be nearly as good as it is without the time, encouragement
// and material help from Richard L Hess, of Toronto, Canada.
//
// This source code is an unpublished proprietary document, and is available 
// for commercial (money making) purposes only under a written license.  If you
// happen to get a copy of this document without agreement, then immediately destroy
// it.
//
// There are trade secrets in this codebase of some value -- there is no
// intent in publishing this code.  Mistaken access to this code is not to be
// construed as being 'published.'
//
// This file contains a suprisingly effective and simple audio expansion processor
//
//
// SOME OF THE CAPABILITY OF THIS CODE IS REDUNDANT -- DUE TO AN EARLIER BUG, I HAD THOUGHT
// THAT IT WAS NECESSARY TO RUN MULTI-AXIS decoding, which might still be helpful for audible
// quality -- but not necessary to decode the targeted compression.  Doing additional
// expansion would require optimized algorithms for that purpose, and such code does exist
// in my finalizing processor program (this code was partially derived from the finalizer.)
//
// THERE ARE LOTS OF DIFFERENCES BETWEEN THIS AN MY FINALIZING PROCESSOR.  The finalizer uses
// 8 bands of audio, and not just 4 like this expander.  It also uses the very advanced
// attack/decay code, and has multiple compression/expansion passes tuned to various purposes,
// with a distortion free hard limiter scheme.  Up to 5-6 of various audio processing passes,
// with no gain control intermod (or at least, very-very little!!!)  Some of the attack/release
// sections can be enabled to have FULL waveform following -- which does not sound ideal, but
// with a little additional modification of the gain control -- works IDEALLY.
//

//
// There is excess precision in the calculations here: vectors of double precision
// instead of using floats...  However, the compiler that I use sometimes chokes on 
// non-full-native vectors -- it sometimes produces worse code with vectors of 4 floats instead
// of vectors of 4 doubles.   Since I am doing very few divisions, the cost of doing
// doubles instead of floats on an Intel architecture isn't very significant, but the cost
// of garbled (but correct) code due to using short vectors of floats can be significant.
// None of the large storage buffers use double precision, so memory bandwidth is also
// not excessively impacted (maybe register spills, short term storage and minimal argument
// passing.)
//
// The finalizing processor just uses floats -- and I am considering a reconfiguration of the
//  existing 8 bands -- possibly up to 16bands.  Rather than being approx octave width, they would
// be approx 1/2 octave -- that opens up some other interesting possibilities, but that
// is a rather aggressive move -- and might be more trouble than it is worth.  It is already
// tricky to properly handle the 8 bands without overly 'flattening' the spectrum.  I am
// using some good methods that allow the processor to maintain the same spectrum shape without
// creating 'policies' which might change the sound.
//
// HINTS ON READING THE VECTOR CODE:
// * the vector type vtype4 is a vector of 4 double precision floats
//
// * when doing a constructor or type cast that looks like 'val = vtype4(-1.0)', it
//       fills all vector members with the specified value.
//
// * when doing a constructor or type cast that looks like 'val = vtype4({a,b,c,d})',
//       it fills in each element as val[0] = a, val[1] = b, val[2] = c,
//       val[3] = d.
//
// * very important: the vector type descriptor is a type in its own right.  That
//       class type describes all aspects of the integer or float vector type
//       of the same shape.  For example, an vector of 4 32bit floating point
//       and a vector of 4 32bit integers use the same descriptor type.  The
//       descriptor describes the type sizes, the underlying types, etc.
//       The actual vector type is usually either varrayT<VDESC> or IvarrayT<VDESC>.
//       There is also a raw vector type like UvarrayT<VDESC> that remembers the
//       shape, but data that has that type can only do conversions or raw assignments.
//       "vtype4" is a typedef of varrayT<TYPEvd4x1>, which means 'a vector of
//       four doubles with a count of one such HW vector.
//
// * math between vectors of the same type are supported.  All of +,-,*,/,
//       and, or, xor, Vmax, Vmin, Vgt, Vlt, Vge, Vle, (Vselect -- does
//			an ?: type operation on each element of the vector),
//       pow(vec,scalar), pow(vec,vec), log, log10, exp, exp10, dB2lin, lin2dB,
//       ([] -- array access of each element), var.fprint(FILEP, fmt),
//       and probably other things.
//
// * some math between vectors and scalars is very partially supported -- I haven't
//      filled in the combinations that support all of the vector/scalar combos.
//      Multiplication that does 'vector *= scalar' does work, but not much more.
//
// * some math between vector types (e.g. vector of 4 doubles vs. vector of 4 floats) is
//      partially supported.  Assignments ARE supported between types, but are meant to be
//      warned/errored if misused -- so mistakes are attempted to be caught.
//
// * some math between vector types that make little sense almost works (e.g. addition
//      between a vector size of 4 and a vector size of 8.) Expect 'interesting'
//      results until this is debugged.
//
// * no attempt is made to support matrices.
//
// * there are integer analogs of all floating point vector types.  Math between such types
//      is limited due to lack of good semantic definition.
//
// * assignment between integer and floating point vector types does the full type conversion,
//      but no special cases are handled, and there is no unsigned integer type yet.
//
// * there is an idiom for RAW copying between integer and floating point types "intvec = intvectype(fpvec.raw())"
//      which devolves the fp vec type to a raw type (with only conversion operations), and then raw copies
//      to the integer.  There are limited purposes for this, but some of the HW vector
//      opcodes (vblendvps/vblendvpd) encourage a certain mix between integer and FP because of the
//      CPU/HW use of an fp value as a logical bitmask.
//
// * the vector code is designed to support multiple HW vectors as one logical one -- but
//      the multiple vectors aren't fully supported yet.  All of the structure does exist.
//


//
// All timings are for 44.1k -- there are some underlying support mechanisms that handle the conversions
// to other sample rates.
//
// For REAL fun, disassemble this turkey -- the vector code can make it almost impossible to trace without
// experience in working with the vector CPU operations. :-).
//

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>

#include <ctype.h>

#include <malloc.h>
#include <signal.h>

#include "vtype.h"

#undef WIDEBAND
#define VTAVNOINLINE __attribute__((noinline))

//
// itqA.h contains manifest constants
//
#include "itqA.h"

//
// general typedefs
//
#include "itq-typedefs.h"

//
// inter-thread communicatsion queues/queue management (thread stuff is pretty much dead for
// the expansion processor -- it is too simple to require threading.)
//
//#include "itq-qdefs.h"

//
// miscellaneous filters (IIR/FIR filter code, filter coefficients/etc.)
//
#include "agcfilters.h"
#include "math.h"

//
// Stupid/poorly written low level audio I/O routines
//
#include "audioio.h"

//
// Data recording/gain logging/etc
//
#include "accuminfo.h"

#include "smfilt.h"


namespace da
{
		static constexpr float SMALLVAL = -1.0e+37;
		float TONEOFFSET = 10.0f;
};

//
// This is a filter class from the very powerful 'smfilt.h' file which contains all kinds of attack/release schemes.
// I have included it here so that the dolbyA code can be kept seperate so that it can be distributed without giving
// away my cool secrets. :-).
//

template <typename DT = double, int NSTEPS = 16>
class atkos
{
	DT atks[NSTEPS];
	DT atksmod[NSTEPS];
	DT aratios[NSTEPS];
	DT atkst, atkend, atkratio;
	float tscale;
	
public:
	atkos(float Atscale, DT Aatkst, DT Aatkend) :
		atks(), atksmod(), aratios(),
		atkst(Aatkst), atkend(Aatkend), atkratio(Aatkend/Aatkst),
		tscale(Atscale)
	{

		tscale = (Atscale < 1.0f) ? 1.0f : Atscale;

		atkst = Aatkst * tscale;
		atkst = Vmin(atkst, DT(1.0));

		atkend = Aatkend * tscale;
		atkend = Vmin(atkend, DT(1.0));

		atkratio = pow(DT(atkend / atkst), (DT(1.0) / DT(NSTEPS-1)));

		DT atktmp{atkst};
		DT atkstart = (atkend / atktmp);

		for (int i = 0; i < NSTEPS; i++)
		{
			aratios[i] = (atkend / atktmp) - atkstart;
			atks[i] = atktmp;
			atksmod[i] = atktmp / (DT(1.0) + atktmp);
			atktmp *= atkratio;
		}
	}

	atkos(void) : atkos(1.0f, 1.0f, 1.0f)
	{
	}

	atkos(DT Aatkst, DT Aatkend) :
		atkos(1.0f, Aatkst, Aatkend) {}

	atkos(const atkos& init) :
		atks(), atksmod(), aratios(),
		atkst(init.atkst), atkend(init.atkend), atkratio(init.atkratio),
		tscale(init.tscale)
	{

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = init.atks[i];
			atksmod[i] = init.atksmod[i];
			aratios[i] = init.aratios[i];
		}
	}

	atkos& NOINLINE
	operator=(const atkos& src)
	{
		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = src.atks[i];
			atksmod[i] = src.atksmod[i];
			aratios[i] = src.aratios[i];
		}
		atkst = src.atkst;
		atkend = src.atkend;
		atkratio = src.atkratio;
		tscale = src.tscale;
		
		return *this;
	}

	void
	setatk(int idx, DT val)
	{
		atks[idx] = val;
		atksmod[idx] = val / (DT(1.0) + val);
		aratios[idx] = atkend / val;
	}

	void
	setdcy(int idx, DT val)
	{
	}

	const DT&
	atk( int idx) const
	{
		return atks[idx];
	}

	const DT&
	atkmod( int idx) const
	{
		return atksmod[idx];
	}

	const DT&
	aratio( int idx) const
	{
		return aratios[idx];
	}

	DT
	aratiorel( int maxidx, int idx) const
	{
		return atksmod[maxidx] / atksmod[idx];
	}

	const DT&
	dcy( int idx) const
	{
	}

	void 
	retunedcys(DT fromdcy, DT todcy)
	{
	}

};


template <typename VDEF, int NSTEPS=16>
class VtavfiltS
{

	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> VIT;
	typedef typename VDEF::FT FT;

public:
	atkos<VT,NSTEPS> ad;

	VT curval[NSTEPS];
	VT zval[NSTEPS];
	VT initval;
	VT atks, atke;
	VT dcy, dcymod;
	VT minaratio;

public:
	typedef VT bsa;
	typedef VIT bsai;
	static constexpr int VS = VDEF::VS;

	VtavfiltS(void) :
		ad(), curval(), zval(), initval(),
		atks(), atke(), dcy(), dcymod(), minaratio()
	{
	}


	VtavfiltS(VT atkst, VT atkend,
		VT dcyval, VT Ainitval=FT(0.0f)) VTAVNOINLINE :
		ad(atkst, atkend),
		curval(), zval(), initval(Ainitval),
		atks(atkst), atke(atkend), dcy(dcyval), dcymod(dcyval / (VT(1.0) + dcyval))
	{
		for (int j = 0; j < NSTEPS; j++)
		{
			curval[j] = VT(initval);
		}
		minaratio = dcymod / ad.atkmod(NSTEPS - 1);
	}

	VtavfiltS(FT atkst, FT atkend, FT dcyval, FT Ainitval=FT(0.0f)) VTAVNOINLINE :
		ad(VT(atkst), VT(atkend)),
		curval(), zval(), initval(Ainitval),
		atks(VT(atkst)), atke(VT(atkend)), dcy(dcyval), dcymod(dcymod)
	{
		for (int j = 0; j < NSTEPS; j++)
		{
			curval[j] = VT(initval);
			zval[j] = VT(initval);
		}
		minaratio = dcymod / ad.atkmod(NSTEPS - 1);
	}

	VtavfiltS(const VtavfiltS<VDEF,NSTEPS>& init) VTAVNOINLINE :
		ad(init.ad), zval(), curval(), initval(init.initval),
		atks(init.atks), atke(init.atke), dcy(init.dcy), dcymod(init.dcymod)
		
	{
		for (int j = 0; j < NSTEPS; j++)
		{
			curval[j] = init.curval[j];
			zval[j] = init.zval[j];
		}
		minaratio = dcymod / ad.atkmod(NSTEPS - 1);
	}

	VtavfiltS&
	operator=(const VtavfiltS<VDEF,NSTEPS>& src)
	{
		ad = src.ad;
		initval = src.initval;
		atks = src.atks;
		atke = src.atke;
		dcy = src.dcy;
		dcymod = src.dcymod;

		for (int j = 0; j < NSTEPS; j++)
		{
			curval[j] = src.curval[j];
			zval[j] = src.zval[j];
		}
		minaratio = dcymod / ad.atkmod(NSTEPS - 1);

		return *this;
	}

	~VtavfiltS()
	{
	}


	VT VTAVNOINLINE
	next(const VT& newval,int aratiomode = 0)
	{
		VT rtvalX(da::SMALLVAL);

		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tzv = zval[i];
			const VT vdiff = newval - Tzv;

			const VT& atkmod(ad.atkmod(i));
			VT& Tcv = curval[i];


			VT aratio(((aratiomode == 1) ? VT(FT(1.0)) : ((aratiomode == 0) ? ad.aratio(i) : ad.aratio(i) * ad.aratio(i))));
			if (aratiomode == 3)
			{
				aratio *= ad.aratio(i);
				VT artmp = aratio * VT(FT(0.66));
				aratio *= ad.aratio(i) * ad.aratio(i);
				artmp += aratio * VT(FT(0.22));
				aratio *= ad.aratio(i) * ad.aratio(i);
				artmp += aratio * VT(FT(0.12));
				aratio = artmp;
			}
				
			aratio = Vmax(aratio, minaratio);
			const VT tmpA = vdiff * atkmod * aratio;
			const VT tmpD = vdiff * dcymod;

			const VT tmp = Vselect(Vgt(newval, Tcv), tmpA, tmpD);
			VT rtvalXtmp = Tzv + tmp;
			Tzv = rtvalXtmp + tmp;
			Tcv = rtvalXtmp;

			rtvalX = Vmax(rtvalX, rtvalXtmp);

		}

		return rtvalX;
	}

	VT VTAVNOINLINE
	next(const VT& newval, int aratiomode, VT decaymult)
	{
		VT rtvalX(da::SMALLVAL);
		VT dcymodx = dcymod;
		VT minaratiox = minaratio;

		if (!alltrue(decaymult == VT(1.0f)))
		{
			VT ndcy = dcy * decaymult;
			ndcy= Vmin(ndcy, VT(1.0f));
			dcymodx = ndcy / (VT(1.0f) + ndcy);
			minaratiox = dcymodx / ad.atkmod(NSTEPS - 1);
		}
			
		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tzv = zval[i];
			const VT vdiff = newval - Tzv;

			const VT& atkmod(ad.atkmod(i));
			VT& Tcv = curval[i];

			VT aratio(((aratiomode == 1) ? VT(FT(1.0)) : ((aratiomode == 0) ? ad.aratio(i) : ad.aratio(i) * ad.aratio(i))));
			if (aratiomode == 3)
			{
				aratio *= ad.aratio(i);
				VT artmp = aratio * VT(FT(0.66));
				aratio *= ad.aratio(i) * ad.aratio(i);
				artmp += aratio * VT(FT(0.22));
				aratio *= ad.aratio(i) * ad.aratio(i);
				artmp += aratio * VT(FT(0.12));
				aratio = artmp;
			}
			aratio = Vmax(aratio, minaratiox);
			const VT tmpA = vdiff * atkmod * aratio;
			const VT tmpD = vdiff * dcymodx;

			const VT tmp = Vselect(Vgt(newval, Tcv), tmpA, tmpD);
			VT rtvalXtmp = Tzv + tmp;
			Tzv = rtvalXtmp + tmp;
			Tcv = rtvalXtmp;

			rtvalX = Vmax(rtvalX, rtvalXtmp);

		}

		return rtvalX;
	}

	VT VTAVNOINLINE
	nextAD(const VT& newval, int aratiomode, VT decaymult)
	{
		VT rtvalX(da::SMALLVAL);
		VT dcymodx = dcymod;
		VT minaratiox = minaratio;

		if (!alltrue(decaymult == VT(1.0f)))
		{
			VT ndcy = dcy * decaymult;
			ndcy= Vmin(ndcy, VT(1.0f));
			dcymodx = ndcy / (VT(1.0f) + ndcy);
			minaratiox = dcymodx / ad.atkmod(NSTEPS - 1);
		}
			
		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tzv = zval[i];
			const VT vdiff = newval - Tzv;

			const VT& atkmod(ad.atkmod(i) * decaymult);
			VT& Tcv = curval[i];

			VT aratio(((aratiomode == 1) ? VT(FT(1.0)) : ((aratiomode == 0) ? ad.aratio(i) : ad.aratio(i) * ad.aratio(i))));
			if (aratiomode == 3)
			{
				aratio *= ad.aratio(i);
				VT artmp = aratio * VT(FT(0.66));
				aratio *= ad.aratio(i) * ad.aratio(i);
				artmp += aratio * VT(FT(0.22));
				aratio *= ad.aratio(i) * ad.aratio(i);
				artmp += aratio * VT(FT(0.12));
				aratio = artmp;
			}
			aratio = Vmax(aratio, minaratiox);
			const VT tmpA = vdiff * atkmod * aratio;
			const VT tmpD = vdiff * dcymodx;

			const VT tmp = Vselect(Vgt(newval, Tcv), tmpA, tmpD);
			VT rtvalXtmp = Tzv + tmp;
			Tzv = rtvalXtmp + tmp;
			Tcv = rtvalXtmp;

			rtvalX = Vmax(rtvalX, rtvalXtmp);

		}

		return rtvalX;
	}

	template <typename PV> 
	PV operator()(const PV& input)
	{
		return next(input);
	}

	template <typename PV> 
	PV operator()(const PV& input, bool skiparatio)
	{
		return next(input, skiparatio);
	}

	template <typename PV> 
	PV operator()(const PV& input, bool skiparatio, const PV& div)
	{
		return next(input, skiparatio, div);
	}
};


#undef NOINLINE

//
// If we are running a recent GCC, IF we want 'noinline', we probably
// want 'noclone' also.
//
#if defined(__GNUC__) && (__GNUC__ > 5)
#define NOINLINE __attribute__((noinline,noclone))
#else
#define NOINLINE __attribute__((noinline))
#endif



//
// The Vector/float types used to be selectable due to problems with the control
// signal filter dynamic ranges.  That problem has been worked around, so we
// use a fixed set of types.  Now, we use the varray series of types, and
// can easily select the types as needed.  Also, dynamic range issues are in the past.
//

namespace da
{

static constexpr int dBNCHANS = 2;
struct dBfilterVAL
{
	float v[dBNCHANS];
};

class dBfilterSIMPLE
{
	dBfilterVAL curval;

	float atk, dcy;

public:
	dBfilterSIMPLE(float atkA, float dcyA) :
		curval(), atk(atkA), dcy(dcyA)
	{
	}

	
	const dBfilterVAL&
	next( dBfilterVAL in)
	{

		static const float deadzone = 0.0;
//		static const float deadzone = 1.0;
		for (int i = 0; i < dBNCHANS; i++)
		{
			float ldiff = in.v[i] - curval.v[i];
	
			if (ldiff > deadzone)
			{
				curval.v[i] += ldiff * atk;
			}
			else if (ldiff < -1.0 * deadzone)
			{
				curval.v[i] += ldiff * dcy;
			}
		}

		return curval;
	}
};
		
			
class proccomp
{
	using VDEF = TYPEvf8x1;
	using vt = varrayT<VDEF>;
	using TfltSASm = VtavfiltS<VDEF8, 16>; // version of Tflt that supports all 8 (L+R) channels processed at a time

	IBPF INfilt;

	VfilterdB<VDEF, 24, 3> rfilt1;
	VfilterdB<VDEF, 24> rfilt1A;
	VfilterdB<VDEF, 24> rfilt1B;
	VfilterdB<VDEF, 24> rfilt1C;
	VfilterdB<VDEF, 24> rfilt3;
	VfilterdB<VDEF, 24> rfilt4;
	VfilterdB<VDEF, 24> rfilt5;
	TfltSASm pf;
	TfltSASm limitfilt0;
	TfltSASm limitfilt1;
	TfltSASm limitfilt2;


public:
	proccomp(void) :
		rfilt1(aspec.inSRATE, 6.0f, 1.00f, 6.0f, 1.00f, -60.0f),
		rfilt1A(aspec.inSRATE, 6.0f, 1.00f, 6.0f, 1.00f, -60.0f),
		rfilt1B(aspec.inSRATE, 6.0f, 1.00f, 6.0f, 1.00f, -60.0f),
		rfilt1C(aspec.inSRATE, 6.0f, 1.00f, 6.0f, 1.00f, -60.0f),
		rfilt3(aspec.inSRATE, 120.0f, 1.0f, 120.0f, 1.0f, -60.0f),
		rfilt4(aspec.inSRATE, 384.0f, 5.0f, 384.0f, 5.0f, -60.0f),
		rfilt5(aspec.inSRATE, 1024.0f, 20.0f, 1024.0f, 20.0f, -60.0f),
		pf(vt((0.45)), sratetmod(0.0330f), sratetmod(0.0330f), vt(0.1)),
		limitfilt0(
			vt(sratetmod(0.01f)),
			sratetmod(vt(0.0010f)),
			sratetmod(vt(0.00030f)),
			vt(0.0)),
		limitfilt1(
			vt(sratetmod(0.10f)),
			sratetmod(vt({0.0010f, 0.00500f, 0.010f, 0.010f, 0.0010f, 0.0050f, 0.010f, 0.010f})),
			sratetmod(vt({0.00038f, 0.0010f, 0.0010f, 0.0010f, 0.00022f, 0.00100f, 0.0010f, 0.0010f})),
			vt(0.0)),
		limitfilt2(vt(0.45f), sratetmod(vt(0.02f)), sratetmod(vt(0.0002f)), vt(0.0))
		{}

	
	Asample
	next(fpdata* fpd, const Asample& src)
	{
		Asample sigin[4];
		Asample sigout[4];
		float lvlL[4];
		float lvlR[4];

		INfilt(src, sigin);
		
		
		for (int i = 0; i < 4; i++)
		{
			lvlL[i] = sigin[i].item0() * sigin[i].item0();
			lvlR[i] = sigin[i].item1() * sigin[i].item1();
			sigout[i] = sigin[i];
		}

		vt lvl(0.0001f);
		for (int i = 0; i < 4; i++)
		{
			lvl[i] = lvlL[i];
			lvl[i+4] = lvlR[i];
		}

#if 0
		if (lvl[2] < lvl[1])
			lvl[2] = lvl[1];
		if (lvl[2+4] < lvl[1+4])
			lvl[2+4] = lvl[1+4];
#endif

		vt dBlvl = lin2dB(lvl);
		
		dBlvl += vt(ac.thresholdlval) * vt(2.0f);
		dBlvl = Vmax(dBlvl, vt(-60.0f) * vt(2.0));


		vt dBrmslvl3 = rfilt3.nextAccelB(dBlvl);
		vt dBrmslvl4 = rfilt4.nextAccelB(dBlvl);
		vt dBrmslvl5 = rfilt5.nextAccelA(dBlvl);

		dBrmslvl3 = Vmax(dBrmslvl3, dBrmslvl4 * 0.67 - vt(1.50f));
		dBrmslvl3 = Vmax(dBrmslvl3, dBrmslvl5 * 0.33 - vt(3.00f));

#if 0
		vt dBrmslvl1 = rfilt1.nextAccelA(dBrmslvl3);
#else
		vt dBrmslvl1A = rfilt1.nextAccelB(dBrmslvl3);
		vt dBrmslvl1B = rfilt1.nextAccelA(Vmax(dBrmslvl3, dBrmslvl1A));
		vt dBrmslvl1C = rfilt1.nextAccelA(Vmax(dBrmslvl1A, dBrmslvl1B));
		vt dBrmslvl1 = Vmax(dBrmslvl1B, dBrmslvl1C);
#endif

		vt dBrmslvl = Vmax(dBrmslvl3, dBrmslvl1);

		//dBrmslvl *= 0.50f;
		
		for (int i = 0; i < 4; i++)
		{
			if (dBrmslvl[i+4] > dBrmslvl[i])
				dBrmslvl[i] = dBrmslvl[i+4];
			if (dBrmslvl[i] > dBrmslvl[i+4])
				dBrmslvl[i+4] = dBrmslvl[i];
		}
		
		vt dBrmsgain = dBrmslvl;
			
		dBrmsgain *= vt(ac.cratio);
		//fprintf(stderr,"(%f)", ac.maxgain);
		dBrmsgain = Vmax(dBrmsgain, vt(-1.0f * ac.maxgain));

		vt LINrmsgain = dB2lin(dBrmsgain);
		LINrmsgain = pf(LINrmsgain,1);
		LINrmsgain = vt(1.0f) / LINrmsgain;

		Asample asgain[3];
		for (int i = 0; i < 4; i++)
		{
			asgain[i] = {LINrmsgain[i], LINrmsgain[i]};
		}

		for (int i = 0; i < 4; i++)
		{
			sigout[i] = sigin[i] * asgain[i];
		}

		Asample limitin0;

		limitin0 = sigout[0];
		limitin0 += sigout[1];
		limitin0 += sigout[2];

		vt LINmaxlevel0 = vt(dB2lin(ac.maxlevel));
		vt lilevel0(0.1);
		lilevel0[0] = limitin0.item0() * limitin0.item0();
		lilevel0[1] = limitin0.item1() * limitin0.item1();
		lilevel0 = limitfilt0(Vmax(LINmaxlevel0, vsqrt(lilevel0)), 2);

		vt ligain0 = (LINmaxlevel0 / Vmax(LINmaxlevel0, lilevel0));
		Asample asligain0{ligain0[0], ligain0[1]};

		Asample limitin1 = limitin0 * asligain0 * -1.0f;

		vt LINmaxlevel1 = vt(dB2lin(ac.maxlevel));
		vt lilevel1(0.1);
		lilevel1[0] = limitin1.item0() * limitin1.item0();
		lilevel1[1] = limitin1.item1() * limitin1.item1();
		lilevel1 = limitfilt1(Vmax(LINmaxlevel1, vsqrt(lilevel1)), 2);

		vt ligain1 = (LINmaxlevel1 / Vmax(lilevel1, vt(LINmaxlevel1)));
		Asample asligain1{ligain1[0], ligain1[1]};

		Asample limitin2 = limitin1 * asligain1 * -1.0f;

		vt LINmaxlevel2 = vt(dB2lin(ac.maxlevel));
		vt lilevel2(0.1);
		lilevel2[0] = limitin2.item0() * limitin2.item0();
		lilevel2[1] = limitin2.item1() * limitin2.item1();
		lilevel2 = limitfilt2(Vmax(LINmaxlevel2, vsqrt(lilevel2)), 0);

		vt ligain2 = (LINmaxlevel2 / Vmax(lilevel2, vt(LINmaxlevel2)));
		Asample asligain2{ligain2[0], ligain2[1]};

		Asample rtval = limitin2 * asligain2;

		vt ligain = ligain0 * ligain1 * ligain2;
		fpd->pgain = LINrmsgain * vt({ligain[0],ligain[0],ligain[0],ligain[0], ligain[1], ligain[1], ligain[1], ligain[1]});
		fpd->pgain[3] = ligain[0];
		fpd->pgain[3+NBANDS] = ligain[1];
		return rtval;
	}

	Asample
	operator() (fpdata* fpd, const Asample& src)
	{
		return next(fpd, src);
	}


};


class dbexoutput
{

public:
	dbexoutput(void) 
	{
	}

	void NOINLINE
	next(fpdata *fps)
	{

		fps->sigout = fps->sigin;

		vtype4 pgainl{};
		vtype4 pgainr{};
		for (int i = 0; i < NBANDS; i++)
		{
				pgainl[i] = fps->pgain[i];
				pgainr[i] = fps->pgain[i + NBANDS];
		}

		vtype4 avggain = (pgainl + pgainr) * vtype4(VFP4(0.50));

		//
		// looking at the schematic -- the fet has a big resistor in the gate -- enough
		// that it should be accounted for by a little bit of a TC -- I used 1msec.
		//

		if (!ac.filtertest)
		{
			DLBYAst[0].next(avggain);
			DLBYAstl[0].next(pgainl);
			DLBYAstr[0].next(pgainr);
		}

	}
	
	int
	maxdelay(void) const
	{
		return 0;
	}

};

};

void*
THinput(const void * threadinfo)
{

	const thread_init_list *actualthreadargs = (const thread_init_list *) threadinfo;
	setnoht();

	da::proccomp comp;

	paudiochunk pa = nullptr;
	while(1)
	{
		pa = dinput.THQwqdataread();
		if (pa->iseof())
		{
			break;
		}

		audiopayload* ap = &pa->payload();
		int samplecount = ap->samplecount;
#if defined(DBGPRT)
		fprintf(stderr,"THinput got %d samples\n", samplecount);
#endif

		for (int idx = 0; idx < samplecount; idx++)
		{
			da::fpdatap fpd = ap->samples[idx];
			Asample& as = fpd->sigin;
			as.item0() *= ac.LINingainlval;
			as.item1() *= ac.LINingainrval;
			if (!ac.filtertest) INst.next( as.sumsq(), as.msumsq());


			fpd->sigin = comp(fpd, as);
		}
	
		dendproc.THQwqspaceandinsert(pa, false);
	}

	if (pa != nullptr)
	{
		dendproc.THQwqspaceandinsert(pa, false);
	}
		
	pthread_exit((void *) &actualthreadargs->EXITrtval);
	return nullptr;
}

void*
THendproc(const void * threadinfo)
{
	static int totalsamples70;
	static int limit99, limit70, limit50, limit25;

	const thread_init_list *actualthreadargs = (const thread_init_list *) threadinfo;
	setht();

	thread_init_list *til = (thread_init_list *) threadinfo;

	paudiochunk pa;
		
	da::dbexoutput op;
	
	while(1)
	{
		pa = dendproc.THQwqdataread();
		if (pa->iseof())
		{
			doutput.THQwqspaceandeof(pa);
			break;
		}
		int nwritten = 0;

		audiopayload* ap = &pa->payload();
		int samplecount = ap->samplecount;
#if defined(DBGPRT)
		fprintf(stderr,"%s got %d samples\n", actualthreadargs->threadname, samplecount);
#endif

		for (int idx = 0; idx < samplecount; idx++)
		{
			da::fpdatap fpd = ap->samples[idx];
			op.next(fpd);
			nwritten += 1;
			Asample& as = fpd->sigout;

			as.item0() *= ac.LINoutgainlval;
			as.item1() *= ac.LINoutgainrval;

			OUTst.next( as.sumsq(), as.msumsq());
			accuminfo::accumulatelevel((const Asample&) as, totalsamples70, limit99, limit70, limit50, limit25);
		}

		if (nwritten > 0)
			doutput.THQwqspaceandinsert(pa, false);
		else
			demptyqueue.THQwqspaceandinsert(pa, false);
	}

	pthread_exit((void *) &actualthreadargs->EXITrtval);
	return nullptr;
}


//
// vi: set ts=3:
//

