//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 

#include <sys/types.h>
#include <sys/fcntl.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>
// #include <math.h>
#include <ctype.h>
#include <signal.h>
#if !defined(WINDOWS64)
#include <termios.h>
#include <sys/ioctl.h>
#endif

#include <malloc.h>
#include <signal.h>

#include <xmmintrin.h>
#include <pmmintrin.h>
#include <immintrin.h>




#include "itqA.h"
#include "itq-typedefs.h"
//#include "itq-qdefs.h"

#include "agcfilters.h"
#include "audioio.h"

#include "wav.h"

#include "accuminfo.h"

#if defined(LOCKCOMPSTATS)
pthread_mutex_t infolock = PTHREAD_MUTEX_INITIALIZER;
#endif

#undef NOINLINE
#if defined(__GNUC__) && (__GNUC__ > 5)
#define NOINLINE __attribute__((noinline,noclone))
#else
#define NOINLINE __attribute__((noinline))
#endif

const char *lstrerror_r(int errcode, char *bufarg, int bufargsize);

volatile bool newfilewait = false;
thread_local char *newfile = nullptr;

static char inwavheaderbuf[1024];
static int incurdatasize;
static int inptrtoformat;

audiospec aspec;
namespace wavdata
{
	size_t filesize = 0;
	size_t newfilesize = 0;
	size_t datasize = 0;
	size_t newdatasize = 0;
	size_t datataglocation = 0;
	size_t datalocation = 0;
	size_t fmtlen = 0;
	char *fmtptr = nullptr;
	size_t fmtlocation = 0;
	size_t factlen = 0;
	char *factptr = nullptr;
	size_t factlocation = 0;
	size_t ds64len = 0;
	char *ds64ptr = nullptr;
	size_t ds64location = 0;
	size_t chnalen = 0;
	char *chnaptr = nullptr;
	size_t chnalocation = 0;
	size_t axmllen = 0;
	char *axmlptr = nullptr;
	size_t axmllocation = 0;
	size_t iXMLlen = 0;
	char *iXMLptr = nullptr;
	size_t iXMLlocation = 0;
	size_t bextlen = 0;
	char *bextptr = nullptr;
	size_t bextlocation = 0;
	rwav0 wavtop;
	tagsize datatop;
};

void
initwavdata(void)
{
	wavdata::filesize = 0;
	wavdata::newfilesize = 0;
	wavdata::datasize = 0;
	wavdata::newdatasize = 0;
	wavdata::datataglocation = 0;
	wavdata::datalocation = 0;
	wavdata::fmtlen = 0;
	if (wavdata::fmtptr != nullptr)
		free(wavdata::fmtptr);
	wavdata::fmtptr = nullptr;
	wavdata::fmtlocation = 0;

	wavdata::factlen = 0;
	if (wavdata::factptr != nullptr)
		free(wavdata::factptr);
	wavdata::factptr = nullptr;
	wavdata::factlocation = 0 ;

	wavdata::ds64len = 0;
	if (wavdata::ds64ptr != nullptr)
		free(wavdata::ds64ptr);
	wavdata::ds64ptr = nullptr;
	wavdata::ds64location = 0;

	wavdata::chnalen = 0;
	if (wavdata::chnaptr != nullptr)
		free(wavdata::chnaptr);
	wavdata::chnaptr = nullptr;
	wavdata::chnalocation = 0;

	wavdata::axmllen = 0;
	if (wavdata::axmlptr != nullptr)
		free(wavdata::axmlptr);
	wavdata::axmlptr = nullptr;
	wavdata::axmllocation = 0;

	wavdata::iXMLlen = 0;
	if (wavdata::iXMLptr != nullptr)
		free(wavdata::iXMLptr);
	wavdata::iXMLptr = nullptr;
	wavdata::iXMLlocation = 0;

	wavdata::bextlen = 0;
	if (wavdata::bextptr != nullptr)
		free(wavdata::bextptr);
	wavdata::bextptr = nullptr;
	wavdata::bextlocation = 0;

}


int NOINLINE
setupinfile(void)
{
	int headersize;

	aspec.change(44100);

	threadresource::speedreadyreset();
	initwavdata();

	if (isatty(fileno(ac.infp)))
	{
		::fprintf(stderr,"Cannot do audio input from tty\n");
		return 1;
	}

	if (isatty(fileno(ac.outfp)))
	{
		::fprintf(stderr,"Cannot do audio output to tty\n");
		return 1;
	}

	wavdata::newfilesize = 0;
	headersize = finddata(ac.infp, 
		&wavdata::filesize, &wavdata::datasize,
		&wavdata::fmtlen, &wavdata::fmtptr,
		&wavdata::factlen, &wavdata::factptr,
		&wavdata::ds64len, &wavdata::ds64ptr,
		&wavdata::chnalen, &wavdata::chnaptr,
		&wavdata::axmllen, &wavdata::axmlptr,
		&wavdata::bextlen, &wavdata::bextptr,
		&wavdata::iXMLlen, &wavdata::iXMLptr
	);

	if (ac.info > 0x10) fprintf(stderr,"ds64: %ld, %p\n", wavdata::ds64len, wavdata::ds64ptr);

//	fprintf(stderr,"Header size: %d\n", headersize);
	if (headersize == -2)
	{
		::fprintf(stderr,"Error -- definite incompatible input file format\n");
		return 1;
	}

	fmtstruct *fh = (fmtstruct *)wavdata::fmtptr;
	if (fh == nullptr)
	{
		::fprintf(stderr,"Error -- no fmt header found\n");
		return 1;
	}

	int curbps = 16;
	aspec.inBPS = 16;
	int curformat = fh->aformat;
	if (headersize != -1)
	{
		if (curformat == 3)
		{
			ac.rafloat = 1;
			ac.rafloatout = 1;
			ac.raint24out = 0;
			ac.raint24 = 0;
			curbps = 32;
			aspec.inBPS = 32;
		}
		else if (curformat == -2 )
		{
			curbps = fh->bps;
			aspec.inBPS = fh->bps;
			curformat = fh->format;

			if (curformat == 3)
			{
				ac.rafloat = 1;
				ac.rafloatout = 1;
				ac.raint24out = 0;
				ac.raint24 = 0;
			}
			else
			{
				if (curformat == 1)
				{
					if (curbps == 24)
					{
						ac.raint24 = 1;
						ac.raint24out = 1;
						ac.rafloatout = 0;
						ac.rafloat = 0;
					}
					else if (curbps != 16)
					{
						::fprintf(stderr,"Error -- PCM WAV audio format %d at %d bits not supported for input\n", fh->aformat, fh->bps);
						return 1;
					}
				}
			}
		}
		else if (curformat == 1 || curformat == 0)
		{
			aspec.inBPS = fh->bps;
			ac.rafloat = 0;
			ac.rafloatout = 0;
			if (fh->bps == 24)
			{
				ac.raint24out = 1;
				ac.raint24 = 1;
			}
			else
			{
				ac.raint24out = 0;
				ac.raint24 = 0;
			}
		}
		else
		{
			::fprintf(stderr,"Error -- WAV audio format %d not supported for input\n", fh->aformat);
			return 1;
		}

		int cursrate = fh->srate;

#if 0
		if (cursrate != 44100)
		{
			if (cursrate == 48000 || cursrate == 96000 || cursrate == 192000)
			{
				aspec.change(cursrate);
			}
			else
			{
				::fprintf(stderr,"Error -- input data rate %d is not supported\n", cursrate);
				return 1;
			}
		}
#else
		aspec.change(cursrate);
#endif
			
	}
	aspec.outBPS = aspec.inBPS;

	if ((fh->nchannels != 1) && (fh->nchannels != 2))
	{
		::fprintf(stderr,"Error -- only 2 channel audio is supported in this program\n");
		return 1;
	}

	ac.nchannels = fh->nchannels;

	if (ac.rareqfloatout)
	{
		ac.rafloatout = 1;
		ac.raint24out = 0;
		aspec.outBPS = 32;
	}

	if (ac.info > 0x10)
	{
		fprintf(stderr,"bps: %d, srate: %d, fpout: %d, int24out: %d\n",
			curbps, int(aspec.inSRATE), ac.rafloatout, ac.raint24out);
	}

	if (aspec.inSRATE > 224000)
	{
		fprintf(stderr,"***WARNING*** SAMPLE RATE=%d, but program works suboptimally faster than 192k, and will 'choke', but try valiantly above 224k\n", aspec.inSRATE);
		fprintf(stderr," best thing is to use sox to convert, e.g. \"sox infile.wav --type=wav - rate -v 192k | da-avx --thresh=xxx | sox - outfile.wav rate -v %d\"\n", aspec.inSRATE);
	}
	if (aspec.inSRATE < 47000)
	{
		fprintf(stderr,"***WARNING*** SAMPLE RATE=%d, program runs very poorly near the 44.1k sample rate.\nUse sox to convert, e.g. \"sox infile.wav --type=wav - rate -v 96k | da-avx --thresh=xxx | sox - outfile.wav rate -v %d\"\n", aspec.inSRATE, aspec.inSRATE);
	
	}

	if (aspec.inSRATE < 44000)
	{
		fprintf(stderr,"***ERROR*** program not designed to work slower than 44.1k sample rate.\nExperiments show that when program forced to function, it sounds 'crunchy' when run at low sample rates\n");
		fprintf(stderr," best thing is to use sox to convert, e.g. \"sox infile.wav --type=wav - rate -v 48k | da-avx --thresh=xxx | sox - outfile.wav rate -v %.0lf\"\n", double(aspec.inSRATE));
		return 1;
	}
			
//	ac.rafloatout = 1;
//	ac.raint24out = 0;

	return 0;
}

extern void procmusic(paudiochunk pa);

void
buildnewbext(void)
{
#if 0
	char *buf;
	int wval;
	wval = 16;
	if (ac.rafloatout)
		wval = 32;
	else if (ac.raint24out)
		wval = 24;

	const char *mval = "stereo";
	if (ac.nchannels == 1)
		mval = "mono";
	const char *t0val = "Dyson-Hess DA Decoder";

	char *tbext = nullptr;
	if (ac.bext)
	{
		tbext = strdup(ac.bext);
		for(int i = 0; tbext[i] != 0; i++)
		{
			if (tbext[i] == ',')
				tbext[i] = '-';
		}
	}
	int nnew = 0;
	if (tbext == nullptr)
		nnew = asprintf(&buf, "A=PCM,F=%d,W=%d,M=%s,T=%s;Version=%s;Lth=%.2f;Rth=%.2f\r\n", aspec.inSRATE, wval, mval, t0val, programVersion, ac.thresholdlval, ac.thresholdrval);
	else
		nnew = asprintf(&buf, "A=PCM,F=%d,W=%d,M=%s,T=%s;Version=%s;Lth=%.2f;Rth=%.2f;%s\r\n", aspec.inSRATE, wval, mval, t0val, programVersion, ac.thresholdlval, ac.thresholdrval, tbext);
	

	int newlen = wavdata::bextlen + sizeof(tagsize)+nnew+128;
	char *nbuffer = (char *) malloc(newlen);
	memset((void *) nbuffer, 0, newlen);
	if (nbuffer != nullptr)
	{
		memcpy(nbuffer, wavdata::bextptr, wavdata::bextlen + sizeof(tagsize));
		bextraw *bp = (bextraw *) ((char *)nbuffer + sizeof(tagsize));
		char *rawend = nbuffer + wavdata::bextlen + sizeof(tagsize) + nnew;
		for (int i = 0; &bp->CodingHistory[i] < rawend; i++)
		{
			if (bp->CodingHistory[i] == '\0')
			{
				memcpy((void *) &bp->CodingHistory[i], buf, nnew);
				free(wavdata::bextptr);
				wavdata::bextptr = nbuffer;
				size_t newbextsize = ((char *) &bp->CodingHistory[i+nnew] - (char *) nbuffer) - sizeof(tagsize);
	//			fprintf(stderr,"Old bext size: %ld(%lx), New bext size: %ld(%lx)\n",
	//				wavdata::bextlen, wavdata::bextlen, newbextsize, newbextsize);
				newbextsize += 3;
				newbextsize &= ~(3ULL);
				wavdata::bextlen = newbextsize;
				((tagsize *) nbuffer)->len = newbextsize;
				break;
			}
		}
	}

	if (buf != nullptr)
		free(buf);
	if (tbext != nullptr)
		free(tbext);
#endif
}

void
createbext(void)
{
#if 0
	bextraw *br;
	char *br_rawptr;
	int br_rawsize = sizeof(tagsize) + sizeof(bextraw);
	tagsize *ts;
	if (wavdata::bextptr != nullptr)
		return;

	br_rawptr = (char *) malloc(br_rawsize);
	memset(br_rawptr, 0, br_rawsize);

	br = (bextraw *) ((char *) br_rawptr + sizeof(tagsize));

	ts = (tagsize *) br_rawptr;
	memcpy(ts->name.cdata, "bext", 4);
	ts->len = sizeof(bextraw);

	if (ac.bextdesc)
	{
		strncpy(&br->Description[0], ac.bextdesc, sizeof br->Description);
	}
	if (ac.bextorig)
	{
		strncpy(&br->Originator[0], ac.bextorig, sizeof br->Originator);
	}
	if (ac.bextorigref)
	{
		strncpy(&br->OriginatorReference[0], ac.bextorigref, sizeof br->OriginatorReference);
	}

	{
		struct timeval tv;
		struct tm tmval;
		char dbuf[128];
		char tbuf[128];
		gettimeofday( &tv, NULL);

		localtime_r(&tv.tv_sec, &tmval);

		strftime(&dbuf[0], sizeof dbuf, "%F", &tmval);
		strftime(&tbuf[0], sizeof tbuf, "%H:%M:%S", &tmval);

		memcpy(br->OriginationDate, dbuf, sizeof br->OriginationDate);
		memcpy(br->OriginationTime, tbuf, sizeof br->OriginationTime);
	}

	wavdata::bextptr = (char *) br_rawptr;
	wavdata::bextlen = sizeof *br;
	wavdata::bextlocation = 0;
#endif
}
	

wuint64_t
correctsizebps(wuint64_t orig)
{
	wuint64_t newbps;
	newbps = (orig * aspec.outBPS) / aspec.inBPS;
	return newbps;
}

void
build1stpartofoutput(void)
{
	size_t curlocation = 0;
	if ((ac.rafloatout == -1) || (ac.rafloatout == 0))
		ac.rafloatout = 0;

	memcpy(wavdata::datatop.name.cdata, "data", 4);
	if (wavdata::ds64ptr != nullptr)
	{
		memcpy(&wavdata::wavtop.riffsn.cdata, "RF64", 4);
		wavdata::wavtop.rlength = 0xffffffff;
		wavdata::datatop.len = 0xffffffff;
	}
	else
	{
		memcpy(&wavdata::wavtop.riffsn.cdata, "RIFF", 4);
		//wavdata::wavtop.rlength = 0xffffffff;
		wavdata::wavtop.rlength = (uint32_t) correctsizebps(wavdata::filesize);
		wavdata::datatop.len = (uint32_t) correctsizebps(wavdata::datasize);
		if (ac.info > 0x10) fprintf(stderr,"filesize: %lu, datasize: %lu\n", wavdata::filesize, wavdata::datasize);
		if (ac.info > 0x10) fprintf(stderr,"corrected RIFF: filesize: %d, datasize: %d\n", wavdata::wavtop.rlength, wavdata::datatop.len);
	}
	memcpy(&wavdata::wavtop.wavsn.cdata,"WAVE", 4);

	fwrite(&wavdata::wavtop, 1, sizeof wavdata::wavtop, ac.outfp);
	curlocation += sizeof wavdata::wavtop;

	if (wavdata::ds64ptr != nullptr)
	{
		wavdata::ds64location = curlocation;
		curlocation += wavdata::ds64len;
		fwrite((void *) wavdata::ds64ptr, 1, wavdata::ds64len + sizeof(tagsize), ac.outfp);
	}

	if (wavdata::bextptr != nullptr)
	{
		//buildnewbext();
		wavdata::bextlocation = curlocation;
		curlocation += wavdata::bextlen;
		fwrite((void *) wavdata::bextptr, 1, wavdata::bextlen + sizeof(tagsize), ac.outfp);
	}

	if (wavdata::factptr != nullptr)
	{
		wavdata::factlocation = curlocation;
		curlocation += wavdata::factlen;
		fwrite((void *) wavdata::factptr, 1, wavdata::factlen + sizeof(tagsize), ac.outfp);
		if (ac.info > 0x10) fprintf(stderr,"factval: %u\n", ((factraw *) wavdata::factptr)->factdata);
	}
	if (wavdata::fmtptr != nullptr)
	{
		wavdata::fmtlocation = curlocation;
		fmtstruct0 *fmt = (fmtstruct0 *)wavdata::fmtptr;
		if (ac.rafloatout)
			fmt->forcefloat();
#if 0
		int specfmtlen = sizeof (fmtstruct0);
		curlocation += specfmtlen;
		fwrite((void *) wavdata::fmtptr, 1, specfmtlen, ac.outfp);
#else
		curlocation += wavdata::fmtlen;
		fwrite((void *) wavdata::fmtptr, 1, wavdata::fmtlen + sizeof(tagsize), ac.outfp);
#endif
	}

	if (wavdata::chnaptr != nullptr)
	{
		wavdata::chnalocation = curlocation;
		curlocation += wavdata::chnalen;
		fwrite((void *) wavdata::chnaptr, 1, wavdata::chnalen + sizeof(tagsize), ac.outfp);
	}

	if (wavdata::axmlptr != nullptr)
	{
		wavdata::axmllocation = curlocation;
		curlocation += wavdata::axmllen;
		fwrite((void *) wavdata::axmlptr, 1, wavdata::axmllen + sizeof(tagsize), ac.outfp);
	}

	if (wavdata::iXMLptr != nullptr)
	{
		wavdata::iXMLlocation = curlocation;
		curlocation += wavdata::iXMLlen;
		fwrite((void *) wavdata::iXMLptr, 1, wavdata::iXMLlen + sizeof(tagsize), ac.outfp);
	}
	wavdata::datataglocation = curlocation;
	curlocation += sizeof wavdata::datatop;
	fwrite((void *) &wavdata::datatop, 1, sizeof wavdata::datatop, ac.outfp);

	wavdata::datalocation = curlocation;
	wavdata::newfilesize = curlocation;
	wavdata::newdatasize = 0;
}	

void
update1stpart(void)
{
	if (wavdata::ds64ptr == nullptr)
	{
		wavdata::wavtop.rlength = (wuint32_t) correctsizebps(wavdata::filesize);
		wavdata::datatop.len = (wuint32_t) correctsizebps(wavdata::datasize);
	}
	else
	{
		ds64raw* ds64 = (ds64raw *) wavdata::ds64ptr;

		wuint64_t tmp = correctsizebps(wavdata::newfilesize);
		ds64->bw64sizelow = tmp & 0xffffffffULL;
		ds64->bw64sizehigh = (tmp >> 32) & 0xffffffffULL;
		wavdata::filesize = tmp;

		tmp = correctsizebps(wavdata::newdatasize);
		ds64->datasizelow = tmp & 0xffffffffULL;
		ds64->datasizehigh = (tmp >> 32) & 0xffffffffULL;
		wavdata::datasize = tmp;

#if 0
		ds64->bw64sizelow = wavdata::newfilesize & 0xffffffffULL;
		ds64->bw64sizehigh = (wavdata::newfilesize >> 32) & 0xffffffffULL;
		wavdata::filesize = wavdata::newfilesize;

		ds64->datasizelow = wavdata::newdatasize & 0xffffffffULL;
		ds64->datasizehigh = (wavdata::newdatasize >> 32) & 0xffffffffULL;
		wavdata::datasize = wavdata::newdatasize;
#endif
	}
}

void
rewrite1stpart(void)
{
	fflush(ac.outfp);
	if (fseek(ac.outfp, SEEK_SET, 0ULL) == 0)
	{
		update1stpart();
		build1stpartofoutput();
	}
}

void*
THreadfile(const void * threadinfo)
{
	paudiochunk pa;

	const thread_init_list *actualthreadargs = (const thread_init_list *) threadinfo;

	setht();


	if ((ac.rafloatout == -1) || (ac.rafloatout == 0))
		ac.rafloatout = 0;

	if (wavdata::bextptr != nullptr)
		buildnewbext();
	else if (ac.bextorig != nullptr)
	{
		createbext();
		buildnewbext();
	}

	build1stpartofoutput();


	readaudioint24 *raint24 = nullptr;
	readaudiofloat *raf = nullptr;
	readaudio *ra = nullptr;

	if (ac.rafloat == 1)
	{
		raf = new readaudiofloat(ac.infp);
	}
	else if (ac.raint24 == 1)
	{
		raint24 = new readaudioint24(ac.infp);
	}
	else
	{
		ra = new readaudio(ac.infp);
	}

	Asample achunkbuf[AUDIOCHUNKSIZE];

	while (1)
	{
		int totalsamples = 0;

		pa = demptyqueue.THQwqdataread();

		audiopayload* ap = &pa->payload();

		if (raf != nullptr)
			totalsamples = raf->readASbuf(achunkbuf, AUDIOCHUNKSIZE);
		else if (raint24 != nullptr)
			totalsamples = raint24->readASbuf(achunkbuf, AUDIOCHUNKSIZE);
		else
			totalsamples = ra->readASbuf(achunkbuf, AUDIOCHUNKSIZE);

		if (totalsamples == 0 || totalsamples == -1)
		{
			dinput.THQwqspaceandeof(pa);
			break;
		}

		for (int i = 0; i < totalsamples; i++)
		{
			ap->samples[i]->sigin = achunkbuf[i];
			//ap->samples[i].sigin = achunkbuf[i];
		}

		ap->samplecount = totalsamples;
#if defined(DBGPRT)
		fprintf(stderr,"Queueing to input: %d samples\n", totalsamples);
#endif
		dinput.THQwqspaceandinsert(pa, false);

	}
	if (raf != nullptr)
		delete raf;
	if (ra != nullptr)
		delete ra;
	if (raint24 != nullptr)
		delete raint24;

	pthread_exit((void *) &actualthreadargs->EXITrtval);
	return nullptr;
}

void*
THwritefile(const void * threadinfo)
{
	const thread_init_list *actualthreadargs = (const thread_init_list *) threadinfo;
	setht();

	static size_t totalwritten;
	size_t byteswritten = 0;
	writeaudiofloat* wa = nullptr;

	physaudiosamplefloat writebuffer[AUDIOCHUNKSIZE];
	float floatwritebuffer[AUDIOCHUNKSIZE];
	int24 int24writebuffer[AUDIOCHUNKSIZE * 2];
	signed short shwritebuffer[AUDIOCHUNKSIZE * 2];
	paudiochunk pa;

//	if (ac.rafloatout)
//		wa = new writeaudiofloat(ac.outfp);

	while(1)
	{
	
		pa = doutput.THQwqdataread();
		if (pa->iseof())
		{
			demptyqueue.THQwqspaceandinsert(pa, true);
			break;
		}

		audiopayload* ap = &pa->payload();
		int samplecount = ap->samplecount;
#if defined(DBGPRT)
		fprintf(stderr,"%s got %d samples\n", actualthreadargs->threadname, samplecount);
#endif

		if (ac.nchannels == 1)
		{
			if (ac.raint24out)
			{
				for (int idx = 0; idx < int(samplecount); idx++)
				{
					const Asample& tmp = ap->samples[idx]->sigout;
					//const Asample& tmp = ap->samples[idx].sigout;
					float tmpval = tmp.item0();
					if (tmpval >= 0.999999f)
						tmpval = 0.999999f;
					if (tmpval <= -0.999999f)
						tmpval = -0.999999f;
					int24writebuffer[(idx*2)] = int24(tmpval);
				}
				size_t bytestowrite = samplecount * sizeof(int24);
				wavdata::newdatasize += bytestowrite;
				wavdata::newfilesize += bytestowrite;
				byteswritten += bytestowrite;
				totalwritten += samplecount;
		retry0s:
				size_t rt = fwrite(int24writebuffer, sizeof(int24),  int(samplecount), ac.outfp);
				if (rt == size_t(0))
				{
					if ((errno == EINTR) || (errno == EAGAIN))
					{
						usleep(1000);
						goto retry0s;
					}
				}
				
			}
			else if (ac.rafloatout)
			{
				for (int idx = 0; idx < int(samplecount); idx++)
				{
					const Asample& tmp = ap->samples[idx]->sigout;
					//const Asample& tmp = ap->samples[idx].sigout;
					floatwritebuffer[idx] = tmp.item0();
				}
				size_t bytestowrite = samplecount * sizeof(float);
				byteswritten += bytestowrite;
				totalwritten += samplecount;
				wavdata::newdatasize += bytestowrite;
				wavdata::newfilesize += bytestowrite;
		retryfps:
				size_t rt = fwrite(floatwritebuffer, sizeof(float),  int(samplecount), ac.outfp);
				if (rt == size_t(0))
				{
					if ((errno == EINTR) || (errno == EAGAIN))
					{
						usleep(1000);
						goto retryfps;
					}
				}
			}
			else
			{
				for (int idx = 0; idx < int(samplecount); idx++)
				{
					const Asample& tmpout = ap->samples[idx]->sigout;
					//const Asample& tmpout = ap->samples[idx].sigout;
					shwritebuffer[idx] = inltrunc(int((tmpout.item0() * 32767.0f)));
				}
				size_t bytestowrite = samplecount * sizeof(short);
				byteswritten += bytestowrite;
				totalwritten += samplecount;
				wavdata::newdatasize += bytestowrite;
				wavdata::newfilesize += bytestowrite;

		retry1s:
				size_t rt = fwrite(shwritebuffer, 1,  bytestowrite, ac.outfp);
				if (rt == size_t(0))
				{
					if ((errno == EINTR) || (errno == EAGAIN))
					{
						usleep(1000);
						goto retry1s;
					}
				}
			}
		}
		else
		{
			
			if (ac.raint24out)
			{
				for (int idx = 0; idx < int(samplecount); idx++)
				{
					const Asample& tmp = ap->samples[idx]->sigout;
					//const Asample& tmp = ap->samples[idx].sigout;
					float tmpval = tmp.item0();
					if (tmpval >= 0.999999f)
						tmpval = 0.999999f;
					if (tmpval <= -0.999999f)
						tmpval = -0.999999f;
					int24writebuffer[(idx*2)] = int24(tmpval);
					tmpval = tmp.item1();
					if (tmpval >= 0.999999f)
						tmpval = 0.999999f;
					if (tmpval <= -0.999999f)
						tmpval = -0.999999f;
					int24writebuffer[(idx*2)+1] = int24(tmpval);
				}
				size_t bytestowrite = 2 * samplecount * sizeof(int24);
				wavdata::newdatasize += bytestowrite;
				wavdata::newfilesize += bytestowrite;
				byteswritten += bytestowrite;
				totalwritten += samplecount;
		retry0:
				size_t rt = fwrite(int24writebuffer, 2*sizeof(int24),  int(samplecount), ac.outfp);
				if (rt == size_t(0))
				{
					if ((errno == EINTR) || (errno == EAGAIN))
					{
						usleep(1000);
						goto retry0;
					}
				}
				
			}
			else if (ac.rafloatout)
			{
				for (int idx = 0; idx < int(samplecount); idx++)
				{
					const Asample& tmp = ap->samples[idx]->sigout;
					//const Asample& tmp = ap->samples[idx].sigout;
					writebuffer[idx] = tmp;
				}
				size_t bytestowrite = samplecount * sizeof(physaudiosamplefloat);
				byteswritten += bytestowrite;
				totalwritten += samplecount;
				wavdata::newdatasize += bytestowrite;
				wavdata::newfilesize += bytestowrite;
		retryfp:
				size_t rt = fwrite(writebuffer, sizeof(physaudiosamplefloat),  int(samplecount), ac.outfp);
				if (rt == size_t(0))
				{
					if ((errno == EINTR) || (errno == EAGAIN))
					{
						usleep(1000);
						goto retryfp;
					}
				}
			}
			else
			{
				for (int idx = 0; idx < int(samplecount); idx++)
				{
					const Asample& tmpout = ap->samples[idx]->sigout;
					//const Asample& tmpout = ap->samples[idx].sigout;
					shwritebuffer[(2*idx)] = inltrunc(int((tmpout.item0() * 32767.0f)));
					shwritebuffer[(2*idx) + 1] = inltrunc(int((tmpout.item1() * 32767.0f)));
				}
				size_t bytestowrite = samplecount * sizeof(short) * 2;
				byteswritten += bytestowrite;
				totalwritten += samplecount;
				wavdata::newdatasize += bytestowrite;
				wavdata::newfilesize += bytestowrite;

		retry1:
				size_t rt = fwrite(shwritebuffer, 1,  bytestowrite, ac.outfp);
				if (rt == size_t(0))
				{
					if ((errno == EINTR) || (errno == EAGAIN))
					{
						usleep(1000);
						goto retry1;
					}
				}
			}
		}
		demptyqueue.THQwqspaceandinsert(pa, false);
	}

	rewrite1stpart();

	//
	// Close the output file if it was opened by us.
	//
#if 0
	if (ac.rafloatout)
	{
		delete wa;
	}
#endif


	initwavdata();

	pthread_exit((void *) &actualthreadargs->EXITrtval);

	return nullptr;
}

void NOINLINE
ttyerror(UNUSED int errcode)
{
#if defined(DUMPALLINFO)
	static const int maxembufsize=512;
	char msgfromlsterr[maxembufsize];
	const char *msgfromlst = lstrerror_r(errcode, msgfromlsterr, maxembufsize);
	::fprintf(stderr,"tty error: \"%s\", command processor is disabled", msgfromlst);
#endif
}

#if !defined(WINDOWS64)
static bool reqloopexit = false;
static volatile int reqwinch = false;

static volatile int ttyrows, ttycolumns;

void
setwinch(void)
{
	__atomic_store_n(&reqwinch, true, __ATOMIC_SEQ_CST);
	__atomic_thread_fence(__ATOMIC_SEQ_CST);
	__atomic_signal_fence(__ATOMIC_SEQ_CST);
}

int
getresetwinch(void)
{
	__atomic_thread_fence(__ATOMIC_SEQ_CST);
	__atomic_signal_fence(__ATOMIC_SEQ_CST);
	int oldval = __atomic_exchange_n(&reqwinch, false, __ATOMIC_SEQ_CST);
	return oldval;
}

static int
getttysizes(void)
{
	struct winsize ws = {0,0,0,0};
	if (!ioctl(2, TIOCGWINSZ, &ws))
	{
		__atomic_store_n(&ttyrows, ws.ws_row, __ATOMIC_SEQ_CST);
		__atomic_store_n(&ttycolumns, ws.ws_col, __ATOMIC_SEQ_CST);
	}
	__atomic_thread_fence(__ATOMIC_SEQ_CST);
	return __atomic_load_n(&ttycolumns, __ATOMIC_SEQ_CST);
}

int
updatettycolumns(void)
{
	int col;
	if(getresetwinch())
	{
		col = getttysizes();
		return col;
	}
	return ttycolumns;
}

void
forcettyupdate(void)
{
	setwinch();
}

static void
sigrtn(UNUSED int signo, UNUSED siginfo_t* si, UNUSED void *addlinfo)
{
	reqloopexit = true;
}

static void
sigwinch(UNUSED int signo, UNUSED siginfo_t* si, UNUSED void *addlinfo)
{
	forcettyupdate();
}

void
setsignal(void)
{
	struct sigaction newsa{};

	newsa.sa_sigaction = sigwinch;
	newsa.sa_flags = SA_SIGINFO;
	sigaction(SIGWINCH, &newsa, nullptr);
}
#else
void
setsignal(void)
{
}
void
forcettyupdate(void)
{
}
#endif


//
// vi: set ts=4:
//

