//
//
//
// Copyright(C) 2017 John S Dyson
// All rights reserved.
//
// This source code is an unpublished proprietary document, and is available only
// under a written license.
// 
#if !defined(__SSMOOTH_H__)
#define __SSMOOTH_H__ 1

#include <sys/types.h>
#include "varray.h"


#if defined(__clang__)
#define FINLINE0  // inline
#else
#define FINLINE0  // __attribute__((noinline))
#endif

#if defined(__clang__)
#define FINLINE1
#else
#define FINLINE1  // __attribute__((noinline))
#endif

#define FINLINE2  // inline

#if !defined(ALIGNED)
#define ALIGNED __attribute__((aligned(32)))
#endif

static const float SMALLVAL(-1.0f * __builtin_huge_valf());
static const float LARGEVAL(__builtin_huge_valf());

#if 0
template <typename VDEF>
inline varrayT<VDEF>
sratetmod(varrayT<VDEF> tval)
{
	return Vselect((tval == varrayT<VDEF>(1.0)), varrayT<VDEF>(1.0), tval);
}

template <typename T>
inline T
sratetmod(T tval)
{
	return (tval == T(1.0)) ? T(1.0) : (tval * T(aspec.SRATETMULT));
}
#endif


namespace agcfilter
{
	static constexpr int nstepspwr = 3;
	static constexpr int nstepsgrain = (1 << nstepspwr);
	template <typename T>
		static const T ones = T(1.0f);
	template <typename T>
		static const T zeros = T(1.0f);
};

using namespace agcfilter;

template <typename DT = double, int NSTEPS = 16>
class ALIGNED atkdcys
{
	DT ALIGNED atks[NSTEPS];
	DT ALIGNED atksmod[NSTEPS];
	DT ALIGNED aratios[NSTEPS];
	DT ALIGNED dcys[NSTEPS];
	DT ALIGNED dcysmod[NSTEPS];
	DT atkst, atkend, atkratio;
	DT dcyst, dcyend, dcyratio;
	float tscale;
	
public:
	atkdcys(float Atscale, DT Aatkst, DT Aatkend, DT Adcyst, DT Adcyend) FINLINE2 :
		atks(), atksmod(), dcys(), dcysmod(),
		atkst(Aatkst), atkend(Aatkend), atkratio(Aatkend/Aatkst),
		dcyst(Adcyst), dcyend(Adcyend), dcyratio(Adcyend/Adcyst),
		tscale(Atscale)
	{

		tscale = (Atscale < 1.0f) ? 1.0f : Atscale;

		atkst = Aatkst * tscale;
		atkst = Vmin(atkst, DT(1.0));

		atkend = Aatkend * tscale;
		atkend = Vmin(atkend, DT(1.0));

		dcyst = Adcyst * tscale;
		dcyst = Vmin(dcyst, DT(1.0));
	
		dcyend = Adcyend * tscale;
		dcyend = Vmin(dcyend, DT(1.0));

		atkratio = pow(DT(atkend / atkst), (DT(1.0) / DT(NSTEPS-1)));
		dcyratio = pow(DT(dcyend / dcyst), (DT(1.0) / DT(NSTEPS-1)));

		DT atktmp{atkst};
		DT dcytmp{dcyst};

		for (int i = 0; i < NSTEPS; i++)
		{
			DT r = atkend / atktmp;
			aratios[i] = r;
			atks[i] = atktmp;
			atksmod[i] = atktmp / (DT(1.0) + atktmp);
			atktmp *= atkratio;
			dcys[i] = dcytmp;
			dcysmod[i] = dcytmp / (DT(1.0) + dcytmp);
			dcytmp *= dcyratio;
		}
	}

	atkdcys(DT Aatkst, DT Aatkend, DT Adcyst, DT Adcyend) FINLINE2 :
		atkdcys(1.0f, Aatkst, Aatkend, Adcyst, Adcyend) {}

	atkdcys(const atkdcys& init) FINLINE2:
		atks(), atksmod(), aratios(), dcys(), dcysmod(),
		atkst(init.atkst), atkend(init.atkend), atkratio(init.atkratio),
		dcyst(init.dcyst), dcyend(init.dcyend), dcyratio(init.dcyratio),
		tscale(init.tscale)
	{

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = init.atks[i];
			dcys[i] = init.dcys[i];
			aratios[i] = init.aratios[i];
		}
	}

	atkdcys& FINLINE2
	operator=(const atkdcys& src)
	{
		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = src.atks[i];
			atksmod[i] = src.atksmod[i];
			dcys[i] = src.dcys[i];
			dcysmod[i] = src.dcysmod[i];
			aratios[i] = src.aratios[i];
		}
		atkst = src.atkst;
		atkend = src.atkend;
		atkratio = src.atkratio;
		dcyst = src.dcyst;
		dcyend = src.dcyend;
		dcyratio = src.dcyratio;
		tscale = src.tscale;
		
		return *this;
	}

	void
	setatk(int idx, DT val)
	{
		atks[idx] = val;
		aratios[idx] = atkend / val;
	}

	void
	setdcy(int idx, DT val)
	{
		dcys[idx] = val;
	}

	const DT&
	atk( int idx) const
	{
		return atks[idx];
	}

	const DT&
	atkmod( int idx) const
	{
		return atksmod[idx];
	}

	const DT&
	aratio( int idx) const
	{
		return aratios[idx];
	}

	const DT&
	dcy( int idx) const
	{
		return dcys[idx];
	}

	const DT&
	dcymod(int idx) const
	{
		return dcysmod[idx];
	}

	void FINLINE2
	retuneatks(DT fromatk, DT toatk)
	{
		DT newatkratio = pow(DT(toatk / fromatk), DT(1.0/(NSTEPS - 1)));
		DT atktmp{fromatk};
		atkst = fromatk;
		atkend = toatk;

		for (int i = 0; i < NSTEPS; i++)
		{
			atks[i] = atktmp;
			atksmod[i] = atktmp / (DT(1.0) + atktmp);
			aratios[i] = toatk / atktmp;
			atktmp *= newatkratio;
		}
	}

	void FINLINE2
	retunedcys(DT fromdcy, DT todcy)
	{
		DT newdcyratio = pow(DT(todcy / fromdcy), DT(1.0/(NSTEPS - 1)));
		DT dcytmp{fromdcy};

		for (int i = 0; i < NSTEPS; i++)
		{
			dcys[i] = dcytmp;
			dcysmod[i] = dcytmp / (DT(1.0) + dcytmp);
			dcytmp *= newdcyratio;
		}
	}

};

#if 0
template <typename DT = float, int NSTEPS=16, int NF=1, bool NEG = false>
class ALIGNED Sfilter
{
	atkdcys<DT,NSTEPS> ALIGNED ad;
	DT ALIGNED curval[NF][NSTEPS];
	DT totalv[NF];

	Sfilter(const Sfilter&) = delete;
	Sfilter& operator=(const Sfilter&) = delete;

public:
	typedef DT bsa;
	typedef int bsai;
	static constexpr int nsteps = NSTEPS;

	Sfilter(int UNUSED scalet,  DT atkst, DT atkend,
		DT dcyst, DT dcyend, DT initval) FINLINE2 :
		ad(atkst, atkend, dcyst, dcyend), curval(),
		totalv()
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < nsteps; j++)
				curval[i][j] = initval;
			totalv[i] = initval;
		}
	}

	Sfilter(DT Aatkst, DT Aatkend,
		DT Adcyst, DT Adcyend, DT initval) FINLINE0 :
		Sfilter(0, Aatkst, Aatkend, Adcyst, Adcyend, initval) {}

	~Sfilter()
	{
	}

	DT FINLINE1
	next(DT newval)
	{
		DT rtval;
		if (NEG)
		{
			rtval = +1.00e20f;
		}
		else
		{
			rtval = -1.00e20f;
		}
		DT GTnewval{newval};
		

		for (int i = 0; i < NSTEPS; i++)
		{
			const DT& atk(ad.atk(i));
			const DT& dcy(ad.dcy(i));

			const DT& Tnewval = GTnewval;
			DT& Tcv = curval[0][i];

			bool atkflags, dcyflags;
			if (NEG)
			{
				atkflags = (Tnewval < Tcv);
				dcyflags = (Tnewval > Tcv);
			}
			else
			{
				atkflags = (Tnewval > Tcv);
				dcyflags = (Tnewval < Tcv);
			}

			DT atktmp = atk;

			DT atkamt = (Tcv * (DT(1.0) - atktmp)) + ( Tnewval * atktmp);
			DT dcyamt = (Tcv * (DT(1.0) - dcy)) + ( Tnewval * dcy);

			Tcv = dcyflags ? dcyamt : (atkflags ? atkamt : Tcv);
			if (NEG)
			{
				rtval = (Tcv < rtval) ? Tcv : rtval;
			}
			else
			{
				rtval = (Tcv > rtval) ? Tcv : rtval;
			}
		}
		totalv[0] = rtval;

		for (int j = 1; j < NF; j++)
		{
			for (int i = 0; i < NSTEPS; i++)
			{
				const DT& atk(ad.atk(i));
				const DT& dcy(ad.dcy(i));

				const DT& Tnewval = curval[j-1][i];
				DT& Tcv = curval[j][i];

				bool atkflags, dcyflags;
				if (NEG)
				{
					atkflags = (Tnewval < Tcv);
					dcyflags = (Tnewval > Tcv);
				}
				else
				{
					atkflags = (Tnewval > Tcv);
					dcyflags = (Tnewval < Tcv);
				}

				DT atktmp = atk;
				DT atkamt = (Tcv * (DT(1.0) - atktmp)) + ( Tnewval * atktmp);
				DT dcyamt = (Tcv * (DT(1.0) - dcy)) + ( Tnewval * dcy);

#if 0
				Tcv = dcyflags ? dcyamt : (atkflags ? Tnewval : Tcv);
#else
				Tcv = dcyflags ? dcyamt : (atkflags ? atkamt : Tcv);
#endif
				if (NEG)
				{
					rtval = (Tcv < rtval) ? Tcv : rtval;
				}
				else
				{
					rtval = (Tcv > rtval) ? Tcv : rtval;
				}
			}
			totalv[j] = rtval;
		}

		rtval = totalv[NF-1];
		if (NEG)
		{
			for (int i = 0; i < NF; i++)
			{
				rtval = Vmin(rtval, totalv[i]);
			}
		}
		else
		{
			for (int i = 0; i < NF; i++)
			{
				rtval = Vmax(rtval, totalv[i] * DT(0.707));
			}
		}
			
		return rtval;
	}

	DT
	getcurval(void) const
	{
		return totalv;
	}


	template <typename PV>
	PV operator()(PV input)
	{
		return next(input);
	}
};

template <typename DT = float, int NSTEPS=16, int NF=1, bool NEG = false>
class ALIGNED SfilterS
{
	atkdcys<DT,NSTEPS> ALIGNED ad;
	DT ALIGNED curval[NF][NSTEPS];
	DT ALIGNED zval[NF][NSTEPS];
	DT totalv[NF];

	SfilterS(const SfilterS&) = delete;
	SfilterS& operator=(const SfilterS&) = delete;

public:
	typedef DT bsa;
	typedef int bsai;
	static constexpr int nsteps = NSTEPS;

	SfilterS(int UNUSED scalet,  DT atkst, DT atkend,
		DT dcyst, DT dcyend, DT initval) FINLINE2 :
		ad(atkst, atkend, dcyst, dcyend), curval(), zval(),
		totalv()
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < nsteps; j++)
			{
				curval[i][j] = initval;
				zval[i][j] = initval;
			}
			totalv[i] = initval;
		}
	}

	SfilterS(DT Aatkst, DT Aatkend,
		DT Adcyst, DT Adcyend, DT initval) FINLINE0 :
		SfilterS(0, Aatkst, Aatkend, Adcyst, Adcyend, initval) {}

	~SfilterS()
	{
	}

	DT FINLINE1
	next(DT newval)
	{
		DT rtval;
		if (NEG)
		{
			rtval = +1.00e20f;
		}
		else
		{
			rtval = -1.00e20f;
		}
		DT GTnewval{newval};
		

		for (int j = 0; j < NF; j++)
		{
			for (int i = 0; i < NSTEPS; i++)
			{
				const DT& atkmod(ad.atkmod(i));
				const DT& dcymod(ad.dcymod(i));

				const DT& Tnewval = (j == 0) ? GTnewval : curval[j-1][i];
				DT& Tcv = curval[j][i];
				DT& Tzv = zval[j][i];

				DT vdiff = Tnewval - Tzv;

				DT tmp = (Tnewval > Tcv) ? vdiff * atkmod : vdiff * dcymod;
				DT rtvalXtmp = Tzv + tmp;
				Tzv = rtvalXtmp + tmp;
				Tcv = rtvalXtmp;

				DT nrt = rtvalXtmp * ad.aratio(i);
				if (NEG)
				{
					rtval = (nrt < rtval) ? nrt : rtval;
				}
				else
				{
					rtval = (nrt > rtval) ? nrt: rtval;
				}
			}
			totalv[j] = rtval;
		}

		rtval = totalv[NF-1];
		if (NEG)
		{
			for (int i = 0; i < NF; i++)
			{
				rtval = Vmin(rtval, totalv[i]);
			}
		}
		else
		{
			for (int i = 0; i < NF; i++)
			{
				rtval = Vmax(rtval, totalv[i]);
			}
		}
			
		return rtval;
	}

	DT
	getcurval(void) const
	{
		return totalv;
	}


	template <typename PV>
	PV operator()(PV input)
	{
		return next(input);
	}
};

template <typename ST = double, int NSTEPS=16, int NF=1>
class ALIGNED SfilterdB
{
	typedef int IT;
	//
	// nstepsgrain must be a power of two
	//
	atkdcys<ST,NSTEPS> ALIGNED ad;
	ST ALIGNED curval[NF][NSTEPS];
	ST ALIGNED totalv[NF];
	ST maxatkspeed, maxdcyspeed;
	ST samplerate;
	ST initval;

	SfilterdB(const SfilterdB&) = delete;
	SfilterdB& operator=(const SfilterdB&) = delete;

public:
	typedef ST bsa;
	typedef int bsai;

	SfilterdB(ST Asamplerate,
		ST atkst, ST atkend, ST dcyst, ST dcyend, ST Ainitval=float(0.0f)) FINLINE0 :
		ad(atkst/Asamplerate,atkend/Asamplerate,dcyst/Asamplerate,dcyend/Asamplerate),
		curval(), totalv(),
		maxatkspeed(aspec.inSRATE), maxdcyspeed(aspec.inSRATE),
		samplerate(Asamplerate),
		initval(Ainitval)
	{

		atkst = (atkst == ST{bsa(0.0)}) ? ST(Asamplerate) : atkst;
		atkend = (atkend == ST{bsa(0.0)}) ? ST(Asamplerate) : atkend;

		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = initval;
			}
			totalv[i] = initval;
		}

	}

	~SfilterdB() 
	{
	}

	ST
	getcurval(void) const
	{
		return totalv;
	}

	ST FINLINE1
	next(ST Anewval, bool hold = false)
	{
		ST newval(Anewval);
	
		bool dodecay = !hold && (newval < totalv[NF - 1]);
		bool doatack = (newval > totalv[NF - 1]);

		ST rtval(-320.0f);

		if (doatack)
		{
			const ST& Tnewval = newval;

			for (int i = 0; i < NSTEPS; i++)
			{
				ST& Tcv = curval[0][i];
				ST rqatack = Tcv + ad.atk(i);
				Tcv = (Tnewval > Tcv) ? ((rqatack < Tnewval) ? rqatack : Tnewval) : Tcv;
				rtval = rtval > Tcv ? rtval : Tcv;
			}
			totalv[0] = rtval;
		}

		if (dodecay)
		{
			const ST& Tnewval = newval;

			for (int i = 0; i < NSTEPS; i++)
			{
				ST& Tcv = curval[0][i];
				ST rqdecay = Tcv - ad.dcy(i);
				Tcv = (Tnewval < Tcv) ? ((rqdecay > Tnewval) ? rqdecay : Tnewval) : Tcv;
				rtval = rtval > Tcv ? rtval : Tcv;
			}
			totalv[0] = rtval;
		}


		for (int j = 1; j < NF; j++)
		{
			ST Irtval(-320.0f);

			if (doatack)
			{
				for (int i = 0; i < NSTEPS; i++)
				{
					const ST& Tnewval = curval[j-1][i];
					ST& Tcv = curval[j][i];
#if 0
					Tcv = (Tnewval > Tcv) ? Tnewval : Tcv;
#else
					ST rqatack = Tcv + ad.atk(i);
					Tcv = (Tnewval > Tcv) ? ((rqatack < Tnewval) ? rqatack : Tnewval) : Tcv;
#endif
					Irtval = lm::Mmax(Irtval, Tcv);
				}
				totalv[j] = Irtval;
			}

			if (dodecay)
			{
				for (int i = 0; i < NSTEPS; i++)
				{
					const ST& Tnewval = curval[j-1][i];
					ST& Tcv = curval[j][i];
					ST rqdecay = Tcv - ad.dcy(i);
					Tcv = (Tnewval < Tcv) ? ((rqdecay > Tnewval) ? rqdecay : Tnewval) : Tcv;
					Irtval = lm::Mmax(Irtval, Tcv);
				}
				totalv[j] = Irtval;
			}
		}
		rtval = totalv[NF-1];
		for (int i = 0; i < NF; ++i)
			rtval = Vmax(rtval - ST(3.0f), totalv[i]);
		
		return rtval;

		//return totalv[NF-1];
	}

	ST FINLINE1
	nextAccel(ST Anewval, bool hold = false)
	{
		ST newval(Anewval);
	
		bool dodecay = !hold && (newval < totalv[NF - 1]);
		bool doatack = (newval > totalv[NF - 1]);

		ST rtval(-320.0f);

		if (doatack)
		{
			const ST& Tnewval = newval;

			for (int i = 0; i < NSTEPS; i++)
			{
				ST& Tcv = curval[0][i];
				ST rqatack = Tcv + (ad.atk(i) * (Tnewval - Tcv));
				Tcv = (Tnewval > Tcv) ? ((rqatack < Tnewval) ? rqatack : Tnewval) : Tcv;
				rtval = rtval > Tcv ? rtval : Tcv;
			}
			totalv[0] = rtval;
		}

		if (dodecay)
		{
			const ST& Tnewval = newval;

			for (int i = 0; i < NSTEPS; i++)
			{
				ST& Tcv = curval[0][i];
				ST rqdecay = Tcv - ad.dcy(i);
				Tcv = (Tnewval < Tcv) ? ((rqdecay > Tnewval) ? rqdecay : Tnewval) : Tcv;
				rtval = rtval > Tcv ? rtval : Tcv;
			}
			totalv[0] = rtval;
		}


		for (int j = 1; j < NF; j++)
		{
			ST Irtval(-320.0f);

			if (doatack)
			{
				for (int i = 0; i < NSTEPS; i++)
				{
					const ST& Tnewval = curval[j-1][i];
					ST& Tcv = curval[j][i];
#if 0
					Tcv = (Tnewval > Tcv) ? Tnewval : Tcv;
#else
					ST rqatack = Tcv + (ad.atk(i) * (Tnewval - Tcv));
					Tcv = (Tnewval > Tcv) ? ((rqatack < Tnewval) ? rqatack : Tnewval) : Tcv;
#endif
					Irtval = lm::Mmax(Irtval, Tcv);
				}
				totalv[j] = Irtval;
			}

			if (dodecay)
			{
				for (int i = 0; i < NSTEPS; i++)
				{
					const ST& Tnewval = curval[j-1][i];
					ST& Tcv = curval[j][i];
					ST rqdecay = Tcv - ad.dcy(i);
					Tcv = (Tnewval < Tcv) ? ((rqdecay > Tnewval) ? rqdecay : Tnewval) : Tcv;
					Irtval = lm::Mmax(Irtval, Tcv);
				}
				totalv[j] = Irtval;
			}
		}

		rtval = totalv[NF-1];
		for (int i = 0; i < NF; ++i)
			rtval = Vmax(rtval - ST(3.0f), totalv[i]);
		
		return rtval;

//		return totalv[NF-1];
	}

	ST FINLINE0
	smooth(ST in, int hold = false)
	{
		next(in, hold);
		if (NF == 1)
			return totalv[NF - 1];
		float total{0.0f};

		varrayf8 Vtotalv(0.0f), Vtotal;

		for (int i = 0; i < NF; i++)
			Vtotalv[i] = totalv[i];

		Vtotal = dB2lin(Vtotalv);

		for (int i = 0; i < NF; i++)
			total += Vtotal[i];
		
		total *= 1.0f/NF;
		
		float rtdb(lin2dB((float) total));
		return rtdb;
	}

	template <typename PV>
	PV operator()(PV input, int hold = false)
	{
		return next(input, hold);
	}

};

template <typename VDEF, int NSTEPS=16, int NF=1, bool NEG=false>
class ALIGNED Vfilter
{
	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> VIT;
	typedef typename VDEF::FT FT;

	atkdcys<VT,NSTEPS> ALIGNED ad;

	VT ALIGNED curval[NF][NSTEPS];
	VT ALIGNED totalv[NF];
	VT initval;

	Vfilter& operator=(const Vfilter&) = delete;

public:
	typedef VT bsa;
	typedef VIT bsai;
	static constexpr int VS = VDEF::VS;

	Vfilter(VT atkst,VT atkend,
		VT dcyst, VT dcyend, VT Ainitval=FT(0.0f)) FINLINE0 :
		ad(atkst, atkend, dcyst, dcyend),
		curval(), totalv(), initval(Ainitval)
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
			}
			totalv[i] = VT(initval);
		}
	}

	Vfilter(FT atkst, FT atkend,
		FT dcyst, FT dcyend, FT Ainitval=FT(0.0f)) FINLINE0 :
		ad(VT(atkst), VT(atkend), VT(dcyst), VT(dcyend)),
		curval(), totalv(), initval(Ainitval)
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
			}
			totalv[i] = VT(initval);
		}
	}

	Vfilter(const Vfilter<VDEF,NSTEPS,NF>& init) FINLINE0 :
		ad(init.ad), curval(), totalv(), initval(init.initval)
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = init.curval[i][j];
			}
			totalv[i] = init.totalv[i];
		}
	}

	~Vfilter()
	{
	}

	VT FINLINE1
	next(const VT& newval, bool atkdecayall = false)
	{
		for (int j = 0; j < NF; j++)
		{
			VT rtvalX(0.0f);

			if (NEG)
				rtvalX = VT(LARGEVAL);
			else
				rtvalX = VT(SMALLVAL);

			bool atkdecay = (j == 0) || atkdecayall;
			for (int i = 0; i < NSTEPS; i++)
			{
				const VT& atk = ad.atk(i);
				const VT& dcy = ad.dcy(i);
				VT& Tcv = curval[j][i];

				const VT& Tnewval { (j != 0) ? curval[j-1][i] : newval };

				VT atkflags, dcyflags;

				if (NEG)
				{
					atkflags = Vlt(Tnewval, Tcv);
					dcyflags = Vgt(Tnewval, Tcv);
				}
				else
				{
					atkflags = Vgt(Tnewval, Tcv);
					dcyflags = Vlt(Tnewval, Tcv);
				}

				VT atkamt = atkdecay ? (Tcv * (VT(1.0f) - atk)) + (Tnewval * atk) : Tnewval;
				VT nTcv = Vselect(atkflags, atkamt, Tcv);
				VT dcyamt = (Tcv * (VT(1.0f) - dcy)) + (Tnewval * dcy);
				nTcv = Vselect(dcyflags, dcyamt, nTcv);

				if (NEG)
					rtvalX = Vmin(rtvalX, nTcv);
				else
					rtvalX = Vmax(rtvalX, nTcv);

				//curval[j][i] = nTcv;
				Tcv = nTcv;
			}
			totalv[j] = rtvalX;
		}

		VT rtval = totalv[NF-1];
		if (NEG)
		{
			for (int i = 0; i < NF; i++)
			{
				rtval = Vmin(rtval, totalv[i]);
			}
		}
		else
		{
			for (int i = 0; i < NF; i++)
			{
				rtval = Vmax(rtval, totalv[i]);
			}
		}
			
		return rtval;
	}

	VT FINLINE0
	smooth(const VT& newval, bool attackdecayall = false)
	{
		next(newval, attackdecayall);
		if (NF == 1)
			return totalv[0];
		VT total(0.0f);

		for (int i = 0; i < NF; i++)
			total += totalv[i];

		total *= VT(1.0f / NF);

		return total;
	}

	void FINLINE1
	setcurvals(const VT& initv)
	{
		for (int j = 0; j < NF; j++)
		{
			VT *cv = curval[j];
			for (int i = 0; i < NSTEPS; i++)
			{
				cv[i] = initv;
			}
			totalv[j] = initv;
		}
	}

	void
	setcurvals(FT newval)
	{
		VT nv(0.0f);
		for (int i = 0; i < VDEF::VS; i++)
			nv[i] = newval[i];
		setcurvals(nv);
		return;
	}

	VT
	getcurvals(void) const
	{
		return totalv;
	}

	void
	setatk(int idx, VT newval)
	{
		ad.setatk(idx, newval);
	}

	template <typename PV> FINLINE2
	PV operator()(const PV& input)
	{
		return next(input);
	}

	template <typename PV> FINLINE2
	PV operator()(const PV& input, bool atkall)
	{
		return next(input, atkall);
	}

	template <typename PV, typename IV> FINLINE2
	PV operator()(const PV& input, const IV& aa, const IV& adX)
	{
		return next(input, aa, adX);
	}

	template <typename PV, typename IV> FINLINE2
	PV operator()(const PV& input, const IV& aa, const IV& adX, bool atkall)
	{
		return next(input, aa, adX, atkall);
	}
};
#endif

template <typename VDEF = TYPEvf8x1, int NSTEPS = 16, int NF = 1>
class ALIGNED VfilterdB
{
	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> IVT;
	typedef typename VDEF::FT FT;

	ALIGNED atkdcys<VT,NSTEPS> ad;

	ALIGNED VT curval[NF][NSTEPS];
	VT totalv[NF];
	FT samplerate;
	FT initval;
	int scalecnt;
	VT smoothval;

	VfilterdB& operator=(const VfilterdB&) = delete;

public:
	typedef VT bsa;
	typedef IVT bsai;
	static constexpr int nsteps = NSTEPS;
	static constexpr int VS = VDEF::VS;

	VfilterdB(FT Asamplerate, VT atkst, VT atkend,
		VT dcyst, VT dcyend, FT Ainitval=FT(0)) FINLINE2 :
		ad(0, atkst/VT(Asamplerate), atkend/VT(Asamplerate), dcyst/VT(Asamplerate), dcyend/VT(Asamplerate)),
		curval(), totalv(),
		samplerate(Asamplerate), initval(Ainitval),
		scalecnt(0), smoothval()
	{
		VT zval(0.0f);

		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
			}

			totalv[i] = VT(initval);
		}

	}

	VfilterdB(FT Asamplerate, FT atkst, FT atkend,
		FT dcyst, FT dcyend, FT Ainitval=FT(0)) FINLINE2 :
		ad(0, VT(atkst/Asamplerate), VT(atkend/Asamplerate), VT(dcyst/Asamplerate), VT(dcyend/Asamplerate)),
		curval(), totalv(),
		samplerate(Asamplerate), initval(Ainitval),
		scalecnt(0), smoothval()
	{
		VT zval(0.0f);

		atkst = (atkst == FT(0.0)) ? samplerate : atkst;
		atkend = (atkend == FT(0.0)) ? samplerate : atkend;

		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
			}

			totalv[i] = VT(initval);
		}

	}

	VfilterdB(const VfilterdB<VDEF,NSTEPS,NF>& init) FINLINE2 :
		curval(), ad(init.ad), totalv(),
		samplerate(init.samplerate), initval(init.initval),
		scalecnt(init.scalecnt), smoothval(init.smoothval)
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = init.curval[i][j];
			}

			totalv[i] = init.totalv[i];
		}
	}

	~VfilterdB() 
	{
	}

	void
	retuneatks(float fromatk, float toatk)
	{
		ad.retuneatks(fromatk, toatk);
	}

	void
	retunedcys(float fromdcy, float todcy)
	{
		ad.retunedcys(fromdcy, todcy);
	}

	VT FINLINE2
	next(const VT& Anewval, IVT Afreeze)
	{
		VT XTnewval(Anewval);

		IVT notfreeze(~Afreeze);
		VT allowdcy(notfreeze.raw());
		VT allowatk(IVT(-1).raw());

		VT ErtvalX(-320.0);

		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tcv = curval[0][i];

			VT atkflags = Vgt(XTnewval, Tcv) & allowatk;
			VT dcyflags = Vlt(XTnewval, Tcv) & allowdcy;

			Tcv = Vselect(atkflags,
				Vmin((Tcv + VT(ad.atk(i))), XTnewval),
					Vselect(dcyflags,
						Vmax((Tcv - VT(ad.dcy(i))), XTnewval), Tcv) );

			ErtvalX = Vmax(ErtvalX, Tcv);
		}

		totalv[0] = ErtvalX;


		for (int j = 1; j < NF; j++)
		{
			VT rtvalX(-320.0);

			for (int i = 0; i < NSTEPS; i++)
			{
				VT& Tcv( curval[j][i]);

				const VT& Tnewval( curval[j-1][i] );

				VT atkflags = Vgt(Tnewval, Tcv) & allowatk;
				VT dcyflags = Vlt(Tnewval, Tcv) & allowdcy;

				Tcv = Vselect(atkflags, Tnewval,
					Vselect(dcyflags,
						Vmax((Tcv - VT(ad.dcy(i))), Tnewval), Tcv) );

				rtvalX = Vmax(rtvalX, Tcv);
			}

			totalv[j] = rtvalX;
		}

		return totalv[NF-1];
	}

	VT FINLINE2
	next(const VT& Anewval)
	{
		VT XTnewval(Anewval);

		VT ErtvalX(-320.0);

		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tcv = curval[0][i];

			VT atkflags = Vgt(XTnewval, Tcv);
			VT dcyflags = Vlt(XTnewval, Tcv);

			Tcv = Vselect(atkflags,
				Vmin((Tcv + VT(ad.atk(i))), XTnewval),
					Vselect(dcyflags,
						Vmax((Tcv - VT(ad.dcy(i))), XTnewval), Tcv) );

			ErtvalX = Vmax(ErtvalX, Tcv);
		}

		totalv[0] = ErtvalX;

		for (int j = 1; j < NF; j++)
		{
			VT rtvalX(-320.0);

			for (int i = 0; i < NSTEPS; i++)
			{
				VT& Tcv( curval[j][i]);

				const VT& Tnewval( curval[j-1][i] );

				VT atkflags = Vgt(Tnewval, Tcv);
				VT dcyflags = Vlt(Tnewval, Tcv);

				Tcv = Vselect(atkflags, Tnewval,
					Vselect(dcyflags,
						Vmax((Tcv - VT(ad.dcy(i))), Tnewval), Tcv) );

#if 0
				Tcv = Vselect(atkflags,
					Vmin((Tcv + VT(ad.atk(i))), Tnewval),
					Vselect(dcyflags,
						Vmax((Tcv - VT(ad.dcy(i))), Tnewval), Tcv) );
#endif

				rtvalX = Vmax(rtvalX, Tcv);
			}

			totalv[j] = rtvalX;
		}

		VT rtval(totalv[NF-1]);
		for (int i = 0; i < NF; ++i)
			rtval = Vmax(rtval, totalv[i]);
		
		return rtval;

//		return totalv[NF-1];
	}

	VT FINLINE2
	nextAccel(const VT& Anewval)
	{
		VT XTnewval(Anewval);

		VT ErtvalX(-320.0);

		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tcv = curval[0][i];

			VT attackdiff = (XTnewval - Tcv) * VT(2.0);
			VT attackdiffpos = Vmax(attackdiff, VT(0.0f));
			VT attack = attackdiffpos;

			VT attackincr = ad.atk(i) * attack * ad.aratio(i);
			attackincr = Vmax(attackincr, VT(0.0f));
			VT attackTcv = Tcv + attackincr;

			Tcv = Vselect(Vgt(attackdiff, VT(0.0f)),
				Vmin(attackTcv, XTnewval),
				Vmax((Tcv - VT(ad.dcy(i))), XTnewval));

			ErtvalX = Vmax(ErtvalX, Tcv);
		}

		totalv[0] = ErtvalX;

		for (int j = 1; j < NF; j++)
		{
			VT rtvalX(-320.0);

			for (int i = 0; i < NSTEPS; i++)
			{
				VT& Tcv( curval[j][i]);

				const VT& Tnewval( curval[j-1][i] );

				VT atkflags = Vgt(Tnewval, Tcv);

				VT attackdiff = (Tnewval - Tcv) * VT(2.0);
				VT attackdiffpos = Vmax(attackdiff, VT(0.0f));
				VT attack = attackdiffpos;

				VT attackincr = ad.atk(i) * attack * ad.aratio(i);
				attackincr = Vmax(attackincr, VT(0.0f));
				VT attackTcv = Tcv + attackincr;

				Tcv = Vselect(Vgt(attackdiff, VT(0.0f)),
					Vmin(attackTcv, Tnewval),
					Vmax((Tcv - VT(ad.dcy(i))), Tnewval));

				rtvalX = Vmax(rtvalX, Tcv);
			}

			totalv[j] = rtvalX;
		}

		VT rtval(totalv[NF-1]);
		for (int i = 0; i < NF; ++i)
			rtval = Vmax(rtval, totalv[i]);

		return rtval;

//		return totalv[NF-1];
	}

	VT FINLINE2
	nextAccelA(const VT& Anewval)
	{
		VT XTnewval(Anewval);

		VT ErtvalX(-320.0);

		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tcv = curval[0][i];

			VT attackdiff = (XTnewval - Tcv) * VT(2.0);
			VT attackdiffpos = Vmax(attackdiff, VT(0.0f));
			VT attack = attackdiffpos;

			VT attackincr = ad.atk(i) * attack;
			attackincr = Vmax(attackincr, VT(0.0f));
			VT attackTcv = Tcv + attackincr;

			Tcv = Vselect(Vgt(attackdiff, VT(0.0f)),
				Vmin(attackTcv, XTnewval),
				Vmax((Tcv - VT(ad.dcy(i))), XTnewval));

			ErtvalX = Vmax(ErtvalX, Tcv);
		}

		totalv[0] = ErtvalX;

		for (int j = 1; j < NF; j++)
		{
			VT rtvalX(-320.0);

			for (int i = 0; i < NSTEPS; i++)
			{
				VT& Tcv( curval[j][i]);

				const VT& Tnewval( curval[j-1][i] );

				VT attackdiff = (Tnewval - Tcv) * VT(2.0);
				VT attackdiffpos = Vmax(attackdiff, VT(0.0f));
				VT attack = attackdiffpos;

				VT attackincr = ad.atk(i) * attack;
				attackincr = Vmax(attackincr, VT(0.0f));
				VT attackTcv = Tcv + attackincr;

				Tcv = Vselect(Vgt(attackdiff, VT(0.0f)),
					Vmin(attackTcv, Tnewval),
					Vmax((Tcv - VT(ad.dcy(i))), Tnewval));

				rtvalX = Vmax(rtvalX, Tcv);
			}

			totalv[j] = rtvalX;
		}

		VT rtval(totalv[NF-1]);
		for (int i = 0; i < NF; ++i)
			rtval = Vmax(rtval, totalv[i]);

		return rtval;

//		return totalv[NF-1];
	}

	VT FINLINE2
	nextAccelA(const VT& Anewval, float atkspdup)
	{
		VT XTnewval(Anewval);

		VT ErtvalX(-320.0);

		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tcv = curval[0][i];

			VT attackdiff = (XTnewval - Tcv) * VT(2.0);
			VT attackdiffpos = Vmax(attackdiff, VT(0.0f));
			VT attack = attackdiffpos;

			VT attackincr = ad.atk(i) * attack * atkspdup;
			attackincr = Vmax(attackincr, VT(0.0f));
			VT attackTcv = Tcv + attackincr;

			Tcv = Vselect(Vgt(attackdiff, VT(0.0f)),
				Vmin(attackTcv, XTnewval),
				Vmax((Tcv - VT(ad.dcy(i))), XTnewval));

			ErtvalX = Vmax(ErtvalX, Tcv);
		}

		totalv[0] = ErtvalX;

		for (int j = 1; j < NF; j++)
		{
			VT rtvalX(-320.0);

			for (int i = 0; i < NSTEPS; i++)
			{
				VT& Tcv( curval[j][i]);

				const VT& Tnewval( curval[j-1][i] );

				VT attackdiff = (Tnewval - Tcv) * VT(2.0);
				VT attackdiffpos = Vmax(attackdiff, VT(0.0f));
				VT attack = attackdiffpos;

				VT attackincr = ad.atk(i) * attack * atkspdup;
				attackincr = Vmax(attackincr, VT(0.0f));
				VT attackTcv = Tcv + attackincr;

				Tcv = Vselect(Vgt(attackdiff, VT(0.0f)),
					Vmin(attackTcv, Tnewval),
					Vmax((Tcv - VT(ad.dcy(i))), Tnewval));

				rtvalX = Vmax(rtvalX, Tcv);
			}

			totalv[j] = rtvalX;
		}

		VT rtval(totalv[NF-1]);
		for (int i = 0; i < NF; ++i)
			rtval = Vmax(rtval, totalv[i]);

		return rtval;

//		return totalv[NF-1];
	}

	VT FINLINE2
	nextAccelB(const VT& Anewval)
	{
		VT XTnewval(Anewval);

		VT ErtvalX(-320.0);

		for (int i = 0; i < NSTEPS; i++)
		{
			VT& Tcv = curval[0][i];

			VT attackdiff = (XTnewval - Tcv);
			VT attackdiffpos = Vmax(attackdiff, VT(0.0f));
			VT attack = attackdiffpos;

			VT decaydiff = (XTnewval - Tcv);
			VT decaydiffneg = Vmin(decaydiff, VT(0.0f));
			VT decay = decaydiffneg;

			VT attackincr = ad.atk(i) * attack;
			VT decayincr = ad.dcy(i) * decay;
			VT decayincrmax = Vmax(decayincr, ad.dcy(0) * VT(-1.0));
			decayincr = Vmax(decayincr, decayincrmax);
			attackincr = Vmax(attackincr, VT(0.0f));
			decayincr = Vmin(decayincr, VT(0.0f));
			VT attackTcv = Tcv + attackincr;
			VT decayTcv = Tcv + decayincr;

			Tcv = Vselect(Vgt(attackdiff, VT(0.0f)),
				Vmin(attackTcv, XTnewval),
				Vmax(decayTcv, XTnewval));

			ErtvalX = Vmax(ErtvalX, Tcv);
		}

		totalv[0] = ErtvalX;

		for (int j = 1; j < NF; j++)
		{
			VT rtvalX(-320.0);

			for (int i = 0; i < NSTEPS; i++)
			{
				VT& Tcv( curval[j][i]);

				const VT& Tnewval( curval[j-1][i] );

				VT atkflags = Vgt(Tnewval, Tcv);

				VT attackdiff = (Tnewval - Tcv);
				VT attackdiffpos = Vmax(attackdiff, VT(0.0f));
				VT attack = attackdiffpos;

				VT decaydiff = (XTnewval - Tcv);
				VT decaydiffneg = Vmin(decaydiff, VT(0.0f));
				VT decay = decaydiffneg;

				VT attackincr = ad.atk(i) * attack;
				VT decayincr = ad.dcy(i) * decay;
				attackincr = Vmax(attackincr, VT(0.0f));
				decayincr = Vmin(decayincr, VT(0.0f));

				VT attackTcv = Tcv + attackincr;
				VT decayTcv = Tcv + decayincr;

				Tcv = Vselect(Vgt(attackdiff, VT(0.0f)),
					Vmin(attackTcv, Tnewval),
					Vmax(decayTcv, Tnewval));

				rtvalX = Vmax(rtvalX, Tcv);
			}

			totalv[j] = rtvalX;
		}

		VT rtval(totalv[NF-1]);
		for (int i = 0; i < NF; ++i)
			rtval = Vmax(rtval, totalv[i]);

		return rtval;

//		return totalv[NF-1];
	}

	VT FINLINE2 //FINLINE0
	smooth(const VT& newval, IVT freeze = Ivarrayf8(0))
	{
		next(newval, freeze);
		if (NF == 1)
			return totalv[0];

		VT total(0.0f);
		for (int i = 0; i < NF; i++)
			total += dB2lin(varrayf8(totalv[i]));

		total *= VT(1.0f/NF);

		return lin2dB(total);
	}

	template <typename XDEF>
	FINLINE0 varrayT<XDEF>
	smooth(const varrayT<XDEF>& newval, IvarrayT<XDEF> freeze = IvarrayT<XDEF>(0))
	{
		return smooth(VT(newval), freeze);
	}

	void FINLINE0
	setcurvals(const VT& Ainitval)
	{
		VT Linitval(Ainitval);
		for (int j = 0; j < NF; j++)
		{
			for (int i = 0; i < NSTEPS; i++)
			{
				curval[j][i] = Linitval;
			}
		}
		totalv = Linitval;
	}

	void
	setcurvals(float Ainitval)
	{
		VT initv(Ainitval);
		setcurvals(initv);
	}

	template <typename XT>
	void
	setcurvals(XT newval)
	{
		varrayf8 nv{0.0f};
		for (int i = 0; i < 8; i++)
			nv[i] = newval[i];
		setcurvals(nv);
		return;
	}

	VT
	getcurvals(void) const
	{
		return totalv[NF-1];
	}

	template <typename PV>
	PV operator()(const PV& input)
	{
		return next(input);
	}

	template <typename PV, typename IV>
	PV operator()(const PV& input, const IV& aa, const IV& adX)
	{
		return next(input, aa, adX);
	}

	template <typename PV, typename IV>
	PV operator()(const PV& input, const IV& aa, const IV& adX, bool atkall)
	{
		return next(input, aa, adX, atkall);
	}
};

template <typename VDEF, int NSTEPS=16, int NF=1>
class ALIGNED VfilterS
{
	typedef varrayT<VDEF> VT;
	typedef IvarrayT<VDEF> VIT;
	typedef typename VDEF::FT FT;

	atkdcys<VT,NSTEPS> ALIGNED ad;

	VT ALIGNED curval[NF][NSTEPS];
	VT ALIGNED zval[NF][NSTEPS];
	VT ALIGNED totalv[NF];
	VT initval;
	VT atks, atke;

	VfilterS& operator=(const VfilterS&) = delete;

public:
	typedef VT bsa;
	typedef VIT bsai;
	static constexpr int VS = VDEF::VS;

	VfilterS(VT atkst,VT atkend,
		VT dcyst, VT dcyend, VT Ainitval=FT(0.0f)) FINLINE0 :
		ad(atkst, atkend, dcyst, dcyend),
		curval(), zval(), totalv(), initval(Ainitval),
		atks(atkst), atke(atkend)
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
				zval[i][j] = VT(initval);
			}
			totalv[i] = VT(initval);
		}
	}

	VfilterS(FT atkst, FT atkend,
		FT dcyst, FT dcyend, FT Ainitval=FT(0.0f)) FINLINE0 :
		ad(VT(atkst), VT(atkend), VT(dcyst), VT(dcyend)),
		curval(), zval(), totalv(), initval(Ainitval),
		atks(VT(atkst)), atke(VT(atkend))
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = VT(initval);
				zval[i][j] = VT(initval);
			}
			totalv[i] = VT(initval);
		}
	}

	VfilterS(const VfilterS<VDEF,NSTEPS,NF>& init) FINLINE0 :
		ad(init.ad), curval(), zval(), totalv(), initval(init.initval),
		atks(init.atks), atke(init.atke)
		
	{
		for (int i = 0; i < NF; i++)
		{
			for (int j = 0; j < NSTEPS; j++)
			{
				curval[i][j] = init.curval[i][j];
				zval[i][j] = init.zval[i][j];
			}
			totalv[i] = init.totalv[i];
		}
	}

	~VfilterS()
	{
	}

	VT FINLINE1
	next(const VT& newval, bool atkdecayall = false)
	{
		for (int j = 0; j < NF; j++)
		{
			VT rtvalX(0.0f);

			rtvalX = VT(SMALLVAL);

			bool atkdecay = (j == 0) || atkdecayall;
			for (int i = 0; i < NSTEPS; i++)
			{
				const VT& atkmod = ad.atkmod(i);
				const VT& dcymod = ad.dcymod(i);
				VT& Tcv = curval[j][i];
				VT& Tzv = zval[j][i];

				const VT& Tnewval { (j != 0) ? curval[j-1][i] : newval };

				VT vdiff = Tnewval - Tzv;

				VT tmpA = vdiff * atkmod;
				VT tmpD = vdiff * dcymod;
				VT tmp = Vselect(Vgt(vdiff, Tcv), tmpA, tmpD);
				VT rtvalXtmp = Tzv + tmp;
				Tzv = rtvalXtmp + tmp;
				Tcv = rtvalXtmp;
				rtvalX = Vmax(rtvalX, (rtvalXtmp * ad.aratio(i)));

			}
			totalv[j] = rtvalX;
		}

		VT rtval = totalv[NF-1];
		for (int i = 0; i < NF; i++)
		{
			rtval = Vmax(rtval, totalv[i]);
		}
			
		return rtval;
	}

	VT FINLINE0
	smooth(const VT& newval, bool attackdecayall = false)
	{
		next(newval, attackdecayall);
		if (NF == 1)
			return totalv[0];
		VT total(0.0f);

		for (int i = 0; i < NF; i++)
			total += totalv[i];

		total *= VT(1.0f / NF);

		return total;
	}

	void FINLINE1
	setcurvals(const VT& initv)
	{
		for (int j = 0; j < NF; j++)
		{
			VT *cv = curval[j];
			for (int i = 0; i < NSTEPS; i++)
			{
				cv[i] = initv;
			}
			totalv[j] = initv;
		}
	}

	void
	setcurvals(FT newval)
	{
		VT nv(0.0f);
		for (int i = 0; i < VDEF::VS; i++)
			nv[i] = newval[i];
		setcurvals(nv);
		return;
	}

	VT
	getcurvals(void) const
	{
		return totalv;
	}

	void
	setatk(int idx, VT newval)
	{
		ad.setatk(idx, newval);
	}

	template <typename PV> FINLINE2
	PV operator()(const PV& input)
	{
		return next(input);
	}

	template <typename PV> FINLINE2
	PV operator()(const PV& input, bool atkall)
	{
		return next(input, atkall);
	}

	template <typename PV, typename IV> FINLINE2
	PV operator()(const PV& input, const IV& aa, const IV& adX)
	{
		return next(input, aa, adX);
	}

	template <typename PV, typename IV> FINLINE2
	PV operator()(const PV& input, const IV& aa, const IV& adX, bool atkall)
	{
		return next(input, aa, adX, atkall);
	}
};


#endif


//
// vim: set ts=4:
//
