#if !defined(__VARRAY_H__)
#define __VARRAY_H__

#if defined(__x86__) | defined(__x86_64__)
#define X86
#endif

#if defined(X86)
#include <stdio.h>
#endif

#include <stdlib.h>
#include <string.h>

#if 1 || defined(X86)
#include <initializer_list>
#define INITIALIZER_LIST
#endif

//
// BELOW -- v[x][sd][fi] are traditional GCC vector designations.
//

//typedef double v8df __attribute__((vector_size(64)));
typedef double v8df __attribute__((vector_size(8*sizeof(double))));
typedef double v4df __attribute__((vector_size(32)));
typedef double v2df __attribute__((vector_size(16)));

typedef float v16sf __attribute__((vector_size(64)));
//typedef float v8sf __attribute__((vector_size(32)));
typedef float v8sf __attribute__((vector_size(8*sizeof(float))));
typedef float v4sf __attribute__((vector_size(16)));
typedef float v2sf __attribute__((vector_size(8)));

typedef long long v8di __attribute__((vector_size(64)));
typedef long long v4di __attribute__((vector_size(32)));
typedef long long v2di __attribute__((vector_size(16)));

typedef unsigned long long v8du __attribute__((vector_size(64)));
typedef unsigned long long v4du __attribute__((vector_size(32)));
typedef unsigned long long v2du __attribute__((vector_size(16)));

typedef int v16si __attribute__((vector_size(64)));
typedef int v8si __attribute__((vector_size(32)));
typedef int v4si __attribute__((vector_size(16))); 
typedef int v2si __attribute__((vector_size(8))); 

typedef unsigned int v16su __attribute__((vector_size(64)));
typedef unsigned int v8su __attribute__((vector_size(32)));
typedef unsigned int v4su __attribute__((vector_size(16)));
typedef unsigned int v2su __attribute__((vector_size(8))); 

typedef short v16hi __attribute__((vector_size(64)));
typedef short v8hi __attribute__((vector_size(32)));
typedef short v4hi __attribute__((vector_size(16)));

typedef unsigned short v16hu __attribute__((vector_size(64)));
typedef unsigned short v8hu __attribute__((vector_size(32)));
typedef unsigned short v4hu __attribute__((vector_size(16)));

typedef char v16qi __attribute__((vector_size(64)));
typedef char v8qi __attribute__((vector_size(32)));
typedef char v4qi __attribute__((vector_size(16)));

typedef unsigned char v16qu __attribute__((vector_size(64)));
typedef unsigned char v8qu __attribute__((vector_size(32)));
typedef unsigned char v4qu __attribute__((vector_size(16)));

//
// Include the vecmathlib in another directory.
// alot of diagnostics turned off because of some lint in the files.
//
#if defined(X86)
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Weffc++"
#pragma clang diagnostic push
#pragma clang diagnostic ignored "-Weverything"
#include "vecmathlib.h"
#define USEVECMATHLIB
#pragma GCC diagnostic pop
#pragma clang diagnostic pop
#endif


//
// Simple vector init.  Useful especially for the traditional 'v8sf' types
//
#define INITALL(iavtype, iafptype, val) \
	iavtype { (iavtype) { \
			iafptype(val), iafptype(val), iafptype(val), iafptype(val), \
			iafptype(val), iafptype(val), iafptype(val), iafptype(val) \
	} }

//
//
// This vector scheme requires explicit definition of each vector type as follows:
//
// TYPEv[fd][vectorsize]x[#ofvectors]
//
// So, a traditional v8si or v8sf is defined by using:
//   IvarrayT<TYPEvf8x1> for v8si
//  or
//   varrayT<TYPEvf8x1> for v8sf
//
// The vectors of the same 'VDEF' type are mutually very compatible for conversion/etc,
// but vectors that don't match exactly do have some amount of compatibility.
//
// If a bit for bit non-conversion is needed, then something like this should be done:
//
// if you have (varrayT<TYPEvf8x1> fv) and (IvarrayT<TYPEvf8x1> iv) and you want a bit for bit copy,
// then use the following:
//
// fv = iv.raw();
//
// This vector scheme supplies alot of vector information without many layers of indirection and
// inferrence.
//
// First, there is a 'universal' class "UvarrayT" mostly meant to support raw conversions.
// Next, a floating point class "varrayT" supporting vector floating point values
// Finally, an integer class "IvarrayT" supporting vector integer values
//
// This scheme is probably extensible and adaptable to reasonable vector architectures, but is
// currently tuned for X86.  At one time, I did compile the code for ARM, and it appeared to produce
// good code.
//
// Most of my tests are with haswell, but with core2 and silvermont being secondary.
//
//

template <typename VDEF> class varrayT;			// FP vector
template <typename VDEF> class IvarrayT;		// INT vector
template <typename VDEF> union UvarrayT;		// Universal (null) type vector

class TYPEvf1x1
{
public:
	typedef unsigned int TYPEuint;

	static constexpr TYPEuint VS = 1;
	static constexpr TYPEuint VS2 = 1;
	static constexpr TYPEuint VS4 = 1;
	static constexpr TYPEuint NVT = 1;
	static constexpr TYPEuint NVT2 = 1;
	static constexpr TYPEuint NVT4 = 1;
	typedef float FT;
	typedef int IT;

	static constexpr TYPEuint sizeofFT = sizeof(FT);
	static constexpr TYPEuint sizeofVT = sizeofFT * VS;
	static constexpr TYPEuint sizeofVT2 = sizeofFT * VS2;
	static constexpr TYPEuint sizeofVT4 = sizeofFT * VS4;
	static constexpr TYPEuint sizeofAVT = NVT*sizeofVT;

#if defined(__clang__)
	typedef FT VFT __attribute__((ext_vector_type(VS)));
	typedef IT VIT __attribute__((ext_vector_type(VS)));
	typedef FT VFT2 __attribute__((ext_vector_type(VS2)));
	typedef IT VIT2 __attribute__((ext_vector_type(VS2)));
	typedef FT VFT4 __attribute__((ext_vector_type(VS4)));
	typedef IT VIT4 __attribute__((ext_vector_type(VS4)));
#else
	typedef FT VFT __attribute__((vector_size(sizeofVT)));
	typedef IT VIT __attribute__((vector_size(sizeofVT)));
	typedef FT VFT2 __attribute__((vector_size(sizeofVT2)));
	typedef IT VIT2 __attribute__((vector_size(sizeofVT2)));
	typedef FT VFT4 __attribute__((vector_size(sizeofVT4)));
	typedef IT VIT4 __attribute__((vector_size(sizeofVT4)));
#endif

	using MTYPE = TYPEvf1x1;
//	using FTYPE = float;
//	using ITYPE = int;
	using FTYPE = varrayT<MTYPE>;
	using ITYPE = IvarrayT<MTYPE>;
	using UTYPE = UvarrayT<MTYPE>;
	using TYPE2 = TYPEvf1x1;

	typedef const FT cFT;
	typedef const IT cIT;
	typedef const VFT cVFT;
	typedef const VIT cVIT;

	typedef VFT& raVFT;
	typedef VIT& raVIT;

	static constexpr TYPEuint nFTperVFT = sizeof(VFT) / sizeof(FT);
	static constexpr TYPEuint nFT = (NVT * sizeof(VFT)) / sizeof(FT);
	static constexpr TYPEuint nITperVIT = sizeof(VIT) / sizeof(IT);
	static constexpr TYPEuint nIT = (NVT * sizeof(VIT)) / sizeof(IT);
};

class TYPEvf2x1
{
public:
	typedef unsigned int TYPEuint;

	static constexpr TYPEuint VS = 2;
	static constexpr TYPEuint VS2 = 1;
	static constexpr TYPEuint VS4 = 1;
	static constexpr TYPEuint NVT = 1;
	static constexpr TYPEuint NVT2 = 2;
	static constexpr TYPEuint NVT4 = 2;
	typedef float FT;
	typedef int IT;

	static constexpr TYPEuint sizeofFT = sizeof(FT);
	static constexpr TYPEuint sizeofVT = sizeofFT * VS;
	static constexpr TYPEuint sizeofVT2 = sizeofFT * VS2;
	static constexpr TYPEuint sizeofVT4 = sizeofFT * VS4;
	static constexpr TYPEuint sizeofAVT = NVT*sizeofVT;

#if defined(__clang__)
	typedef FT VFT __attribute__((ext_vector_type(VS)));
	typedef IT VIT __attribute__((ext_vector_type(VS)));
	typedef FT VFT2 __attribute__((ext_vector_type(VS2)));
	typedef IT VIT2 __attribute__((ext_vector_type(VS2)));
	typedef FT VFT4 __attribute__((ext_vector_type(VS4)));
	typedef IT VIT4 __attribute__((ext_vector_type(VS4)));
#else
	typedef FT VFT __attribute__((vector_size(sizeofVT)));
	typedef IT VIT __attribute__((vector_size(sizeofVT)));
	typedef FT VFT2 __attribute__((vector_size(sizeofVT2)));
	typedef IT VIT2 __attribute__((vector_size(sizeofVT2)));
	typedef FT VFT4 __attribute__((vector_size(sizeofVT4)));
	typedef IT VIT4 __attribute__((vector_size(sizeofVT4)));
#endif

	using MTYPE = TYPEvf2x1;
	using FTYPE = varrayT<MTYPE>;
	using ITYPE = IvarrayT<MTYPE>;
	using UTYPE = UvarrayT<MTYPE>;
	using TYPE2 = TYPEvf1x1;

	typedef const FT cFT;
	typedef const IT cIT;
	typedef const VFT cVFT;
	typedef const VIT cVIT;

	typedef VFT& raVFT;
	typedef VIT& raVIT;

	static constexpr TYPEuint nFTperVFT = sizeof(VFT) / sizeof(FT);
	static constexpr TYPEuint nFT = (NVT * sizeof(VFT)) / sizeof(FT);
	static constexpr TYPEuint nITperVIT = sizeof(VIT) / sizeof(IT);
	static constexpr TYPEuint nIT = (NVT * sizeof(VIT)) / sizeof(IT);
};

class TYPEvf4x1
{
public:
	typedef unsigned int TYPEuint;

	static constexpr TYPEuint VS = 4;
	static constexpr TYPEuint VS2 = 2;
	static constexpr TYPEuint VS4 = 1;
	static constexpr TYPEuint NVT = 1;
	static constexpr TYPEuint NVT2 = 2;
	static constexpr TYPEuint NVT4 = 2;
	typedef float FT;
	typedef int IT;

	static constexpr TYPEuint sizeofFT = sizeof(FT);
	static constexpr TYPEuint sizeofVT = sizeofFT * VS;
	static constexpr TYPEuint sizeofVT2 = sizeofFT * VS2;
	static constexpr TYPEuint sizeofVT4 = sizeofFT * VS4;
	static constexpr TYPEuint sizeofAVT = NVT*sizeofVT;

#if defined(__clang__)
	typedef FT VFT __attribute__((ext_vector_type(VS)));
	typedef IT VIT __attribute__((ext_vector_type(VS)));
	typedef FT VFT2 __attribute__((ext_vector_type(VS2)));
	typedef IT VIT2 __attribute__((ext_vector_type(VS2)));
	typedef FT VFT4 __attribute__((ext_vector_type(VS4)));
	typedef IT VIT4 __attribute__((ext_vector_type(VS4)));
#else
	typedef FT VFT __attribute__((vector_size(sizeofVT)));
	typedef IT VIT __attribute__((vector_size(sizeofVT)));
	typedef FT VFT2 __attribute__((vector_size(sizeofVT2)));
	typedef IT VIT2 __attribute__((vector_size(sizeofVT2)));
	typedef FT VFT4 __attribute__((vector_size(sizeofVT4)));
	typedef IT VIT4 __attribute__((vector_size(sizeofVT4)));
#endif

	using MTYPE = TYPEvf4x1;
	using FTYPE = varrayT<MTYPE>;
	using ITYPE = IvarrayT<MTYPE>;
	using UTYPE = UvarrayT<MTYPE>;
	using TYPE2 = TYPEvf2x1;

	typedef const FT cFT;
	typedef const IT cIT;
	typedef const VFT cVFT;
	typedef const VIT cVIT;

	typedef VFT& raVFT;
	typedef VIT& raVIT;

	static constexpr TYPEuint nFTperVFT = sizeof(VFT) / sizeof(FT);
	static constexpr TYPEuint nFT = (NVT * sizeof(VFT)) / sizeof(FT);
	static constexpr TYPEuint nITperVIT = sizeof(VIT) / sizeof(IT);
	static constexpr TYPEuint nIT = (NVT * sizeof(VIT)) / sizeof(IT);
};

class TYPEvf8x1
{
public:
	typedef unsigned int TYPEuint;

	static constexpr TYPEuint VS = 8;
	static constexpr TYPEuint VS2 = 4;
	static constexpr TYPEuint VS4 = 2;
	static constexpr TYPEuint NVT = 1;
	static constexpr TYPEuint NVT2 = 2;
	static constexpr TYPEuint NVT4 = 4;
	typedef float FT;
	typedef int IT;

	static constexpr TYPEuint sizeofFT = sizeof(FT);
	static constexpr TYPEuint sizeofVT = sizeofFT * VS;
	static constexpr TYPEuint sizeofVT2 = sizeofFT * VS2;
	static constexpr TYPEuint sizeofVT4 = sizeofFT * VS4;
	static constexpr TYPEuint sizeofAVT = NVT*sizeofVT;

#if defined(__clang__)
	typedef FT VFT __attribute__((ext_vector_type(VS)));
	typedef IT VIT __attribute__((ext_vector_type(VS)));
	typedef FT VFT2 __attribute__((ext_vector_type(VS2)));
	typedef IT VIT2 __attribute__((ext_vector_type(VS2)));
	typedef FT VFT4 __attribute__((ext_vector_type(VS4)));
	typedef IT VIT4 __attribute__((ext_vector_type(VS4)));
#else
	typedef FT VFT __attribute__((vector_size(sizeofVT)));
	typedef IT VIT __attribute__((vector_size(sizeofVT)));
	typedef FT VFT2 __attribute__((vector_size(sizeofVT2)));
	typedef IT VIT2 __attribute__((vector_size(sizeofVT2)));
	typedef FT VFT4 __attribute__((vector_size(sizeofVT4)));
	typedef IT VIT4 __attribute__((vector_size(sizeofVT4)));
#endif

	using MTYPE = TYPEvf8x1;
	using FTYPE = varrayT<MTYPE>;
	using ITYPE = IvarrayT<MTYPE>;
	using UTYPE = UvarrayT<MTYPE>;
	using TYPE2 = TYPEvf4x1;

	typedef const FT cFT;
	typedef const IT cIT;
	typedef const VFT cVFT;
	typedef const VIT cVIT;

	typedef VFT& raVFT;
	typedef VIT& raVIT;

	static constexpr TYPEuint nFTperVFT = sizeof(VFT) / sizeof(FT);
	static constexpr TYPEuint nFT = (NVT * sizeof(VFT)) / sizeof(FT);
	static constexpr TYPEuint nITperVIT = sizeof(VIT) / sizeof(IT);
	static constexpr TYPEuint nIT = (NVT * sizeof(VIT)) / sizeof(IT);
};

class TYPEvd1x1
{
public:
	typedef unsigned int TYPEuint;

	static constexpr TYPEuint VS = 1;
	static constexpr TYPEuint VS2 = 1;
	static constexpr TYPEuint VS4 = 1;
	static constexpr TYPEuint NVT = 1;
	static constexpr TYPEuint NVT2 = 1;
	static constexpr TYPEuint NVT4 = 1;
	typedef double FT;
	typedef long long IT;

	static constexpr TYPEuint sizeofFT = sizeof(FT);
	static constexpr TYPEuint sizeofVT = sizeofFT * VS;
	static constexpr TYPEuint sizeofVT2 = sizeofFT * VS2;
	static constexpr TYPEuint sizeofVT4 = sizeofFT * VS4;
	static constexpr TYPEuint sizeofAVT = NVT*sizeofVT;

#if defined(__clang__)
	typedef FT VFT __attribute__((ext_vector_type(VS)));
	typedef IT VIT __attribute__((ext_vector_type(VS)));
	typedef FT VFT2 __attribute__((ext_vector_type(VS2)));
	typedef IT VIT2 __attribute__((ext_vector_type(VS2)));
	typedef FT VFT4 __attribute__((ext_vector_type(VS4)));
	typedef IT VIT4 __attribute__((ext_vector_type(VS4)));
#else
	typedef FT VFT __attribute__((vector_size(sizeofVT)));
	typedef IT VIT __attribute__((vector_size(sizeofVT)));
	typedef FT VFT2 __attribute__((vector_size(sizeofVT2)));
	typedef IT VIT2 __attribute__((vector_size(sizeofVT2)));
	typedef FT VFT4 __attribute__((vector_size(sizeofVT4)));
	typedef IT VIT4 __attribute__((vector_size(sizeofVT4)));
#endif

	using MTYPE = TYPEvd1x1;
	//using FTYPE = double;
	//using ITYPE = long long;
	using FTYPE = varrayT<MTYPE>;
	using ITYPE = IvarrayT<MTYPE>;
	using UTYPE = UvarrayT<MTYPE>;
	using TYPE2 = TYPEvd1x1;

	typedef const FT cFT;
	typedef const IT cIT;
	typedef const VFT cVFT;
	typedef const VIT cVIT;

	typedef VFT& raVFT;
	typedef VIT& raVIT;

	static constexpr TYPEuint nFTperVFT = sizeof(VFT) / sizeof(FT);
	static constexpr TYPEuint nFT = (NVT * sizeof(VFT)) / sizeof(FT);
	static constexpr TYPEuint nITperVIT = sizeof(VIT) / sizeof(IT);
	static constexpr TYPEuint nIT = (NVT * sizeof(VIT)) / sizeof(IT);
};

class TYPEvd2x1
{
public:
	typedef unsigned int TYPEuint;

	static constexpr TYPEuint VS = 2;
	static constexpr TYPEuint VS2 = 1;
	static constexpr TYPEuint VS4 = 1;
	static constexpr TYPEuint NVT = 1;
	static constexpr TYPEuint NVT2 = 2;
	static constexpr TYPEuint NVT4 = 2;
	typedef double FT;
	typedef long long IT;

	static constexpr TYPEuint sizeofFT = sizeof(FT);
	static constexpr TYPEuint sizeofVT = sizeofFT * VS;
	static constexpr TYPEuint sizeofVT2 = sizeofFT * VS2;
	static constexpr TYPEuint sizeofVT4 = sizeofFT * VS4;
	static constexpr TYPEuint sizeofAVT = NVT*sizeofVT;

#if defined(__clang__)
	typedef FT VFT __attribute__((ext_vector_type(VS)));
	typedef IT VIT __attribute__((ext_vector_type(VS)));
	typedef FT VFT2 __attribute__((ext_vector_type(VS2)));
	typedef IT VIT2 __attribute__((ext_vector_type(VS2)));
	typedef FT VFT4 __attribute__((ext_vector_type(VS4)));
	typedef IT VIT4 __attribute__((ext_vector_type(VS4)));
#else
	typedef FT VFT __attribute__((vector_size(sizeofVT)));
	typedef IT VIT __attribute__((vector_size(sizeofVT)));
	typedef FT VFT2 __attribute__((vector_size(sizeofVT2)));
	typedef IT VIT2 __attribute__((vector_size(sizeofVT2)));
	typedef FT VFT4 __attribute__((vector_size(sizeofVT4)));
	typedef IT VIT4 __attribute__((vector_size(sizeofVT4)));
#endif

	using MTYPE = TYPEvd2x1;
	using FTYPE = varrayT<MTYPE>;
	using ITYPE = IvarrayT<MTYPE>;
	using UTYPE = UvarrayT<MTYPE>;
	using TYPE2 = TYPEvd1x1;

	typedef const FT cFT;
	typedef const IT cIT;
	typedef const VFT cVFT;
	typedef const VIT cVIT;

	typedef VFT& raVFT;
	typedef VIT& raVIT;

	static constexpr TYPEuint nFTperVFT = sizeof(VFT) / sizeof(FT);
	static constexpr TYPEuint nFT = (NVT * sizeof(VFT)) / sizeof(FT);
	static constexpr TYPEuint nITperVIT = sizeof(VIT) / sizeof(IT);
	static constexpr TYPEuint nIT = (NVT * sizeof(VIT)) / sizeof(IT);
};

class TYPEvd4x1
{
public:
	typedef unsigned int TYPEuint;

	static constexpr TYPEuint VS = 4;
	static constexpr TYPEuint VS2 = 2;
	static constexpr TYPEuint VS4 = 2;
	static constexpr TYPEuint NVT = 1;
	static constexpr TYPEuint NVT2 = 2;
	static constexpr TYPEuint NVT4 = 4;
	typedef double FT;
	typedef long long IT;

	static constexpr TYPEuint sizeofFT = sizeof(FT);
	static constexpr TYPEuint sizeofVT = sizeofFT * VS;
	static constexpr TYPEuint sizeofVT2 = sizeofFT * VS2;
	static constexpr TYPEuint sizeofVT4 = sizeofFT * VS4;
	static constexpr TYPEuint sizeofAVT = NVT*sizeofVT;

#if defined(__clang__)
	typedef FT VFT __attribute__((ext_vector_type(VS)));
	typedef IT VIT __attribute__((ext_vector_type(VS)));
	typedef FT VFT2 __attribute__((ext_vector_type(VS2)));
	typedef IT VIT2 __attribute__((ext_vector_type(VS2)));
	typedef FT VFT4 __attribute__((ext_vector_type(VS4)));
	typedef IT VIT4 __attribute__((ext_vector_type(VS4)));
#else
	typedef FT VFT __attribute__((vector_size(sizeofVT)));
	typedef IT VIT __attribute__((vector_size(sizeofVT)));
	typedef FT VFT2 __attribute__((vector_size(sizeofVT2)));
	typedef IT VIT2 __attribute__((vector_size(sizeofVT2)));
	typedef FT VFT4 __attribute__((vector_size(sizeofVT4)));
	typedef IT VIT4 __attribute__((vector_size(sizeofVT4)));
#endif

	using MTYPE = TYPEvd4x1;
	using FTYPE = varrayT<MTYPE>;
	using ITYPE = IvarrayT<MTYPE>;
	using UTYPE = UvarrayT<MTYPE>;
	using TYPE2 = TYPEvd2x1;

	typedef const FT cFT;
	typedef const IT cIT;
	typedef const VFT cVFT;
	typedef const VIT cVIT;

	typedef VFT& raVFT;
	typedef VIT& raVIT;

	static constexpr TYPEuint nFTperVFT = sizeof(VFT) / sizeof(FT);
	static constexpr TYPEuint nFT = (NVT * sizeof(VFT)) / sizeof(FT);
	static constexpr TYPEuint nITperVIT = sizeof(VIT) / sizeof(IT);
	static constexpr TYPEuint nIT = (NVT * sizeof(VIT)) / sizeof(IT);
};

class TYPEvd8x1
{
public:
	typedef unsigned int TYPEuint;

	static constexpr TYPEuint VS = 8;
	static constexpr TYPEuint VS2 = 4;
	static constexpr TYPEuint VS4 = 2;
	static constexpr TYPEuint NVT = 1;
	static constexpr TYPEuint NVT2 = 2;
	static constexpr TYPEuint NVT4 = 4;
	typedef double FT;
	typedef long long IT;

	static constexpr TYPEuint sizeofFT = sizeof(FT);
	static constexpr TYPEuint sizeofVT = sizeofFT * VS;
	static constexpr TYPEuint sizeofVT2 = sizeofFT * VS2;
	static constexpr TYPEuint sizeofVT4 = sizeofFT * VS4;
	static constexpr TYPEuint sizeofAVT = NVT*sizeofVT;

#if defined(__clang__)
	typedef FT VFT __attribute__((ext_vector_type(VS)));
	typedef IT VIT __attribute__((ext_vector_type(VS)));
	typedef FT VFT2 __attribute__((ext_vector_type(VS2)));
	typedef IT VIT2 __attribute__((ext_vector_type(VS2)));
	typedef FT VFT4 __attribute__((ext_vector_type(VS4)));
	typedef IT VIT4 __attribute__((ext_vector_type(VS4)));
#else
	typedef FT VFT __attribute__((vector_size(sizeofVT)));
	typedef IT VIT __attribute__((vector_size(sizeofVT)));
	typedef FT VFT2 __attribute__((vector_size(sizeofVT2)));
	typedef IT VIT2 __attribute__((vector_size(sizeofVT2)));
	typedef FT VFT4 __attribute__((vector_size(sizeofVT4)));
	typedef IT VIT4 __attribute__((vector_size(sizeofVT4)));
#endif

	using MTYPE = TYPEvd8x1;
	using FTYPE = varrayT<MTYPE>;
	using ITYPE = IvarrayT<MTYPE>;
	using UTYPE = UvarrayT<MTYPE>;
	using TYPE2 = TYPEvd4x1;

	typedef const FT cFT;
	typedef const IT cIT;
	typedef const VFT cVFT;
	typedef const VIT cVIT;

	typedef VFT& raVFT;
	typedef VIT& raVIT;

	static constexpr TYPEuint nFTperVFT = sizeof(VFT) / sizeof(FT);
	static constexpr TYPEuint nFT = (NVT * sizeof(VFT)) / sizeof(FT);
	static constexpr TYPEuint nITperVIT = sizeof(VIT) / sizeof(IT);
	static constexpr TYPEuint nIT = (NVT * sizeof(VIT)) / sizeof(IT);
};

template <typename VDEF>
union UvarrayT
{
public:
	typedef unsigned int TYPEuint;

	using FT = typename VDEF::FT;
	using IT = typename VDEF::IT;

	using cFT = typename VDEF::cFT;
	using cIT = typename VDEF::cIT;

	using VFT = typename VDEF::VFT;
	using VIT = typename VDEF::VIT;

	using VFT2 = typename VDEF::VFT2;
	using VIT2 = typename VDEF::VIT2;

	using VFT4 = typename VDEF::VFT4;
	using VIT4 = typename VDEF::VIT4;

	using cVFT = typename VDEF::cVFT;
	using cVIT = typename VDEF::cVIT;

	using raVFT = typename VDEF::raVFT;
	using raVIT = typename VDEF::raVIT;

	using MTYPE = typename VDEF::MTYPE;

	typedef VFT& rVFT;
	typedef VIT& rVIT;

	static constexpr TYPEuint VS = VDEF::VS;
	static constexpr TYPEuint NVT = VDEF::NVT;
	static constexpr TYPEuint NVT2 = VDEF::NVT2;
	static constexpr TYPEuint NVT4 = VDEF::NVT4;

	static constexpr TYPEuint nFTperVFT = VDEF::nFTperVFT;
	static constexpr TYPEuint nFT = VDEF::nFT;

	static constexpr TYPEuint nITperVIT = VDEF::nFTperVIT;
	static constexpr TYPEuint nIT = VDEF::nIT;

	union 
	{
		VFT vdata[NVT];
		VFT2 vdata2[NVT2];
		VFT4 vdata4[NVT4];
		FT sdata[nFT];
		VIT vdatai[NVT];
		VIT2 vdata2i[NVT2];
		VIT4 vdata4i[NVT4];
		IT sdatai[nFT];
	};

	UvarrayT() : vdata() {}

//
// Code gets much larger if this is enabled
//
#if 0
	UvarrayT(const UvarrayT& init) : vdatai()
	{
		for (int i = 0; i < NVT; i++)
			vdatai[i] = init.vdatai[i];
	}

	UvarrayT&
	operator=(const UvarrayT& src)
	{
		for (int i = 0; i < NVT; i++)
			vdatai[i] = src.vdatai[i];
		return *this;
	}
#endif
};


//
// IvarrayT is mostly a helper class to support comparisons.
// it is also useful to support integer vectors, but is less supported
// than the varrayT classes for floating point
//
template <typename VDEF>
class IvarrayT
{
public:
	typedef unsigned int TYPEuint;

	using FT = typename VDEF::FT;
	using IT = typename VDEF::IT;

	using cFT = typename VDEF::cFT;
	using cIT = typename VDEF::cIT;

	using VFT = typename VDEF::VFT;
	using VIT = typename VDEF::VIT;

	using VFT2 = typename VDEF::VFT2;
	using VIT2 = typename VDEF::VIT2;

	using VFT4 = typename VDEF::VFT4;
	using VIT4 = typename VDEF::VIT4;

	using cVFT = typename VDEF::cVFT;
	using cVIT = typename VDEF::cVIT;

	using raVFT = typename VDEF::raVFT;
	using raVIT = typename VDEF::raVIT;

	using MTYPE = typename VDEF::MTYPE;

	typedef VFT& rVFT;
	typedef VIT& rVIT;

	typedef const IvarrayT<VDEF> cIvarrayT;

	static constexpr TYPEuint VS = VDEF::VS;
	static constexpr TYPEuint NVT = VDEF::NVT;
	static constexpr TYPEuint NVT2 = VDEF::NVT2;
	static constexpr TYPEuint NVT4 = VDEF::NVT4;

	static constexpr TYPEuint nFTperVFT = VDEF::nFTperVFT;
	static constexpr TYPEuint nFT = VDEF::nFT;

	static constexpr TYPEuint nITperVIT = VDEF::nFTperVIT;
	static constexpr TYPEuint nIT = VDEF::nIT;

//private:

	UvarrayT<VDEF> u;

public:
	IvarrayT(void) : u()
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdatai[i] = VIT{0};
		}
	}

	IvarrayT(const typename VDEF::VIT& init) : u()
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdatai[i] = init;
		}
	}


	IvarrayT(IT init) : u()
	{
		for (TYPEuint i = 0; i < nIT; i++)
		{
			u.sdatai[i] = init;
		}
	}

	IvarrayT(const UvarrayT<VDEF>& init) : u()
	{
		u = init;
	}
	
#if defined(INITIALIZER_LIST)
	IvarrayT(const std::initializer_list<IT>& initargs) : u()
	{
		TYPEuint i = 0;
		IT tmp{0};
		for (auto p = initargs.begin(); p != initargs.end(); ++p, ++i)
		{
			if (i < nIT)
			{
				tmp = *p;
				u.sdatai[i] = tmp;
			}
		}
		while (i < nIT)
		{
			u.sdatai[i] = tmp;
			i++;
		}
	}

#if defined(__clang__)
	IvarrayT(const std::initializer_list<VIT>& initargs) : u()
	{
		TYPEuint i = 0;
		VIT tmp{0};
		for (auto p = initargs.begin(); p != initargs.end(); ++p, ++i)
		{
			if (i < NVT)
			{
				tmp = *p;
				u.vdatai[i] = tmp;
			}
		}
		while (i < NVT)
		{
			u.vdatai[i] = tmp;
			i++;
		}
	}
#endif
#endif

	IvarrayT<VDEF>&
	operator=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < nFT; i++)
		{
			u.sdatai[i] = IT(src.u.sdata[i]);
		}
		return *this;
	}

	template <typename XVDEF>
	IvarrayT(const IvarrayT<XVDEF>& init) : u()
	{
		for (TYPEuint i = 0; i < nIT; i++)
			u.sdatai[i] = init.Svali(i);
	}

	IvarrayT(const varrayT<VDEF>& init) : u()
	{
		for (TYPEuint i = 0; i < nFT; i++)
		{
			u.sdatai[i] = IT(init.u.sdata[i]);
		}
	}

	UvarrayT<VDEF>&
	raw(void)
	{
		return u;
	}

	const UvarrayT<VDEF>&
	raw(void) const
	{
		return u;
	}

	operator VIT()
	{
		return u.vdatai[0];
	}

	cVIT&
	Vvali(TYPEuint index) const
	{
		return u.vdatai[index];
	}

	VIT&
	Vvali(TYPEuint index)
	{
		return u.vdatai[index];
	}

	void
	setVvali(TYPEuint index, VIT newval)
	{
		u.vdatai[index] = newval;
	}

	cIT&
	Svali(TYPEuint index) const
	{
		return u.sdatai[index];
	}

	IT&
	Svali(TYPEuint index)
	{
		return u.sdatai[index];
	}

	void
	setSvali(TYPEuint index, IT newval)
	{
		u.sdatai[index] = newval;
	}

	cIT&
	operator[](TYPEuint index) const
	{
		return u.sdatai[index];
	}

	IT&
	operator[](TYPEuint index)
	{
		return u.sdatai[index];
	}

	IvarrayT&
	operator+=(const IvarrayT& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] += src.u.vdatai[i];
		return *this;
	}

	IvarrayT&
	operator-=(const IvarrayT& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] -= src.u.vdatai[i];
		return *this;
	}

	IvarrayT&
	operator*=(const IvarrayT& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] *= src.u.vdatai[i];
		return *this;
	}

	IvarrayT&
	operator/=(const IvarrayT& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] /= src.u.vdatai[i];
		return *this;
	}

	IvarrayT&
	operator*=(FT src)
	{
		IvarrayT multval(src);
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] *= multval.u.vdatai[i];
		return *this;
	}

	IvarrayT&
	operator/=(FT src)
	{
		IvarrayT divval(FT(1.0) / src);
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] *= divval.u.vdatai[i];
		return *this;
	}

	template <typename XDEF>
	IvarrayT<XDEF>&
	operator|=(const IvarrayT<XDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] |= src.u.vdatai[i];
		return *this;
	}

	template <typename XDEF>
	IvarrayT<XDEF>&
	operator&=(const IvarrayT<XDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] &= src.u.vdatai[i];
		return *this;
	}

	template <typename XDEF>
	IvarrayT<XDEF>&
	operator^=(const IvarrayT<XDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] ^= src.u.vdatai[i];
		return *this;
	}

	bool __attribute__((noinline))
	isanytrue(void) const
	{
		bool rtval = false;
		for (TYPEuint i = 0; i < NVT; i++)
		{
			for (TYPEuint j = 0; j < nIT; j++)
			{
				if (u.vdatai[i][j] != 0)
					rtval = true;
			}
		}
		return rtval;
	}

	template <typename XVDEF>
	typename IvarrayT<XVDEF>::VIT Vvalslice(TYPEuint index) const
	{
		using RT= typename XVDEF::VIT;

		RT rtval;

		for (TYPEuint i = 0; i < XVDEF::nIT; i++)
		{
			rtval[i] = u.sdatai[index + i];
		}

		return rtval;
	}

	template <typename XVDEF>
	void setVvalslice(TYPEuint index, const IvarrayT<XVDEF>& src)
	{
		for (TYPEuint i = 0; i < XVDEF::nIT; i++)
		{
			u.sdatai[index + i] = src[i];
		}

		return;
	}


static constexpr int maxfmtsize = ((nFT * 64) + NVT * 4 + 2);

	static char *
	correctfmt(char *fmtproposed)
	{
		int nchars = strlen(fmtproposed);
		if ((nchars >= 2) && fmtproposed[nchars - 2] != 'l')
		{
			int ncharsout = 0;
			char *newfmt = (char *) ::malloc(size_t(nchars + 16));
			int i = 0;
			for ( ; i < nchars - 2; i++)
			{
				char newc = fmtproposed[i];
				newfmt[ncharsout] = newc;
				ncharsout += 1;
			}
			if (fmtproposed[i] != 'l')
			{
				newfmt[ncharsout] = fmtproposed[i];
				ncharsout += 1;
				i += 1;
				newfmt[ncharsout] = 'l';
				ncharsout += 1;
			}

			while (i < nchars)
			{
				newfmt[ncharsout] = fmtproposed[i];
				ncharsout += 1;
				i += 1;
			}

			newfmt[ncharsout] = '\0';

			::free(fmtproposed);
			return newfmt;
		}

		return fmtproposed;
	}

	char*
	fmtoutval(const char *basefmt, bool intmode = 0) const
	{
		const char *fmtin = (basefmt != nullptr) ? basefmt : (intmode ? "%d" : "%f");

		char *fmt = ::strdup(fmtin);

		fmt = correctfmt(fmt);

		const int bfsz = ::strlen(fmt);
		char tmpfmt[maxfmtsize + 1] = {0};

		int cidx = 0;
		int cleft = maxfmtsize;
		int vidx = 0;
		char *cfp = tmpfmt;

		auto dpinc = [&cidx, &cleft](int cnt = 1)
		{
			cidx += cnt;
			cleft -= cnt;
		};

		auto dpchr = [&cfp,&cidx,&cleft,&dpinc](char nv)
		{
			cfp[cidx] = nv;
			dpinc();
		};

		dpchr('{');

		for (TYPEuint i = 0; i < NVT; i++)
		{
			dpchr('{');
			for (TYPEuint j = 0; j < nFTperVFT; j++)
			{
				if ((cidx + bfsz + 3) > maxfmtsize)
				{
					::free(fmt);
					return ::strdup("{format memory size error}");
				}
				int nsp = 0;

				if (sizeof(IT) == 4)
				{
					if (intmode)
					{
						long long sdi = (IT) u.sdatai[vidx];
//						sdi &= 0xffffffffUL;
						nsp = ::snprintf(&cfp[cidx], cleft, fmt, sdi);
					}
					else
					{
						nsp = ::snprintf(&cfp[cidx], cleft, fmt, (double) (u.sdata[vidx]));
					}
				}
				else
				{
					if (intmode)
						nsp = ::snprintf(&cfp[cidx], cleft, fmt, (long long) (u.sdatai[vidx]));
					else
						nsp = ::snprintf(&cfp[cidx], cleft, fmt, (double) (u.sdata[vidx]));
				}

				dpinc(nsp);

				vidx += 1;
	
				if (j == (nFTperVFT - 1))
					dpchr('}');
				else
					dpchr(',');
			}
			dpchr('}');
			if (i != (NVT - 1))
			{
				dpchr(',');
			}
		}

		dpchr('\0');

		::free(fmt);
		char *rtval = strdup(tmpfmt);
		return rtval;
	}

	char*
	snprint(char *buf, int bufsize, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt, true);
		::strncpy(buf, outval, (size_t) bufsize);
		::free(outval);
		return buf;
	}
		
	char*
	snprinti(char *buf, int bufsize, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt, true);
		::strncpy(buf, outval, (size_t) bufsize);
		::free(outval);
		return buf;
	}

	char*
	snprintfp(char *buf, int bufsize, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt, false);
		::strncpy(buf, outval, (size_t) bufsize);
		::free(outval);
		return buf;
	}

	void
	fprint(FILE *fp, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt, true);
		::fprintf(fp,"%s", outval);
		::free(outval);
		return;
	}

	void
	fprinti(FILE *fp, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt, true);
		::fprintf(fp,"%s", outval);
		::free(outval);
		return;
	}

	void
	fprintfp(FILE *fp, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt, false);
		::fprintf(fp,"%s", outval);
		::free(outval);
		return;
	}

	IvarrayT<VDEF>&
	operator~(void)
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdatai[i] = ~u.vdatai[i];
		}
		return *this;
	}

};

#if 0
template <typename VDEF>
	inline varrayT<VDEF> operator~(const varrayT<VDEF>& src0)
	{
		typedef unsigned int TYPEuint;
		varrayT <VDEF> rtval;
		for (TYPEuint i = 0; i < VDEF::nIT; i++)
			rtval.setSvali(i, ~(src0.Svali(i)));

		return rtval;
	}
#endif

#if 0 && defined(__AVX__)
	IvarrayT<TYPEvf8x1>
	operator~(const IvarrayT<TYPEvf8x1>& src0)
	{
		IvarrayT<TYPEvf8x1> rtval;
		const IvarrayT<TYPEvf8x1> src1(~0);
		asm("vpxor %2,%1,%0" : "=x"(rtval.u.vdatai[0]) : "%x"(src1.u.vdatai[0]), "x"(src0.u.vdatai[0]));
		return rtval;
	}
#endif


//
// This supports a vector of any size multiple of the intrinsic
// vector support on the running CPU.
//
// For example, one can use a v8sf on a current Intel machine,
// or even a v16sf (if you really want to -- but very dependent upon
// compiler support.)  Instead, one can use a v8sf[2] type,
// and perhaps get better results from the compiler.
//
// As it is, one can access the vector by individual scalar
// components, vector items, or the full vector itself.
//
template <typename VDEF>
class varrayT
{
public:

	typedef unsigned int TYPEuint;

	using FT = typename VDEF::FT;
	using IT = typename VDEF::IT;

	using cFT = typename VDEF::cFT;
	using cIT = typename VDEF::cIT;

	using VFT = typename VDEF::VFT;
	using VIT = typename VDEF::VIT;

	using VFT2 = typename VDEF::VFT2;
	using VIT2 = typename VDEF::VIT2;

	using VFT4 = typename VDEF::VFT4;
	using VIT4 = typename VDEF::VIT4;

	using cVFT = typename VDEF::cVFT;
	using cVIT = typename VDEF::cVIT;

	using raVFT = typename VDEF::raVFT;
	using raVIT = typename VDEF::raVIT;

	using MTYPE = typename VDEF::MTYPE;

	//using VFT2T = typename varrayT<typename VDEF::TYPE2>;
	//using VIT2T = typename IvarrayT<typename VDEF::TYPE2>;
	using TYPE2 = typename VDEF::TYPE2;

	typedef VFT& rVFT;
	typedef VIT& rVIT;

	typedef const varrayT<VDEF> cvarrayT;

	static constexpr TYPEuint NVT = VDEF::NVT;
	static constexpr TYPEuint NVT2 = VDEF::NVT2;
	static constexpr TYPEuint NVT4 = VDEF::NVT4;
	static constexpr TYPEuint VS = VDEF::VS;

	static constexpr TYPEuint nFTperVFT = VDEF::nFTperVFT;
	static constexpr TYPEuint nFT = VDEF::nFT;

	static constexpr TYPEuint nITperVIT = VDEF::nFTperVIT;
	static constexpr TYPEuint nIT = VDEF::nIT;

	//
	// general VFT
	//

	template <typename XVDEF>
	class buildvt
	{
	private:
		static constexpr int XVFTbytes = XVDEF::VS * sizeof(XVDEF::FT);
	public:
#if defined(__clang__)
		typedef typename XVDEF::FT GVFT __attribute__((ext_vector_type(XVDEF::VS)));
		typedef typename XVDEF::IT GVIT __attribute__((ext_vector_type(XVDEF::XVS)));
#else
		typedef typename XVDEF::FT GVFT __attribute__((vector_size(XVDEF::sizeofVT)));
		typedef typename XVDEF::IT GVIT __attribute__((vector_size(XVDEF::sizeofVT)));
#endif
	};

//private:

	UvarrayT<VDEF> u;

public:
	friend varrayT<typename VDEF::TYPE2>;

	varrayT(void) : u()
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdata[i] = VFT{FT(0.0)};
		}
	}

// Smaller code if we don't define this!!!
#if 0
	varrayT(const varrayT<VDEF>& init) : u()
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdata[i] = init.u.vdata[i];
		}
	}
#endif

#if !defined(__INTEL_COMPILER)
	varrayT(FT init) : u()
	{
		for (TYPEuint i = 0; i < nFT; i++)
		{
			u.sdata[i] = init;
		}
	}
#endif

//
// Initializers shorter than the array will fill the rest of the
// array with the last initilizer value.
//
#if defined(INITIALIZER_LIST)
	varrayT(const std::initializer_list<FT>& initargs) : u()
	{
		TYPEuint i = 0;
		FT tmp = FT(0.0);
		for (auto p = initargs.begin(); p != initargs.end(); ++p, ++i)
		{
			if (i < nFT)
			{
				tmp = *p;
				u.sdata[i] = tmp;
			}
		}
		while (i < nFT)
		{
			u.sdata[i] = tmp;
			i++;
		}
	}

#if defined(__clang__)
	varrayT(const std::initializer_list<VFT>& initargs) : u()
	{
		TYPEuint i = 0;
		VFT tmp = FT(0.0);
		for (auto p = initargs.begin(); p != initargs.end(); ++p, ++i)
		{
			if (i < NVT)
			{
				tmp = *p;
				u.vdata[i] = tmp;
			}
		}
		while (i < NVT)
		{
			u.vdata[i] = tmp;
			i++;
		}
	}
#endif
#endif

	varrayT(const VFT& init) : u()
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdata[i] = init;
		}
	}

#define EXTRASLOPPY 
#if defined(EXTRASLOPPY)
	template <typename XVDEF>
	varrayT(const varrayT<XVDEF>& init) : u()
	{
		for (TYPEuint i = 0; i < nFT; i++)
			u.sdata[i] = init.Sval(i);
	}
#endif

	varrayT(const UvarrayT<VDEF>& init) : u()
	{
		u = init;
	}

#if defined(USEVECMATHLIB)
	varrayT(vecmathlib::realvec<typename VDEF::FT, VDEF::VS> init) : u()
	{
		for (TYPEuint i = 0; i < VDEF::NVT; i++)
		{
			for (TYPEuint j = 0; j < VDEF::VS; j++)
			{
				u.vdata[i][j] = init[int(j)];
			}
		}
	}
#endif


#if defined(EXTRASLOPPY)
#if 1 || defined(__clang__)
	template <typename XVDEF>
	varrayT(const typename XVDEF::VFT& init) : u()
	{
		for (TYPEuint i = 0; i < nFT; i++)
			u.sdata[i] = init.Sval(i);
	}
#endif
#endif

	varrayT(const cFT* init, TYPEuint maxinit = nFT) : u()
	{
		TYPEuint nfpinit = (maxinit > nFT) ? nFT : maxinit;

		for (TYPEuint i = 0; i < nfpinit; i++)
		{
			u.sdata[i] = init[i];
		}
	}

	template <typename FFT>
		varrayT(const FFT* init, TYPEuint maxinit = nFT) : u()
		{
			TYPEuint nfpinit = (maxinit > nFT) ? nFT : maxinit;

			for (TYPEuint i = 0; i < nfpinit; i++)
			{
				u.sdata[i] = cFT(init[i]);
			}
		}

	varrayT(const IvarrayT<VDEF>& init) : u()
	{
		for (TYPEuint i = 0; i < nFT; i++)
			u.sdata[i] = FT(init.u.sdatai[i]);
	}

	UvarrayT<VDEF>&
	raw(void)
	{
		return u;
	}

	const UvarrayT<VDEF>&
	raw(void) const
	{
		return u;
	}

	void
	copyout(FT* dest, TYPEuint maxsize = nFT) const
	{
		for (TYPEuint i = 0; i < maxsize; i++)
		{
			dest[i] = u.sdata[i];
		}
	}

/*
	varrayT<VDEF>&
	operator=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdata[i] = src.u.vdata[i];
		}
		return *this;
	}

	template <typename XVDEF>
		varrayT&
		operator=(const varrayT<XVDEF>& src)
	{
		int nft = (nFT <= XVDEF::nFT) ? nFT : XVDEF::nFT;
		for (TYPEuint i = 0; i < NVT; i++)
		{
			for (TYPEuint j = 0; j < VS; j++)
			{
				u.sdata[i][j] = src[j];
			}
		}
		return *this;
	}
*/

	varrayT<VDEF>&
	operator=(const varrayT<VDEF>& src) const
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdatai[i] = src.u.vdatai[i];
		}
		return *this;
	}

	varrayT<VDEF>&
	operator=(const varrayT<VDEF>& src) 
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdatai[i] = src.u.vdatai[i];
		}
		return *this;
	}

	varrayT<VDEF>&
	operator=(const IvarrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < nFT; i++)
			u.sdata[i] = FT(src.u.sdatai[i]);
		return *this;
	}

	const varrayT&
	operator=(const VFT& src) const
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdata[i] = src;
		}
		return *this;
	}

	varrayT&
	operator=(const VFT& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
		{
			u.vdata[i] = src;
		}
		return *this;
	}

	operator VFT() 
	{
		return u.vdata[0];
	}
		
	cVFT&
	Vval(TYPEuint index) const
	{
		return u.vdata[index];
	}

	VFT&
	Vval(TYPEuint index)
	{
		return u.vdata[index];
	}

	cVIT&
	Vvali(TYPEuint index) const
	{
		return u.vdatai[index];
	}

	VIT&
	Vvali(TYPEuint index)
	{
		return u.vdatai[index];
	}

	void
	setVval(TYPEuint index, VFT newval)
	{
		u.vdata[index] = newval;
	}

	void
	setVvali(TYPEuint index, VIT newval)
	{
		u.vdatai[index] = newval;
	}

	cFT&
	Sval(TYPEuint index) const
	{
		return u.sdata[index];
	}

	FT&
	Sval(TYPEuint index) 
	{
		return u.sdata[index];
	}

	cIT&
	Svali(TYPEuint index) const
	{
		return u.sdatai[index];
	}

	IT&
	Svali(TYPEuint index)
	{
		return u.sdatai[index];
	}

	void
	setSval(TYPEuint index, FT newval)
	{
		u.sdata[index] = newval;
	}

	void
	setSvali(TYPEuint index, IT newval)
	{
		u.sdatai[index] = newval;
	}

	bool
	alltrue(void)
	{
		for (int i = 0; i < nFT; i++)
		{
			if (u.sdatai[i] != ~(IT(0)))
				return false;
		}
		return true;
	}

	template <typename XVDEF>
	typename varrayT<XVDEF>::VFT Vvalslice(TYPEuint index) const
	{
		using RT= typename XVDEF::VFT;

		RT rtval;

		for (TYPEuint i = 0; i < XVDEF::nFT; i++)
		{
			rtval[i] = u.sdata[index + i];
		}

		return rtval;
	}

	template <typename XVDEF>
	void setVvalslice(TYPEuint index, const varrayT<XVDEF>& src)
	{
		for (TYPEuint i = 0; i < XVDEF::nFT; i++)
		{
			u.sdata[index + i] = src[i];
		}

		return;
	}

	cFT&
	operator[](TYPEuint index) const
	{
		return u.sdata[index];
	}

	FT&
	operator[](TYPEuint index)
	{
		return u.sdata[index];
	}

	cFT&
	operator[](int index) const
	{
		return u.sdata[index];
	}

	FT&
	operator[](int index)
	{
		return u.sdata[index];
	}

	varrayT<VDEF>&
	operator+=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdata[i] += src.u.vdata[i];
		return *this;
	}

	varrayT<VDEF>&
	operator-=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdata[i] -= src.u.vdata[i];
		return *this;
	}

	varrayT<VDEF>&
	operator*=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdata[i] *= src.u.vdata[i];
		return *this;
	}

	varrayT<VDEF>&
	operator/=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdata[i] /= src.u.vdata[i];
		return *this;
	}

	varrayT<VDEF>&
	operator*=(FT src)
	{
		varrayT multval(src);
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdata[i] *= multval.u.vdata[i];
		return *this;
	}

	varrayT<VDEF>&
	operator/=(FT src)
	{
		varrayT divval(FT(1.0) / src);
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdata[i] *= divval.u.vdata[i];
		return *this;
	}

	varrayT<VDEF>&
	operator|=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] |= src.u.vdatai[i];
		return *this;
	}

	varrayT<VDEF>&
	operator&=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] &= src.u.vdatai[i];
		return *this;
	}

	varrayT<VDEF>&
	operator^=(const varrayT<VDEF>& src)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] ^= src.u.vdatai[i];
		return *this;
	}

	bool
	isanytrue(void) const
	{
		bool rtval = false;
		for (TYPEuint i = 0; i < NVT; i++)
		{
			for (TYPEuint j = 0; j < nIT; j++)
			{
				if (u.vdatai[i][j] != 0)
					rtval = true;
			}
		}
		return rtval;
	}

static constexpr int maxfmtsize = ((nFT * 64) + NVT * 4 + 2);

	static char *
	correctfmt(char *fmtproposed)
	{
		size_t nchars = ::strlen(fmtproposed);
		if ((nchars >= 2) && fmtproposed[nchars - 2] != 'l')
		{
			int ncharsout = 0;
			char *newfmt = (char *) ::malloc(size_t(nchars + 16));
			size_t i = 0;
			for ( ; i < nchars - 2; i++)
			{
				char newc = fmtproposed[i];
				newfmt[ncharsout] = newc;
				ncharsout += 1;
			}
			if (fmtproposed[i] != 'l')
			{
				newfmt[ncharsout] = fmtproposed[i];
				ncharsout += 1;
				i += 1;
				newfmt[ncharsout] = 'l';
				ncharsout += 1;
			}

			while (i < nchars)
			{
				newfmt[ncharsout] = fmtproposed[i];
				ncharsout += 1;
				i += 1;
			}

			newfmt[ncharsout] = '\0';

			if (fmtproposed != nullptr) ::free(fmtproposed);
			return newfmt;
		}

		return fmtproposed;
	}

	char*
	fmtoutval(const char *basefmt, bool intmode = 0) const
	{
		const char *fmtin = (basefmt != nullptr) ? basefmt : (intmode ? "%d" : "%f");

		char *fmt = ::strdup(fmtin);

		fmt = correctfmt(fmt);

		const size_t bfsz = ::strlen(fmt);
		char tmpfmt[maxfmtsize + 1] = {0};

		unsigned int cidx = 0;
		unsigned int cleft = maxfmtsize;
		unsigned int vidx = 0;
		char *cfp = tmpfmt;

		auto dpinc = [&cidx, &cleft](size_t cnt = 1)
		{
			cidx += cnt;
			cleft -= cnt;
		};

		auto dpchr = [&cfp,&cidx,&dpinc](char nv)
		{
			cfp[cidx] = nv;
			dpinc();
		};

		dpchr('{');

		for (TYPEuint i = 0; i < NVT; i++)
		{
			dpchr('{');
			for (TYPEuint j = 0; j < nFTperVFT; j++)
			{
				if ((cidx + bfsz + 3) > maxfmtsize)
				{
					if (fmt != nullptr) ::free(fmt);
					return ::strdup("{format memory size error}");
				}
				size_t nsp = 0;

				if (intmode)
					nsp = ::snprintf(&cfp[cidx], cleft, fmt, (long long) (u.sdatai[vidx]));
				else
					nsp = ::snprintf(&cfp[cidx], cleft, fmt, (double) (u.sdata[vidx]));

				dpinc(nsp);

				vidx += 1;
	
				if (j == (nFTperVFT - 1))
					dpchr('}');
				else
					dpchr(',');
			}
			dpchr('}');
			if (i != (NVT - 1))
			{
				dpchr(',');
			}
		}

		dpchr('\0');

		if (fmt != nullptr)
			::free(fmt);
		char *rtval = ::strdup(tmpfmt);
		return rtval;
	}

	char*
	snprint(char *buf, int bufsize, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt);
		::strncpy(buf, outval, (size_t) bufsize);
		::free(outval);
		return buf;
	}
		
	char*
	snprinti(char *buf, int bufsize, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt, true);
		::strncpy(buf, outval, (size_t) bufsize);
		::free(outval);
		return buf;
	}

	void
	fprint(FILE *fp, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt);
		::fprintf(fp,"%s", outval);
		::free(outval);
		return;
	}

	void
	fprinti(FILE *fp, const char *efmt = nullptr) const
	{
		char *outval = fmtoutval(efmt, true);
		::fprintf(fp,"%s", outval);
		::free(outval);
		return;
	}

// JSD
	varrayT<VDEF>&
	operator~(void)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdatai[i] = ~u.vdatai[i];
		return *this;
	}

	varrayT<VDEF>&
	operator-(void)
	{
		for (TYPEuint i = 0; i < NVT; i++)
			u.vdata[i] = FT(-1.0) * u.vdata[i];
		return *this;
	}

};

typedef long long vall;
typedef varrayT<TYPEvf8x1> varrayf8x1;
typedef varrayT<TYPEvd8x1> varrayd8x1;
typedef IvarrayT<TYPEvf8x1> Ivarrayf8x1;
typedef IvarrayT<TYPEvd8x1> Ivarrayd8x1;
typedef varrayf8x1 varrayf8;
typedef varrayd8x1 varrayd8;
typedef Ivarrayf8x1 Ivarrayf8;
typedef Ivarrayd8x1 Ivarrayd8;

typedef varrayT<TYPEvf4x1> varrayf4x1;
typedef varrayT<TYPEvd4x1> varrayd4x1;
typedef varrayT<TYPEvd2x1> varrayd2x1;
typedef IvarrayT<TYPEvf4x1> Ivarrayf4x1;
typedef IvarrayT<TYPEvd4x1> Ivarrayd4x1;
typedef IvarrayT<TYPEvd2x1> Ivarrayd2x1;
typedef varrayf4x1 varrayf4;
typedef varrayd4x1 varrayd4;
typedef varrayd2x1 varrayd2;
typedef Ivarrayf4x1 Ivarrayf4;
typedef Ivarrayd4x1 Ivarrayd4;
typedef Ivarrayd2x1 Ivarrayd2;



#if defined(__AVX__)
	template<>
	inline varrayT<TYPEvf4x1>&
	varrayT<TYPEvf4x1>::operator|=(const varrayT<TYPEvf4x1>& src0)
	{
		asm("vorps %2,%1,%0" : "=x"(u.vdata[0]) : "%x"(u.vdata[0]), "x"(src0.u.vdata[0]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvf4x1>&
	varrayT<TYPEvf4x1>::operator&=(const varrayT<TYPEvf4x1>& src0)
	{
		asm("vandps %2,%1,%0" : "=x"(u.vdata[0]) : "%x"(u.vdata[0]), "x"(src0.u.vdata[0]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvf4x1>&
	varrayT<TYPEvf4x1>::operator^=(const varrayT<TYPEvf4x1>& src0)
	{
		asm("vxorps %2,%1,%0" : "=x"(u.vdata[0]) : "%x"(u.vdata[0]), "x"(src0.u.vdata[0]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvf8x1>&
	varrayT<TYPEvf8x1>::operator|=(const varrayT<TYPEvf8x1>& src0)
	{
		asm("vorps %2,%1,%0" : "=&x"(u.vdata[0]) : "%x"(u.vdata[0]), "x"(src0.u.vdata[0]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvf8x1>&
	varrayT<TYPEvf8x1>::operator&=(const varrayT<TYPEvf8x1>& src0)
	{
		asm("vandps %2,%1,%0" : "=&x"(u.vdata[0]) : "%x"(u.vdata[0]), "x"(src0.u.vdata[0]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvf8x1>&
	varrayT<TYPEvf8x1>::operator^=(const varrayT<TYPEvf8x1>& src0)
	{
		asm("vxorps %2,%1,%0" : "=&x"(u.vdata[0]) : "%x"(u.vdata[0]), "x"(src0.u.vdata[0]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvd4x1>&
	varrayT<TYPEvd4x1>::operator|=(const varrayT<TYPEvd4x1>& src0)
	{
		asm("vorpd %2,%1,%0" : "=&x"(u.vdata[0]) : "%x"(u.vdata[0]), "xm"(src0.u.vdata[0]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvd4x1>&
	varrayT<TYPEvd4x1>::operator&=(const varrayT<TYPEvd4x1>& src0)
	{
		asm("vandpd %2,%1,%0" : "=&x"(u.vdata[0]) : "%x"(u.vdata[0]), "xm"(src0.u.vdata[0]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvd4x1>&
	varrayT<TYPEvd4x1>::operator^=(const varrayT<TYPEvd4x1>& src0)
	{
		asm("vxorpd %2,%1,%0" : "=&x"(u.vdata[0]) : "%x"(u.vdata[0]), "xm"(src0.u.vdata[0]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvd8x1>&
	varrayT<TYPEvd8x1>::operator|=(const varrayT<TYPEvd8x1>& src0)
	{
		asm("vorpd %2,%1,%0" : "=&x"(u.vdata2[0]) : "%x"(u.vdata2[0]), "xm"(src0.u.vdata2[0]));
		asm("vorpd %2,%1,%0" : "=&x"(u.vdata2[1]) : "%x"(u.vdata2[1]), "xm"(src0.u.vdata2[1]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvd8x1>&
	varrayT<TYPEvd8x1>::operator&=(const varrayT<TYPEvd8x1>& src0)
	{
		asm("vandpd %2,%1,%0" : "=&x"(u.vdata2[0]) : "%x"(u.vdata2[0]), "xm"(src0.u.vdata2[0]));
		asm("vandpd %2,%1,%0" : "=&x"(u.vdata2[1]) : "%x"(u.vdata2[1]), "xm"(src0.u.vdata2[1]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvd8x1>&
	varrayT<TYPEvd8x1>::operator^=(const varrayT<TYPEvd8x1>& src0)
	{
		asm("vxorpd %2,%1,%0" : "=&x"(u.vdata2[0]) : "%x"(u.vdata2[0]), "xm"(src0.u.vdata2[0]));
		asm("vxorpd %2,%1,%0" : "=&x"(u.vdata2[1]) : "%x"(u.vdata2[1]), "xm"(src0.u.vdata2[1]));
		return *this;
	}

//
// Fix this -- three args doesn't work on SSE2 - simple fix, just haven't done so yet.
//
#elif defined(__SSE2__)

	template<>
	inline varrayT<TYPEvf4x1>&
	varrayT<TYPEvf4x1>::operator|=(const varrayT<TYPEvf4x1>& src0)
	{
		asm("orps %1,%0,%0" : "=&x"(u.vdata[0]) : "x"(src0.u.vdata[0]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvf4x1>&
	varrayT<TYPEvf4x1>::operator&=(const varrayT<TYPEvf4x1>& src0)
	{
		asm("andps %1,%0,%0" : "=&x"(u.vdata[0]) : "x"(src0.u.vdata[0]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvf4x1>&
	varrayT<TYPEvf4x1>::operator^=(const varrayT<TYPEvf4x1>& src0)
	{
		asm("xorps %1,%0,%0" : "=&x"(u.vdata[0]) : "x"(src0.u.vdata[0]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvf8x1>&
	varrayT<TYPEvf8x1>::operator|=(const varrayT<TYPEvf8x1>& src0)
	{
		asm("orps %1,%0,%0" : "=&x"(u.vdata2[0]) : "x"(src0.u.vdata2[0]));
		asm("orps %1,%0,%0" : "=&x"(u.vdata2[1]) : "x"(src0.u.vdata2[2]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvf8x1>&
	varrayT<TYPEvf8x1>::operator&=(const varrayT<TYPEvf8x1>& src0)
	{
		asm("andps %1,%0,%0" : "=&x"(u.vdata2[0]) : "x"(src0.u.vdata2[0]));
		asm("andps %1,%0,%0" : "=&x"(u.vdata2[1]) : "x"(src0.u.vdata2[1]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvf8x1>&
	varrayT<TYPEvf8x1>::operator^=(const varrayT<TYPEvf8x1>& src0)
	{
		asm("xorps %1,%0,%0" : "=&x"(u.vdata2[0]) : "x"(src0.u.vdata2[0]));
		asm("xorps %1,%0,%0" : "=&x"(u.vdata2[1]) : "x"(src0.u.vdata2[1]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvd4x1>&
	varrayT<TYPEvd4x1>::operator|=(const varrayT<TYPEvd4x1>& src0)
	{
		asm("orpd %1,%0,%0" : "=&x"(u.vdata4[0]) : "x"(src0.u.vdata4[0]));
		asm("orpd %1,%0,%0" : "=&x"(u.vdata4[1]) : "x"(src0.u.vdata4[1]));
		asm("orpd %1,%0,%0" : "=&x"(u.vdata4[2]) : "x"(src0.u.vdata4[2]));
		asm("orpd %1,%0,%0" : "=&x"(u.vdata4[3]) : "x"(src0.u.vdata4[3]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvd4x1>&
	varrayT<TYPEvd4x1>::operator&=(const varrayT<TYPEvd4x1>& src0)
	{
		asm("andpd %1,%0,%0" : "=&x"(u.vdata4[0]) : "x"(src0.u.vdata4[0]));
		asm("andpd %1,%0,%0" : "=&x"(u.vdata4[1]) : "x"(src0.u.vdata4[1]));
		asm("andpd %1,%0,%0" : "=&x"(u.vdata4[2]) : "x"(src0.u.vdata4[2]));
		asm("andpd %1,%0,%0" : "=&x"(u.vdata4[3]) : "x"(src0.u.vdata4[3]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvd4x1>&
	varrayT<TYPEvd4x1>::operator^=(const varrayT<TYPEvd4x1>& src0)
	{
		asm("xorpd %1,%0,%0" : "=&x"(u.vdata4[0]) : "x"(src0.u.vdata4[0]));
		asm("xorpd %1,%0,%0" : "=&x"(u.vdata4[1]) : "x"(src0.u.vdata4[1]));
		asm("xorpd %1,%0,%0" : "=&x"(u.vdata4[2]) : "x"(src0.u.vdata4[2]));
		asm("xorpd %1,%0,%0" : "=&x"(u.vdata4[3]) : "x"(src0.u.vdata4[3]));
		return *this;
	}

	template<>
	inline varrayT<TYPEvd8x1>&
	varrayT<TYPEvd8x1>::operator|=(const varrayT<TYPEvd8x1>& src0)
	{
		asm("orpd %1,%0,%0" : "=&x"(u.vdata4[0]) : "x"(src0.u.vdata4[0]));
		asm("orpd %1,%0,%0" : "=&x"(u.vdata4[1]) : "x"(src0.u.vdata4[1]));
		asm("orpd %1,%0,%0" : "=&x"(u.vdata4[2]) : "x"(src0.u.vdata4[2]));
		asm("orpd %1,%0,%0" : "=&x"(u.vdata4[3]) : "x"(src0.u.vdata4[3]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvd8x1>&
	varrayT<TYPEvd8x1>::operator&=(const varrayT<TYPEvd8x1>& src0)
	{
		asm("andpd %1,%0,%0" : "=&x"(u.vdata4[0]) : "x"(src0.u.vdata4[0]));
		asm("andpd %1,%0,%0" : "=&x"(u.vdata4[1]) : "x"(src0.u.vdata4[1]));
		asm("andpd %1,%0,%0" : "=&x"(u.vdata4[2]) : "x"(src0.u.vdata4[2]));
		asm("andpd %1,%0,%0" : "=&x"(u.vdata4[3]) : "x"(src0.u.vdata4[3]));
		return *this;
	}
	template<>
	inline varrayT<TYPEvd8x1>&
	varrayT<TYPEvd8x1>::operator^=(const varrayT<TYPEvd8x1>& src0)
	{
		asm("xorpd %1,%0,%0" : "=&x"(u.vdata4[0]) : "x"(src0.u.vdata4[0]));
		asm("xorpd %1,%0,%0" : "=&x"(u.vdata4[1]) : "x"(src0.u.vdata4[1]));
		asm("xorpd %1,%0,%0" : "=&x"(u.vdata4[2]) : "x"(src0.u.vdata4[2]));
		asm("xorpd %1,%0,%0" : "=&x"(u.vdata4[3]) : "x"(src0.u.vdata4[3]));
		return *this;
	}
#endif

template <typename VDEF>
varrayT<VDEF> operator+(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVval(i, (src0.Vval(i) + src1.Vval(i)));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> operator-(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVval(i, (src0.Vval(i) - src1.Vval(i)));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> operator*(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVval(i, (src0.Vval(i) * src1.Vval(i)));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> operator/(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVval(i, (src0.Vval(i) / src1.Vval(i)));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> operator*(const varrayT<VDEF>& src0, typename VDEF::FT src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVval(i, (src0.Vval(i) * src1));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> operator/(const varrayT<VDEF>& src0, typename VDEF::FT src1)
{
	typename VDEF::FT newsrc1 = typename VDEF::FT(1.0) / src1;
	return src0 * newsrc1;
}

template <typename VDEF>
varrayT<VDEF> operator&(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.u.vdatai[i] = src0.u.vdatai[i] & src1.u.vdatai[i];

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> operator|(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.u.vdatai[i] = src0.u.vdatai[i] | src1.u.vdatai[i];

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> operator^(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.u.vdatai[i] = src0.u.vdatai[i] ^ src1.u.vdatai[i];

	return rtval;
}

#if !defined(__AVX__)
template <typename VDEF>
inline varrayT<VDEF> operator!=(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < varrayT<VDEF>::nFT; i++)
		rtval.setSvali(i, (src0.Sval(i) != src1.Sval(i)) ? typename VDEF::IT(~0) : typename VDEF::IT(0));

	return rtval;
}

template <typename VDEF>
inline varrayT<VDEF> operator==(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < varrayT<VDEF>::nFT; i++)
		rtval.setSvali(i, (src0.Sval(i) == src1.Sval(i)) ? typename VDEF::IT(~0) : typename VDEF::IT(0));

	return rtval;
}
#endif

template <typename VDEF>
varrayT<VDEF> Vgt(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < varrayT<VDEF>::nFT; i++)
		rtval.setSvali(i, (src0.Sval(i) > src1.Sval(i)) ? typename VDEF::IT(~0) : typename VDEF::IT(0));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> operator>(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < varrayT<VDEF>::nFT; i++)
		rtval.setSvali(i, (src0.Sval(i) > src1.Sval(i)) ? typename VDEF::IT(~0) : typename VDEF::IT(0));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> Vge(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < varrayT<VDEF>::nFT; i++)
		//rtval.setSvali(i, (src0.Sval(i) >= src1.Sval(i)));
		rtval.setSvali(i, (src0.Sval(i) >= src1.Sval(i)) ? typename VDEF::IT(~0) : typename VDEF::IT(0));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> Vlt(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < varrayT<VDEF>::nFT; i++)
		//rtval.setSvali(i, (src0.Sval(i) < src1.Sval(i)));
		rtval.setSvali(i, (src0.Sval(i) < src1.Sval(i)) ? typename VDEF::IT(~0) : typename VDEF::IT(0));

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> Vle(const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < varrayT<VDEF>::nFT; i++)
		//rtval.setSvali(i, (src0.Sval(i) <= src1.Sval(i)));
		rtval.setSvali(i, (src0.Sval(i) <= src1.Sval(i)) ? typename VDEF::IT(~0) : typename VDEF::IT(0));

	return rtval;
}

template <typename VDEF>
bool
alltrue(const varrayT<VDEF>& src)
{
	for (int i = 0; i < varrayT<VDEF>::nFT; i++)
	{
		if (src.u.sdatai[i] != ~(typename varrayT<VDEF>::IT(0)))
			return false;
	}
	return true;
}


//
// The Vselect routine is  optimized specifically for each compiler.
// The clang compiler doesn't seem to be able to deal with the direct use of the vectors as well as GCC, and
// produces a syntax error when using the GCC technique below.
// 
// GCC produces very poor vector code on the per element method needed by clang.
//
#if defined(__clang__)

template <typename VDEF>
varrayT<VDEF> Vselect(const varrayT<VDEF>& isel, const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
	{
		typedef typename VDEF::FT FT;
		typedef typename VDEF::VFT VFT;
		typedef typename VDEF::VIT VIT;
		static constexpr TYPEuint nFTperVFT = VDEF::nFTperVFT;

		VFT vt0{src0.Vval(i)};
		VFT vt1{src1.Vval(i)};
		VIT is{isel.Vvali(i)};
		VFT rt{FT(0.0)};
		for (TYPEuint j = 0; j < nFTperVFT; j++)
		{
			rt[j] = is[j] ? vt0[j] : vt1[j];
		}
		rtval.setVval(i, rt);
	}

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> Vselect(IvarrayT<VDEF>& isel, const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
	{
		typedef typename VDEF::FT FT;
		typedef typename VDEF::VFT VFT;
		typedef typename VDEF::VIT VIT;
		static constexpr TYPEuint nFTperVFT = VDEF::nFTperVFT;

		VFT vt0{src0.Vval(i)};
		VFT vt1{src1.Vval(i)};
		VIT is{isel.Vval(i)};
		VFT rt{FT(0.0)};
		for (TYPEuint j = 0; j < nFTperVFT; j++)
		{
			rt[j] = is[j] ? vt0[j] : vt1[j];
		}
		rtval.setVval(i, rt);
	}

	return rtval;
}

#else
template <typename VDEF>
varrayT<VDEF> Vselect(const varrayT<VDEF>& isel, const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;

	for (TYPEuint i = 0; i < VDEF::NVT; i++)
	{
		rtval.setVval(i, isel.Vvali(i) ? src0.Vval(i) : src1.Vval(i));
	}

	return rtval;
}

template <typename VDEF>
varrayT<VDEF> Vselect(const IvarrayT<VDEF>& isel, const varrayT<VDEF>& src0, const varrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;

	for (TYPEuint i = 0; i < VDEF::NVT; i++)
	{
		rtval.setVval(i, isel.Vvali(i) ? src0.Vval(i) : src1.Vval(i));
	}

	return rtval;
}

#endif

namespace varray
{

	template <typename DT, typename ST=typeof DT{}[0]>
	inline DT
	vasqrt(DT val, ST unused)
	{
		DT rtval;
		static constexpr int vs = sizeof(DT)/sizeof(DT[0]);
		for (int i = 0; i < vs; i++)
		{
			rtval[i] = ::sqrt((double) val[i]);
		}
		return rtval;
	}

//
// if an unsupported vsqrt type is called (the routine outsid of varray that were are
// calling, then we'll try to do the sqrt by calling the default sqrt routine.  If
// the sqrt is not supported, then the error message pointing to this function can
// give a hint as to what additional needs to be supported...
//
	template <typename DT>
	inline DT
	vasqrt(DT val)
	{
		return sqrt(val);
	}

	template <>
	inline float
	vasqrt<float>(float val)
	{
		return ::sqrtf(val);
	}

	template <>
	inline double
	vasqrt<double>(double val)
	{
		return ::sqrt(val);
	}

	template <>
	inline long double
	vasqrt<long double>(long double val)
	{
		return ::sqrtl(val);
	}
};

//
// These two following templates named 'vsqrt' play a template trick.
// The first template is actually requesting a type similar to
// 'varrayT<VDEF>', so that we can infer the underlying floating
// type.  The first template is operating on a vector and not a
// scalar type, so I explicitly look for something that works
// with a VDEF template argument (which is a way to tell between
// a varray vector and a normal scalar (or even compiler supported
// vector!!!)
//
// The way that I make sure that VDEF is a 'varray' or 'Ivarray'
// is a type specifier for varray or Ivarray, I make the 2nd
// template argument a default that is one of the underlying
// definitions in VDEF (in this case it is 'FT', which I am
// using anyway.  If the inferred type isn't a template type that
// has an 'FT' memeber defintion, then this template is rejected,
// and the next template takes effect.  (According to C++ standard,
// if you have an error in the specification of a
// template -- including return value, then the template is rejected
// without an error indication.  Another template is then looked for
// to try to fit.)
//
// This is also a tricky way of pre-selecting whether or not
// the type of the data item is a vector or a scalar.
//
template <typename VDEF, typename FT = typename VDEF::FT>
inline varrayT<VDEF>
vsqrt(const varrayT<VDEF>& src)
{
	varrayT<VDEF> rtval;
	for (int i = 0; i < VDEF::nFT; i++)
	{
		rtval[i] = varray::vasqrt<FT>(src[i]);
	}
	return rtval;
}

//
// This is the simpler direct template that assumes that the
// type is a scalar or naturally supported data type.
//
template <typename VDEF>
inline VDEF
vsqrt(VDEF src)
{
	return varray::vasqrt<VDEF>(src);
}

#if defined(__AVX__)
template<>
varrayT<TYPEvf8x1>
inline vsqrt(varrayT<TYPEvf8x1> src)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vsqrtps %1,%0" : "=x"(rtval.u.vdata[0]) : "x"(src.u.vdata[0]));
	return rtval;
}

template<>
varrayT<TYPEvf4x1>
inline vsqrt(varrayT<TYPEvf4x1> src)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vsqrtps %1,%0" : "=x"(rtval.u.vdata[0]) : "x"(src.u.vdata[0]));
	return rtval;
}

template<>
varrayT<TYPEvd4x1>
inline vsqrt(varrayT<TYPEvd4x1> src)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vsqrtpd %1,%0" : "=x"(rtval.u.vdata[0]) : "x"(src.u.vdata[0]));
	return rtval;
}

template<>
varrayT<TYPEvd8x1>
inline vsqrt(const varrayT<TYPEvd8x1>& src)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vsqrtpd %1,%0" : "=x"(rtval.u.vdata2[0]) : "x"(src.u.vdata2[0]));
	asm("vsqrtpd %1,%0" : "=x"(rtval.u.vdata2[1]) : "x"(src.u.vdata2[1]));
	return rtval;
}

#elif defined(__SSE3__)
#if 1
template<>
varrayT<TYPEvf4x1>
inline vsqrt(varrayT<TYPEvf4x1>src)
{
	varrayT<TYPEvf4x1> rtval;
	asm("sqrtps %1,%0" : "=x"(rtval.u.vdata[0]) : "x"(src.u.vdata[0]));
	return rtval;
}
template<>
varrayT<TYPEvf8x1>
inline vsqrt(const varrayT<TYPEvf8x1>& src)
{
	varrayT<TYPEvf8x1> rtval;
	asm("sqrtps %1,%0" : "=x"(rtval.u.vdata2[0]) : "x"(src.u.vdata2[0]));
	asm("sqrtps %1,%0" : "=x"(rtval.u.vdata2[1]) : "x"(src.u.vdata2[1]));
	return rtval;
}
#endif

#endif


#if defined(__AVX__)

varrayT<TYPEvf4x1>
inline operator==(varrayT<TYPEvf4x1> src0, varrayT<TYPEvf4x1> src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vcmpeqps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src1.u.vdata[0]), "x"(src0.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvf4x1>
inline operator!=(varrayT<TYPEvf4x1> src0, varrayT<TYPEvf4x1> src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vcmpeqps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src1.u.vdata[0]), "x"(src0.u.vdata[0]));
	rtval.u.vdatai[0] = ~rtval.u.vdatai[0];
	return rtval;
}

varrayT<TYPEvd4x1>
inline operator==(varrayT<TYPEvd4x1> src0, varrayT<TYPEvd4x1> src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vcmpeqpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src1.u.vdata[0]), "x"(src0.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvd4x1>
inline operator!=(varrayT<TYPEvd4x1> src0, varrayT<TYPEvd4x1> src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vcmpeqpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src1.u.vdata[0]), "x"(src0.u.vdata[0]));
	rtval.u.vdatai[0] = ~rtval.u.vdatai[0];
	return rtval;
}

varrayT<TYPEvf8x1>
inline operator!=(varrayT<TYPEvf8x1> src0, varrayT<TYPEvf8x1> src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vcmpneps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src1.u.vdata[0]), "x"(src0.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvf8x1>
inline operator==(varrayT<TYPEvf8x1> src0, varrayT<TYPEvf8x1> src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vcmpeqps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src1.u.vdata[0]), "x"(src0.u.vdata[0]));
	return rtval;
}

IvarrayT<TYPEvf8x1>
inline operator!=(IvarrayT<TYPEvf8x1> src0, IvarrayT<TYPEvf8x1> src1)
{
	IvarrayT<TYPEvf8x1> rtval;
	asm("vpcmpeqs %2,%1,%0" : "=x"(rtval.u.vdatai[0]) : "%x"(src1.u.vdatai[0]), "x"(src0.u.vdatai[0]));
	rtval.u.vdatai[0] = ~rtval.u.vdatai[0];
	return rtval;
}

IvarrayT<TYPEvf8x1>
inline operator==(IvarrayT<TYPEvf8x1> src0, IvarrayT<TYPEvf8x1> src1)
{
	IvarrayT<TYPEvf8x1> rtval;
	asm("vpcmpeqs %2,%1,%0" : "=x"(rtval.u.vdatai[0]) : "%x"(src1.u.vdatai[0]), "x"(src0.u.vdatai[0]));
	return rtval;
}

IvarrayT<TYPEvf8x1>
inline operator&(IvarrayT<TYPEvf8x1> src0, IvarrayT<TYPEvf8x1> src1)
{
	IvarrayT<TYPEvf8x1> rtval;
	asm("vpand %2,%1,%0" : "=x"(rtval.u.vdatai[0]) : "%x"(src1.u.vdatai[0]), "x"(src0.u.vdatai[0]));
	return rtval;
}

IvarrayT<TYPEvf8x1>
inline operator|(IvarrayT<TYPEvf8x1> src0, IvarrayT<TYPEvf8x1> src1)
{
	IvarrayT<TYPEvf8x1> rtval;
	asm("vpor %2,%1,%0" : "=x"(rtval.u.vdatai[0]) : "%x"(src1.u.vdatai[0]), "x"(src0.u.vdatai[0]));
	return rtval;
}

IvarrayT<TYPEvf8x1>
inline operator^(IvarrayT<TYPEvf8x1> src0, IvarrayT<TYPEvf8x1> src1)
{
	IvarrayT<TYPEvf8x1> rtval;
	asm("vpxor %2,%1,%0" : "=x"(rtval.u.vdatai[0]) : "%x"(src1.u.vdatai[0]), "x"(src0.u.vdatai[0]));
	return rtval;
}


varrayT<TYPEvf4x1>
inline	operator|(varrayT<TYPEvf4x1> src0, varrayT<TYPEvf4x1> src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vorps %2,%1,%0" : "=x"(rtval) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvf4x1>
inline	operator&(varrayT<TYPEvf4x1> src0, varrayT<TYPEvf4x1> src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vandps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvf4x1>
inline	operator^(varrayT<TYPEvf4x1> src0, varrayT<TYPEvf4x1> src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vxorps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvf8x1>
inline	operator|(varrayT<TYPEvf8x1> src0, varrayT<TYPEvf8x1> src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vorps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvf8x1>
inline	operator&(varrayT<TYPEvf8x1> src0, varrayT<TYPEvf8x1> src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vandps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvf8x1>
inline	operator^(varrayT<TYPEvf8x1> src0, varrayT<TYPEvf8x1> src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vxorps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvd4x1>
inline	operator|(varrayT<TYPEvd4x1> src0, varrayT<TYPEvd4x1> src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vorpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvd4x1>
inline	operator&(varrayT<TYPEvd4x1> src0, varrayT<TYPEvd4x1> src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vandpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvd4x1>
inline	operator^(varrayT<TYPEvd4x1> src0, varrayT<TYPEvd4x1> src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vxorpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

varrayT<TYPEvd8x1>
inline	operator|(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vorpd %2,%1,%0" : "=x"(rtval.u.vdata2[0]) : "%x"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("vorpd %2,%1,%0" : "=x"(rtval.u.vdata2[1]) : "%x"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

varrayT<TYPEvd8x1>
inline	operator&(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vandpd %2,%1,%0" : "=x"(rtval.u.vdata2[0]) : "%x"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("vandpd %2,%1,%0" : "=x"(rtval.u.vdata2[1]) : "%x"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

varrayT<TYPEvd8x1>
inline	operator^(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vxorpd %2,%1,%0" : "=x"(rtval.u.vdata2[0]) : "%x"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("vxorpd %2,%1,%0" : "=x"(rtval.u.vdata2[1]) : "%x"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

inline varrayT<TYPEvf8x1>
Vgt(varrayT<TYPEvf8x1> src0, varrayT<TYPEvf8x1> src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vcmpgtps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

inline varrayT<TYPEvf8x1>
Vlt(varrayT<TYPEvf8x1> src0, varrayT<TYPEvf8x1> src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vcmpltps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

template<>
inline varrayT<TYPEvf8x1>
Vselect(const varrayT<TYPEvf8x1>& isel, const varrayT<TYPEvf8x1>& src0, const varrayT<TYPEvf8x1>& src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vblendvps %1, %2, %3, %0" : "=x"(rtval.u.vdata[0]) : "x" ( isel.u.vdata[0]), "x" (src0.u.vdata[0]), "x" (src1.u.vdata[0]));
	return rtval;
}

template<>
inline varrayT<TYPEvf4x1>
Vselect(const varrayT<TYPEvf4x1>& isel, const varrayT<TYPEvf4x1>& src0, const varrayT<TYPEvf4x1>& src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vblendvps %1, %2, %3, %0" : "=x"(rtval.u.vdata[0]) : "x" ( isel.u.vdata[0]), "x" (src0.u.vdata[0]), "x" (src1.u.vdata[0]));
	return rtval;
}

inline varrayT<TYPEvd4x1>
Vgt(varrayT<TYPEvd4x1> src0, varrayT<TYPEvd4x1> src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vcmpgtpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

inline varrayT<TYPEvd8x1>
Vgt(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vcmpgtpd %2,%1,%0" : "=x"(rtval.u.vdata2[0]) : "x"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("vcmpgtpd %2,%1,%0" : "=x"(rtval.u.vdata2[1]) : "x"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

inline varrayT<TYPEvd4x1>
Vlt(varrayT<TYPEvd4x1> src0, varrayT<TYPEvd4x1> src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vcmpltpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

inline varrayT<TYPEvd8x1>
Vlt(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vcmpltpd %2,%1,%0" : "=x"(rtval.u.vdata2[0]) : "x"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("vcmpltpd %2,%1,%0" : "=x"(rtval.u.vdata2[1]) : "x"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

template<>
inline varrayT<TYPEvd4x1>
Vselect(const varrayT<TYPEvd4x1>& isel, const varrayT<TYPEvd4x1>& src0, const varrayT<TYPEvd4x1>& src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vblendvpd %1, %2, %3, %0" : "=x"(rtval.u.vdata[0]) : "x" ( isel.u.vdata[0]), "x" (src0.u.vdata[0]), "x" (src1.u.vdata[0]));
	return rtval;
}

template<>
inline varrayT<TYPEvd8x1>
Vselect(const varrayT<TYPEvd8x1>& isel, const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vblendvpd %1, %2, %3, %0" : "=x"(rtval.u.vdata2[0]) : "x" ( isel.u.vdata2[0]), "x" (src0.u.vdata2[0]), "x" (src1.u.vdata2[0]));
	asm("vblendvpd %1, %2, %3, %0" : "=x"(rtval.u.vdata2[1]) : "x" ( isel.u.vdata2[1]), "x" (src0.u.vdata2[1]), "x" (src1.u.vdata2[1]));
	return rtval;
}

#endif

//
// Here we have a 'Vmax' and 'Vmin' that work with general variables also --
// mostly for semantic source compatibility.
//
template <typename T0, typename T1>
T0
Vmax(T0 src0, T1 src1)
{
	return (src0 > src1) ? src0 : src1;
}

template <typename T0, typename T1>
T0
Vmin(T0 src0, T1 src1)
{
	return (src0 < src1) ? src0 : src1;
}


template < typename VDEF, template <typename > class varrayX >
varrayX<VDEF>
Vmax(const varrayX<VDEF>& src0, const varrayX<VDEF>& src1)
{
	return Vselect(Vgt(src0, src1), src0, src1);
}

template < typename VDEF, template <typename > class varrayX >
varrayX<VDEF>
Vmin(const varrayX<VDEF>& src0, const varrayX<VDEF>& src1)
{
	return Vselect(Vlt(src0, src1), src0, src1);
}


#if defined(__AVX__)

template <>
inline varrayT<TYPEvd8x1>
Vmin<>(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vminpd %2,%1,%0" : "=x"(rtval.u.vdata2[0]) : "%x"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("vminpd %2,%1,%0" : "=x"(rtval.u.vdata2[1]) : "%x"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

template <>
inline varrayT<TYPEvd8x1>
Vmax<>(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	asm("vmaxpd %2,%1,%0" : "=x"(rtval.u.vdata2[0]) : "%x"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("vmaxpd %2,%1,%0" : "=x"(rtval.u.vdata2[1]) : "%x"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

template <>
inline varrayT<TYPEvf8x1>
Vmax<>(const varrayT<TYPEvf8x1>& src0, const varrayT<TYPEvf8x1>& src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vmaxps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

template <>
inline varrayT<TYPEvf8x1>
Vmin<>(const varrayT<TYPEvf8x1>& src0, const varrayT<TYPEvf8x1>& src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("vminps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

template <>
inline varrayT<TYPEvf4x1>
Vmax<>(const varrayT<TYPEvf4x1>& src0, const varrayT<TYPEvf4x1>& src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vmaxps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

template <>
inline varrayT<TYPEvf4x1>
Vmin<>(const varrayT<TYPEvf4x1>& src0, const varrayT<TYPEvf4x1>& src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("vminps %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

template <>
inline varrayT<TYPEvd4x1>
Vmax<>(const varrayT<TYPEvd4x1>& src0, const varrayT<TYPEvd4x1>& src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vmaxpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

template <>
inline varrayT<TYPEvd4x1>
Vmin<>(const varrayT<TYPEvd4x1>& src0, const varrayT<TYPEvd4x1>& src1)
{
	varrayT<TYPEvd4x1> rtval;
	asm("vminpd %2,%1,%0" : "=x"(rtval.u.vdata[0]) : "%x"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}

#elif defined(__clang__) && defined(__SSE3__)
template <>
inline varrayT<TYPEvf4x1>
Vmin<>(varrayT<TYPEvf4x1> src0, varrayT<TYPEvf4x1> src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("minps %2,%0" : "=x"(rtval.u.vdata[0]) : "%0"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}
template <>
inline varrayT<TYPEvf8x1>
Vmin<>(const varrayT<TYPEvf8x1>& src0, const varrayT<TYPEvf8x1>& src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("minps %2,%0" : "=x"(rtval.u.vdata2[0]) : "%0"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("minps %2,%0" : "=x"(rtval.u.vdata2[1]) : "%0"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

template <>
inline varrayT<TYPEvf4x1>
Vmax<>(varrayT<TYPEvf4x1> src0, varrayT<TYPEvf4x1> src1)
{
	varrayT<TYPEvf4x1> rtval;
	asm("maxps %2,%0" : "=x"(rtval.u.vdata[0]) : "%0"(src0.u.vdata[0]), "x"(src1.u.vdata[0]));
	return rtval;
}
template <>
inline varrayT<TYPEvf8x1>
Vmax<>(const varrayT<TYPEvf8x1>& src0, const varrayT<TYPEvf8x1>& src1)
{
	varrayT<TYPEvf8x1> rtval;
	asm("maxps %2,%0" : "=x"(rtval.u.vdata2[0]) : "%0"(src0.u.vdata2[0]), "x"(src1.u.vdata2[0]));
	asm("maxps %2,%0" : "=x"(rtval.u.vdata2[1]) : "%0"(src0.u.vdata2[1]), "x"(src1.u.vdata2[1]));
	return rtval;
}

template <>
inline varrayT<TYPEvd8x1>
Vmax<>(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	for (int i = 0; i < TYPEvd8x1::nFT; i++)
	{
		rtval[i] = (src0[i] > src1[i]) ? src0[i] : src1[i];
	}
	return rtval;
}

template <>
inline varrayT<TYPEvd8x1>
Vmin<>(const varrayT<TYPEvd8x1>& src0, const varrayT<TYPEvd8x1>& src1)
{
	varrayT<TYPEvd8x1> rtval;
	for (int i = 0; i < TYPEvd8x1::nFT; i++)
	{
		rtval[i] = (src0[i] < src1[i]) ? src0[i] : src1[i];
	}
	return rtval;
}

#endif

//
// Below are the external Ivarray functions
//

#if 0
template <typename VDEF>
IvarrayT<VDEF> operator~(const IvarrayT<VDEF>& src0)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::nIT; i++)
		rtval.setSvali(i, ~(src0.Svali(i)));

	return rtval;
}
#endif

template <typename VDEF>
IvarrayT<VDEF> operator+(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) + src1.Vvali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator-(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) - src1.Vvali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator*(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) * src1.Vvali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator/(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) / src1.Vvali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator*(const IvarrayT<VDEF>& src0, typename VDEF::IT src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) * src1));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator&(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) & src1.Vvali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator|(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) | src1.Vvali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator^(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) ^ src1.Vvali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator<<(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) << src1.Vvali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator>>(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
		rtval.setVvali(i, (src0.Vvali(i) >> src1.Vvali(i)));

	return rtval;
}

#if !defined(__AVX__)
template <typename VDEF>
IvarrayT<VDEF> operator!=(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::nIT; i++)
		rtval.setSvali(i, src0.Svali(i) != src1.Svali(i));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> operator==(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::nIT; i++)
		rtval.setSvali(i, src0.Svali(i) == src1.Svali(i));

	return rtval;
}
#endif

template <typename VDEF>
IvarrayT<VDEF> Vgt(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::nIT; i++)
		rtval.setSvali(i, src0.Svali(i) > src1.Svali(i));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> Vge(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::nIT; i++)
		rtval.setSvali(i, (src0.Svali(i) >= src1.Svali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> Vlt(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::nIT; i++)
		rtval.setSvali(i, (src0.Svali(i) < src1.Svali(i)));

	return rtval;
}

template <typename VDEF>
IvarrayT<VDEF> Vle(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	IvarrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::nIT; i++)
		rtval.setSvali(i, (src0.Svali(i) <= src1.Svali(i)));

	return rtval;
}

#if 0
varrayT<TYPEvf8x1> operator= (varrayT<TYPEvf4x1>& src0, varrayT<TYPEvf4x1>& src1)
{
	varrayT<TYPEvf8x1> rtval;
	rtval = varrayT<TYPEvf8x1>({
		src0.sval(0), src0,sval(1), src0.sval(2), src0.sval3(),
		src1.sval(0), src1,sval(1), src1.sval(2), src1.sval3()
	});
	return rtval;
}
#endif

//
// The Vselect routine is  optimized specifically for each compiler.
// The clang compiler doesn't seem to be able to deal with the direct use of the vectors as well as GCC, and
// produces a syntax error when using the GCC technique below.
// 
// GCC produces very poor vector code on the per element method needed by clang.
//
#if defined(__clang__)
template <typename VDEF, typename VRES>
IvarrayT<VRES> Vselect(const IvarrayT<VDEF>& isel, const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;
	for (TYPEuint i = 0; i < VDEF::NVT; i++)
	{
		typedef typename VDEF::VIT VIT;
		static constexpr TYPEuint nITperVIT = VDEF::nITperVIT;

		VIT vt0{src0.Vvali(i)};
		VIT vt1{src1.Vvali(i)};
		VIT is{isel.Vvali(i)};
		VIT rt(0);
		for (TYPEuint j = 0; j < nITperVIT; j++)
		{
			rt[j] = is[j] ? vt0[j] : vt1[j];
		}
		rtval.setVvali(i, rt);
	}

	return rtval;
}

#else
template <typename VDEF>
IvarrayT<VDEF> Vselect(const IvarrayT<VDEF>& isel, const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	typedef unsigned int TYPEuint;
	varrayT <VDEF> rtval;

	for (TYPEuint i = 0; i < VDEF::NVT; i++)
	{
		typedef typename VDEF::VIT VIT;

		VIT vt0{src0.Vvali(i)};
		VIT vt1{src1.Vvali(i)};
		VIT is{isel.Vvali(i)};

		rtval.setVvali(i, is ? vt0 : vt1);
	}

	return rtval;
}

#endif

template <typename VDEF>
IvarrayT<VDEF> 
Vmax(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	return Vselect(Vgt(src0, src1), src0, src1);
}

template <typename VDEF>
IvarrayT<VDEF> 
Vmin(const IvarrayT<VDEF>& src0, const IvarrayT<VDEF>& src1)
{
	return Vselect(Vlt(src0, src1), src0, src1);
}


#if 0
class sigvop
{
	typedef float (*afcnff)(float);

	const afcnff fcn;

public:
	sigvop(afcnff op) : fcn(op) {}

	varrayT
	operator()(const varrayT& src)
	{
		varrayT rtval;
		for (int i = 0; i < SIGARRAYmaxidx; i++)
		{
			rtval[i] = (*fcn)(src[i]);
		}
		return rtval;
	}
};

class sigvopus
{
	typedef float (*afcnff)(float);

	const afcnff fcn;

public:
	sigvopus(afcnff op) : fcn(op) {}

	varrayT
	operator()(const varrayT& src)
	{
		varrayT rtval;
		for (int i = 0; i < NUMASUSED; i++)
		{
			rtval[i] = (*fcn)(src[i]);
		}
		return rtval;
	}
};
#endif


#endif



//
// vim: set ts=4:
//

