#include <stdio.h>
#include <string.h>

#if !defined(offsetof)
#define offsetof __builtin_offsetof
#endif

static const int wav0_nchannels = 2;
static const int wav0_bits_per_sample = 16;

using wuint16_t = unsigned short;
using wint16_t = signed short;
using wuint32_t = unsigned int;
using wint32_t = signed int;
using wuint64_t = unsigned long;
using wint64_t = signed long;

struct __attribute__((packed)) tag
{
	char cdata[4];

	tag(void) : cdata() {}

	tag(const char *tn) : cdata()
	{
		strncpy(cdata, tn, 4);
	}
};

struct __attribute__((packed)) tagsize
{
	tag name;
	int len;

	tagsize() : name(), len() {}
};

struct rwav0 {
	tag	riffsn;
	int	rlength;
	tag	wavsn;
};

struct __attribute__((packed)) fmtstruct {
	tag fmtsn;
	int flength;
	short aformat;
	short nchannels;
	int srate;
	int byrate;
	short balign;
	short bps;
	short extra;
	short vbps;
	int channelmask;
	short format;

	void printf(FILE *fp)
	{
		if (ac.info > 0x10)
		{
			fprintf(fp,"tag: %4.4s, flength: %d, aformat: %d, nchannels: %d, srate: %d, byrate: %d, balign: %d\n",
				fmtsn.cdata, flength, aformat, nchannels, srate, byrate, balign);
			fprintf(fp,"bps: %d, extra: %d, vbps: %d, channelmask: 0x%x, format: %d\n",
				bps, extra, vbps, (unsigned) channelmask, format);
		}
	}
};

struct __attribute__((packed)) fmtstruct0 {
	tag fmtsn;
	int flength;
	short aformat;
	short nchannels;
	int srate;
	int byrate;
	short balign;
	short bps;
	short extra;
	//char extradata[22];
	short	vbps;
	int	channelmask;
	short	format;
	char	guid[14];

	void printf(FILE *fp)
	{
		if (ac.info > 0x10)
		{
			fprintf(fp,"tag: %4.4s, flength: %d, aformat: %d, nchannels: %d, srate: %d, byrate: %d, balign: %d\n",
				fmtsn.cdata, flength, aformat, nchannels, srate, byrate, balign);
			fprintf(fp,"bps: %d, extra: %d\n",
				bps, extra);
		}
	}

	void
	forcefloat(void)
	{
		aformat = 3;
		format = 3;
		bps = 32;
		vbps = 32;
		byrate = nchannels * srate * (bps / 8);
		balign = nchannels * (bps / 8);
	}
		
};

struct __attribute__((packed)) ds64raw {
	tagsize	ds64sn;
	wuint32_t	bw64sizelow;
	wuint32_t	bw64sizehigh;
	wuint32_t	datasizelow;
	wuint32_t	datasizehigh;
	wuint32_t	dummylow;
	wuint32_t	dummyhigh;
	wuint32_t	tablelength;
};

struct ds64logical {
	tagsize	ds64sn;
	long int bw64size;
	long int datasize;
};

struct __attribute__((packed)) chunksize64 {
	char	chunkid[4];
	uint32_t	chunksizelow;
	uint32_t	chunksizehigh;
};

struct __attribute__((packed)) bextraw {
	char Description[256];
	char Originator[32];
	char OriginatorReference[32];
	char OriginationDate[10];
	char OriginationTime[8];
	wuint32_t TimeReferenceLow;
	wuint32_t TimeReferenceHigh;
	wuint16_t Version;
	char UMID[64];
	wuint16_t LoudnessValue;
	wuint16_t LoudnessRange;
	wuint16_t MaxTruePeakLevel;
	wuint16_t MaxMomentaryLoudness;
	wuint16_t MaxShortTermLoudness;
	char Reserved0[180];
	char CodingHistory[1];
};

	
struct __attribute__((packed)) factraw {
	tag factsn;
	int factlen;
	int factdata;
};


struct __attribute__((packed)) wavh {

	tag	riffsn;
	int	rlength;
	tag	wavsn;
	tag	fmtsn;
	int	flength;
	short	aformat;
	short	nchannels;
	int	srate;
	int	byrate;
	short	balign;
	short	bps;
	short	extra;
	short	vbps;
	int	channelmask;
	short	format;
	char	guid[14];
	tag	factsn;
	int	factlen;
	int	factdata;
	tag	datasn;
	int	datalen;

	wavh(int wavtype = 1, int asrate = baseSRATE) :
		riffsn("RIFF"), rlength(0), wavsn("WAVE"), fmtsn("fmt "),
		flength(40), aformat(1), nchannels(wav0_nchannels), srate(asrate),
		byrate(wav0_nchannels*asrate*wav0_bits_per_sample/8),
		balign(wav0_nchannels*wav0_bits_per_sample/8),
		bps(wav0_bits_per_sample), extra(22), vbps(wav0_bits_per_sample),
		channelmask(0x3),
		format(1), 
		factsn("fact"), factlen(4), factdata(0),
		datasn("data"), datalen(0)
	{
		guid[0] = 0x00;
		guid[1] = 0x00;
		guid[2] = 0x00;
		guid[3] = 0x00;
		guid[4] = 0x10;
		guid[5] = 0x00;
		guid[6] = 0x80;
		guid[7] = 0x00;
		guid[8] = 0x00;
		guid[9] = 0xaa;
		guid[10] = 0x00;
		guid[11] = 0x38;
		guid[12] = 0x9b;
		guid[13] = 0x71;
		
/*
		guid[0] = 0x0;
		guid[1] = 0x0;
		guid[2] = 0x10;
		guid[3] = 0x00;
		guid[4] = 0x80;
		guid[5] = 0x00;
		guid[6] = 0x00;
		guid[7] = 0xaa;
		guid[8] = 0x00;
		guid[9] = 0x38;
		guid[10] = 0x9b;
		guid[11] = 0x71;
*/
// "\x00\x00\x00\x00\x10\x00\x80\x00\x00\xAA\x00\x38\x9B\x71"
		//
		// Default above is signed integer 16 bit, type 3 is fp 32bit
		//
		if (wavtype == 3)
		{
			aformat = 3;
			format = 3;
			bps = 32;
			vbps = 32;
			byrate = nchannels * srate * (bps / 8);
			balign = nchannels * (bps / 8);
		}
		else if (wavtype == -2)
		{
			aformat = -2;
			format = 1;
			vbps = 24;
			bps = 24;
			byrate = nchannels * srate * (bps / 8);
			balign = nchannels * (bps / 8);
		}
	}

// header type: fmt , size: 18 found
// header type: fact, size: 4 found
// Wave file data found at 58, size: 153904344, fmt at: 20, wav header:
// riffsn: "RIFF", rlength: 153904394, wavsn: "WAVE", fmtsn: "RIFF", flength: 153904394
//   audioformat: 3, nchannels: 2, srate: 96000, byrate: 768000, balign: 8
//   bps: 32, datasn: "    ", datalen: 1996488706

	void
	setfsize(int nsamples, int length)
	{
		int dlength;
		dlength = length - sizeof(wavh);
		datalen = dlength;

		rlength = length - 8;
		factdata = nsamples * nchannels;
	}


	void
	snprint(char *buf, int bufsize)
	{
		snprintf(buf, bufsize,
			"riffsn: \"%4.4s\", rlength: %d, wavsn: \"%4.4s\", fmtsn: \"%4.4s\", flength: %d\n"
			"  audioformat: %d, nchannels: %d, srate: %d, byrate: %d, balign: %d\n"
			"  bps: %d, extra: %d, datasn: \"%4.4s\", datalen: %d\n",
			riffsn.cdata, rlength, wavsn.cdata, fmtsn.cdata, flength,
			aformat, nchannels, srate, byrate,
			balign, bps, extra, datasn.cdata, datalen);
	}
};



//
// Find the audio data in a wav file.  Return the basic wav header
// in the structure pointed to by the 2nd arg.  Return the FP
// state pointing to the audio data, and also the location of the
// audio data is returned.  If an error, then return -1ll.
//
static int __attribute__((unused))
finddata(FILE *fp, size_t* filesize, size_t* datasize, size_t* fmtlen, char **fmtptr, size_t* factlen, char **factptr,
size_t* ds64len, char **ds64ptr, size_t *chnalen, char **chnaptr,
size_t* axmllen, char **axmlptr, size_t* bextlen, char **bextptr,
size_t* iXMLlen = nullptr, char **iXMLptr = nullptr)
{
	char tbuf[16384];
	char rawbuf[16384];
	int rawbptr = 0;
	bool bw64flag = false;

	if (filesize != nullptr)
		*filesize = size_t(0);
	if (datasize != nullptr)
		*datasize = size_t(0);

	if (fmtlen != nullptr)
		*fmtlen = size_t(0);
	if (fmtptr != nullptr)
		*fmtptr = nullptr;
	

	if (ds64len != nullptr)
		*ds64len = size_t(0);
	if (ds64ptr != nullptr)
		*ds64ptr = nullptr;

	if (chnalen != nullptr)
		*chnalen = size_t(0);
	if (chnaptr != nullptr)
		*chnaptr = nullptr;

	if (axmllen != nullptr)
		*axmllen = size_t(0);
	if (axmlptr != nullptr)
		*axmlptr = nullptr;

	if (bextlen != nullptr)
		*bextlen = size_t(0);
	if (bextptr != nullptr)
		*bextptr = nullptr;
	
	if (iXMLlen != nullptr)
		*iXMLlen = size_t(0);
	if (iXMLptr != nullptr)
		*iXMLptr = nullptr;

	
//
// rewind the file (Ignored for pipes)
//
	fseek(fp, 0L, SEEK_SET);

//
// read the file header
//
	int nbr = fread(tbuf, 1, sizeof (rwav0), fp);

	if (nbr < ((int) sizeof (rwav0)))
	{
		fprintf(stderr, "finddata: file too small for a wav file\n");
		return -1L;
	}

	memcpy(&rawbuf[rawbptr], tbuf, nbr);
	rawbptr += nbr;

	//
	// First part of wav header
	//

	if (!memcmp(&rawbuf[offsetof(wavh, riffsn)], "OggS", 4))
	{
		fprintf(stderr, "wav file type is OggS -- error\n");
		return -2;
	}

	if (memcmp(&rawbuf[offsetof(wavh, riffsn)], "RIFF", 4))
	{
		if (memcmp(&rawbuf[offsetof(wavh, riffsn)], "BW64", 4) && memcmp(&rawbuf[offsetof(wavh, riffsn)], "RF64", 4))
		{
			fprintf(stderr, "wav file type isn't riff or bw64/rf64 -- error\n");
			return -1;
		}
		if (ac.info > 0x10) fprintf(stderr,"BW64 or RF64 file type\n");
		bw64flag = true;
	}
	else
	{
		rwav0 *rwp = (rwav0 *)&rawbuf[0];
		*filesize = rwp->rlength;
		if (ac.info > 0x10) fprintf(stderr,"file size: %ld\n", *filesize);
	}

	if (memcmp(&rawbuf[offsetof(wavh, wavsn)], "WAVE", 4))
	{
		fprintf(stderr, "%s file type is not \"WAVE\" -- error\n",
			bw64flag ? "BW64" : "RIFF");
		return -1;
	}

	tagsize ts;
		
	while (true)
	{
		bool fmtflag = false;
		bool ds64flag = false;
		bool chnaflag = false;
		bool axmlflag = false;
		bool iXMLflag = false;
		bool bextflag = false;
		bool dataflag = false;
		bool factflag = false;
		if (( nbr = fread(&ts, 1, sizeof ts, fp)) != sizeof ts)
		{
			fprintf(stderr, "Cannot read data tag in wav file\n");
			return -1;
		}

#if 1
		if (ac.info > 0x10) fprintf(stderr,"header type: %4.4s, size: %d found\n", ts.name.cdata, ts.len);
#endif

		if (!memcmp(ts.name.cdata, "fmt ", 4))
		{
			fmtflag = true;
		}
		if (!memcmp(ts.name.cdata, "fact", 4))
		{
			factflag = true;
		}
		else if (!memcmp(ts.name.cdata, "ds64", 4))
		{
			ds64flag = true;
		}
		else if (!memcmp(ts.name.cdata, "chna", 4))
		{
			chnaflag = true;
		}
		else if (!memcmp(ts.name.cdata, "axml", 4))
		{
			axmlflag = true;
		}
		else if (!memcmp(ts.name.cdata, "bext", 4))
		{
			bextflag = true;
		}
		else if (!memcmp(ts.name.cdata, "iXML", 4))
		{
			iXMLflag = true;
		}
		else if (!memcmp(ts.name.cdata, "data", 4))
		{
			dataflag = true;
			if (!bw64flag && (datasize != nullptr))
			{
				*datasize = wuint64_t(ts.len);
			if (ac.info > 0x10) 	fprintf(stderr,"tag: old data, size: %lu\n", wuint64_t(ts.len));
			}
			else
			{
				if (ac.info > 0x10) fprintf(stderr,"tag: new data, size: %lu\n", *datasize);
			}
			break;
		}

		//
		// Hack alert below
		//
		if (ts.len == 0 || ts.len > 1024*1024*1024)
		{
#if 0
			fprintf(stderr,"probable data found: \"%4.4s\"\n", ts.name.cdata);
#endif
			*datasize = -1;
			break;
		}
		char * tb = (char *) malloc(size_t(ts.len + 128));
		memset(tb, 0, ts.len + 128);
		char * tbp = tb;

		memcpy((void *)tb, (void *)&ts, sizeof ts);
		tbp += sizeof ts;


		//int needtoread = fmtflag ? (sizeof (fmtstruct0) - sizeof ts): ts.len;
		int needtoread = ts.len;
		int toread = needtoread;

		bool stopcopy = false;
		while (needtoread > 0)
		{
			nbr = fread(tbp, 1, needtoread, fp);
			if (nbr != needtoread)
			{
				fprintf(stderr, "Data read skipped off end of wav file -- error\n");
				return -1;
			}

			tbp += nbr;
			needtoread -= nbr;
		}

		if (fmtflag)
		{
			if (fmtptr != nullptr)
				*fmtptr = (char *) tb;
			if (fmtlen != nullptr)
				*fmtlen = ts.len;
				//*fmtlen = toread;
			((fmtstruct *)tb)->printf(stderr);
			if (ac.info > 0x10) fprintf(stderr,"tag: fmt, size: %lu, toread: %u\n", *fmtlen, toread);
		}
		else if (factflag)
		{
			if (factptr != nullptr)
				*factptr = (char *) tb;
			if (factlen != nullptr)
				*factlen = ts.len;
			if (ac.info > 0x10) fprintf(stderr,"tag: fact, size: %lu\n", *factlen);
		}
		else if (ds64flag)
		{
			if (ds64ptr != nullptr)
				*ds64ptr = (char *) tb;
			if (ds64len != nullptr)
				*ds64len = ts.len;

			ds64raw* drp = (ds64raw*) tb;
			wuint64_t fs = (wuint64_t(drp->bw64sizehigh) << 32) + drp->bw64sizelow;
			wuint64_t ds = (wuint64_t(drp->datasizehigh) << 32) + drp->datasizelow;
			if (filesize != nullptr)
				*filesize = fs;
			if (datasize != nullptr)
				*datasize = ds;
			if (ac.info > 0x10) fprintf(stderr,"tag: ds64, size: %lu, filesize: %lu, datasize: %lu\n", *ds64len, *filesize, *datasize);
			
		}
		else if (chnaflag)
		{
			if (chnaptr != nullptr)
				*chnaptr = (char *) tb;
			if (chnalen != nullptr)
				*chnalen = ts.len;
			if (ac.info > 0x10) fprintf(stderr,"tag: chna, size: %lu\n", *chnalen);
		}
		else if (axmlflag)
		{
			if (axmlptr != nullptr)
				*axmlptr = (char *) tb;
			if (axmllen != nullptr)
				*axmllen = ts.len;
			if (ac.info > 0x10) fprintf(stderr,"tag: axml, size: %lu\n", *axmllen);
		}
		else if (iXMLflag)
		{
			if (iXMLptr != nullptr)
				*iXMLptr = (char *) tb;
			if (iXMLlen != nullptr)
				*iXMLlen = ts.len;
			if (ac.info > 0x10) fprintf(stderr,"tag: iXML, size: %lu\n", *iXMLlen);
		}
		else if (bextflag)
		{
			if (bextptr != nullptr)
				*bextptr = (char *) tb;
			if (bextlen != nullptr)
				*bextlen = ts.len;
			if (ac.info > 0x10) fprintf(stderr,"tag: bext, size: %lu\n", *bextlen);
		}
		else if (dataflag)
		{
			if (!bw64flag && (datasize != nullptr))
			{
				*datasize = wuint64_t(ts.len);
				if (ac.info > 0x10) fprintf(stderr,"tag: old data, size: %lu\n", wuint64_t(ts.len));
			}
			else
			{
				if (ac.info > 0x10) fprintf(stderr,"tag: new data, size: %lu\n", *datasize);
			}
			break;
		}
		else
		{
			if (ac.info > 0x10) fprintf(stderr,"tag (unknown): %4.4s, size: %lu\n", ts.name.cdata, (uint64_t) ts.len);
			if (tb != nullptr)
				::free(tb);
			tb = nullptr;
		}
	}

	return rawbptr;
}


